//! Utility for converting and packing game assets for use with the engine.
//! Run with no arguments for usage information.

extern crate engine;

use engine::{MeshData, pretty_print};
use engine::ffi_sdl::{
	SDL_RWFromFile, SDL_LoadBMP_RW, SDL_ConvertSurfaceFormat,
	SDL_PIXELFORMAT_RGBA8888, SDL_FreeSurface,
};
use std::convert::TryFrom;
use std::ffi::CString;
use std::fs::File;
use std::io::{BufReader, Error, ErrorKind, Result, Write};
use std::os::raw::c_char;

fn main() {
	let mut iter = std::env::args();
	let binary = iter.next().unwrap_or_else(|| String::from("program"));
	let in_file_maybe = iter.next();
	let out_file_maybe = iter.next();
	if in_file_maybe.is_none() || out_file_maybe.is_none() || iter.next().is_some() {
		println!("Usage: {} in_file out_file", binary);
		println!("Convert asset files to internal formats convenient for embedding in binaries.");
		println!("Supported formats:");
		println!("- .bmp (RGBA image, a power of two transparent texture");
		println!("- .obj (3D mesh, triangles with texture coordinates)");
		println!("- .stl (3D mesh, just triangles)");
		return;
	}
	let in_file = in_file_maybe.unwrap();
	let out_file = out_file_maybe.unwrap();

	let extension = in_file.rsplit('.').next();
	match extension {
		Some("obj") => convert_obj(&in_file, &out_file).unwrap(),
		Some("stl") => convert_stl(&in_file, &out_file).unwrap(),
		Some("bmp") => convert_bmp(&in_file, &out_file).unwrap(),
		Some(extension) => eprintln!("Unrecognised file extension {}", extension),
		None => eprintln!("Input file has no extension, unable to detect file format"),
	}
}

fn convert_bmp(in_file: &str, out_file: &str) -> Result<()> { unsafe {
	// Load image.
	let c_file = CString::new(in_file).unwrap();
	// This is an expansion of the `SDL_LoadBMP()` macro.
	let rwops = SDL_RWFromFile(c_file.as_ptr(), b"rb\0".as_ptr() as *const c_char);
	let surface = SDL_LoadBMP_RW(rwops, 1);
	if surface.is_null() {
		return Err(Error::new(ErrorKind::InvalidData, "failed to load bitmap"));
	}
	let rgba_surface = SDL_ConvertSurfaceFormat(surface, SDL_PIXELFORMAT_RGBA8888, 0);
	assert!(!rgba_surface.is_null());

	// Copy data, reversing rows from top-to-bottom into bottom-to-top.
	let planes = 4;
	let width = (*rgba_surface).w as usize;
	let height = (*rgba_surface).h as usize;
	let pitch = (*rgba_surface).pitch as usize;
	assert!(pitch > 0);
	assert!(pitch >= width*planes);
	let pixels = std::slice::from_raw_parts((*rgba_surface).pixels as *const u8, width*height*planes);
	let row = width*planes;
	let mut rgba = vec![0 as u8; width*height*planes].into_boxed_slice();
	let mut from = 0;
	let mut to = row*height;
	while to > 0 {
		to -= row;
		rgba[to..to+row].copy_from_slice(&pixels[from..from+row]);
		from += pitch;
	}
	debug_assert!(from == pitch*height);
	// Reverse chanels in each pixel because SDL seems to use ABGR.
	rgba.chunks_exact_mut(4).for_each(|pixel| pixel.reverse());

	// Free manual allocations.
	SDL_FreeSurface(rgba_surface);
	SDL_FreeSurface(surface);

	// Write header.
	if !(width.is_power_of_two() && height.is_power_of_two()) {
		return Err(Error::new(ErrorKind::InvalidData, "image dimensions are not power of two"));
	}
	let exponents = [
		u8::try_from(width.trailing_zeros()).unwrap(),
		u8::try_from(height.trailing_zeros()).unwrap(),
	];
	let mut file = File::create(out_file)?;
	file.write_all(&exponents)?;

	// Helper to encode runs of colour values.
	let write_run = |file: &mut File, run: u32, value: u8| -> Result<()> {
		debug_assert!(run > 0);
		if run == 1 && value&0b11000000 == 0 {
			file.write_all(&[value])
		} else {
			file.write_all(&[
				0b10000000|u8::try_from(run-1).unwrap(),
				value,
			])
		}
	};

	// Write run length encoded colour planes.
	for offset in 0..planes {
		let mut index = offset;
		let mut value = None;
		let mut run = 0 as u32;
		while index < rgba.len() {
			if Some(rgba[index]) == value && run < 0b01000000 {
				run += 1;
			} else {
				if let Some(value) = value {
					write_run(&mut file, run, value)?;
				}
				value = Some(rgba[index]);
				run = 1;
			}
			index += planes;
		}
		write_run(&mut file, run, value.unwrap())?;
	}

	Ok(())
} }

fn convert_obj(in_file: &str, out_file: &str) -> Result<()> {
	let file = File::open(in_file)?;
	let mut reader = BufReader::new(file);
	let data = MeshData::read_obj(&mut reader)?;
	print_mesh_info(&data);
	let mut out_file = File::create(out_file)?;
	data.write(&mut out_file)?;
	Ok(())
}

fn convert_stl(in_file: &str, out_file: &str) -> Result<()> {
	let file = File::open(in_file)?;
	let mut reader = BufReader::new(file);
	let data = MeshData::read_stl(&mut reader)?;
	print_mesh_info(&data);
	let mut out_file = File::create(out_file)?;
	data.write(&mut out_file)?;
	Ok(())
}

fn print_mesh_info(data: &MeshData) {
	println!("{} triangles", pretty_print::Thousands(data.indices.len() as u64/3));
	println!("{} vertices", pretty_print::Thousands(data.vertices.len() as u64));
}
