//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use super::ffi_gles::*;
use super::window::Window;

/// Different ways of combining colours into those already shown.
/// To draw transparent images "normally" you probably want `Alpha`.
/// `PreAlpha` may be slightly more efficient but makes interpolation awkward
/// and requires textures to be pre-mulitplied before uploading.
#[derive(Clone, Copy, Debug, Eq, PartialEq)]
pub enum Blend {
	/// Additive, always makes the displayed value brighter.
	/// Useful for lighting and particle effects.
	Add,
	/// Over, normal transparent blending.
	Alpha,
	/// Multiplicative, always makes the displayed value darker.
	/// Useful for shadowing effects.
	Multiply,
	/// Alpha with pre-multiplied colours.
	/// This saves a multiply during blending, but requires the texture's red,
	/// green, and blue channels to be multiplied by the alpha channel before
	/// uploading.
	PreAlpha,
	/// Ignored the colour already on the display.
	/// This does nothing as it is the default mode you get if you do nothing.
	Replace,
}
impl Blend {
	/// Set blend mode until the next call to this function.
	pub fn draw(&self, window: &mut Window) { unsafe {
		let gl = window.gl();
		match self {
			Blend::Add => {
				(gl.Enable)(GL_BLEND);
				(gl.BlendFunc)(GL_ONE, GL_ONE);
			},
			Blend::Alpha => {
				(gl.Enable)(GL_BLEND);
				(gl.BlendFunc)(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			},
			Blend::Multiply => {
				(gl.Enable)(GL_BLEND);
				(gl.BlendFunc)(GL_ZERO, GL_SRC_COLOR);
			},
			Blend::PreAlpha => {
				(gl.Enable)(GL_BLEND);
				(gl.BlendFunc)(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
			},
			Blend::Replace => {
				(gl.Disable)(GL_BLEND);
			},
		}
		gl.check();
	} }
}
