//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use super::mesh_data::MeshData;
use super::vector3::{F3, f3};
use super::vertex::Index;

pub struct CollisionMesh {
	edges: Box<[Edge]>,
	triangles: Box<[Triangle]>,
	vertices: Box<[Vertex]>,
}
impl CollisionMesh {
	/// Generate collision data for a static mesh.
	pub fn new(data: &MeshData) -> Self {
		debug_assert!(data.indices.len()%3 == 0);
		let mut triangles = Vec::with_capacity(data.indices.len()/3);
		let mut edges = Vec::with_capacity(triangles.len()*3);
		for triangle in data.indices.chunks_exact(3) {
			let indices = [
				triangle[0],
				triangle[1],
				triangle[2],
			];
			let vertices = [
				F3::from_array(data.vertices[triangle[0] as usize].position),
				F3::from_array(data.vertices[triangle[1] as usize].position),
				F3::from_array(data.vertices[triangle[2] as usize].position),
			];
			triangles.push(Triangle::new(vertices, indices));
			edges.push((triangle[0], triangle[1]));
			edges.push((triangle[1], triangle[2]));
			edges.push((triangle[2], triangle[0]));
		}
		edges.sort_unstable_by(|a, b| a.cmp(b));
		edges.dedup();
		let edges = edges.iter().copied()
			.map(|(a, b)| {
				let a_data = &data.vertices[a as usize];
				let b_data = &data.vertices[b as usize];
				Edge::new(
					[F3::from_array(a_data.position), F3::from_array(b_data.position)],
					[F3::from_array(a_data.normal), F3::from_array(b_data.normal)],
					[a, b],
				)
			})
			.collect::<Vec<Edge>>()
			.into_boxed_slice();
		Self {
			edges,
			triangles: triangles.into_boxed_slice(),
			vertices: data.vertices.iter()
				.map(|vertex| Vertex {
					position: F3::from_array(vertex.position),
					normal: F3::from_array(vertex.normal),
				})
				.collect::<Vec<Vertex>>()
				.into_boxed_slice(),
		}
	}
	/// Test a ray against the mesh.
	pub fn ray(&self, start: F3, direction: F3, range: Option<f32>) -> Option<RayHit> {
		let mut best = None;
		let mut best_time = 0.0;
		for cache in self.triangles.iter() {
			if let Some(ray_hit) = cache.ray(start, direction, range) {
				let time = ray_hit.time;
				if best.is_none() || time < best_time {
					best = Some(ray_hit);
					best_time = time;
				}
			}
		}
		best
	}
	/// Test a sphere against the surface of the mesh.
	pub fn sphere(&self, centre: F3, radius: f32, scratch: &mut Vec<F3>) -> Option<SphereHit> {
		scratch.clear();
		let radius2 = radius*radius;
		let mut direction = f3(0.0, 0.0, 0.0);
		let mut smooth = f3(0.0, 0.0, 0.0);
		for hit in self.triangles.iter().map(|triangle| triangle.sphere(&self.vertices, centre, radius))
			.chain(self.edges.iter().map(|edge| edge.sphere(&self.vertices, centre, radius2)))
			.chain(self.vertices.iter().map(|vertex| vertex.sphere(centre, radius2)))
		{
			if let Some(hit) = hit {
				scratch.push(hit.direction*hit.distance);
				direction += hit.direction;
				smooth += hit.smooth;
			}
		}
		if scratch.len() == 0 { return None; }
		let scale = (scratch.len() as f32).recip();
		let direction = direction*scale;
		let smooth = (smooth*scale).unit_zero();
		let mut distance = 0.0;
		for separate in scratch.iter().copied() {
			let bound = separate.dot(direction);
			if bound > distance { distance = bound; }
		}
		Some(SphereHit {
			direction,
			distance,
			smooth,
		})
	}
}

/// Information about a ray-mesh intersection.
pub struct RayHit {
	/// Ray direction.
	pub direction: F3,
	/// Where to find the original vertices in the mesh.
	/// This is useful for interpolating vertex attributes.
	pub indices: [Index; 3],
	/// Normal of face hit (unit vector).
	/// You may want to interpolate vertex normals instead if simulating a smooth
	/// surface (you'll need to use `u` and `v` to do this).
	pub normal: F3,
	/// Start of ray.
	pub start: F3,
	/// Time to intersection along ray (also distance if ray has length of one).
	pub time: f32,
	/// Barycentric coordinates of contact.
	/// These are a trio of values all >= 0.0 and <= 1.0 that sum to 1.0.
	/// We only store two here because the third can be easily calculated:
	/// `w = 1.0-u-v`
	/// They give the amount of effect each respective vertex has on the
	/// attributes of the point of intersection.
	/// This is useful for interpolating attributes like vertex normals for
	/// smooth surfaces.
	/// `u` corresponds to `indices[1]`, `v` corresponds to `indices[2]`.
	pub u: f32,
	/// See `u`.
	pub v: f32,
}
impl RayHit {
	/// Return position of intersection in local coordinates.
	pub fn point(&self) -> F3 { self.start+self.direction*self.time }
	/// Return interpolated vertex normal for point.
	/// This may be useful for simulating smooth surfaces.
	pub fn smooth(&self, mesh: &CollisionMesh) -> F3 {
		let a = mesh.vertices[self.indices[0] as usize].normal;
		let b = mesh.vertices[self.indices[1] as usize].normal;
		let c = mesh.vertices[self.indices[2] as usize].normal;
		(a+(b-a)*self.u+(c-a)*self.v).unit_zero()
	}
}

/// Information about a sphere-mesh intersection.
pub struct SphereHit {
	/// Distance to move sphere to separate collision (unit vector).
	pub distance: f32,
	/// Direction to move sphere to separate collision.
	pub direction: F3,
	/// Interpolated surface normal at point of contact (unit vector).
	pub smooth: F3,
}

/// Pre-calculated collision data for a triangle.
/// This is used for ray and sphere queries.
/// Ray queries use Havel and Herout's algorithm.
/// Sphere queries are implemented as modified ray queries, and rely on edge
/// and vertex cases being handled separately.
/// See: Yet Faster Ray-Triangle Intersection (Using SSE4) (Havel and Herout, 2010)
/// See: https://stackoverflow.com/questions/13163129/ray-triangle-intersection
struct Triangle {
	/// Vertex indices for looking up original data.
	indices: [Index; 3],
	/// `normals[0].length().recip()` to make sphere intersections tidier and
	/// let us return normals as unit vectors cheaply.
	normal_recip: f32,
	/// Normal for triangle and two edge-planes scaled for use in Havel and Herout.
	/// AKA N, N1, N2 in the paper.
	/// Note that the triangle's normal (element zero) points inward and is not a
	/// unit vector.
	normals: [F3; 3],
	/// Constants combined with `normals` to describe planes.
	/// Called d, d1, d2 in the paper.
	offsets: [f32; 3],
}
impl Triangle {
	pub fn new(vertices: [F3; 3], indices: [Index; 3]) -> Self {
		let u = vertices[1]-vertices[0]; // AKA A->B
		let v = vertices[2]-vertices[0]; // AKA A->C
		let normal = u.cross(v);
		let length2 = normal.length2();
		assert!(length2 > 0.0, "degenerate (flat) triangle in collision mesh");
		let recip_length2 = length2.recip();
		let n1 = v.cross(normal)*recip_length2;
		let n2 = normal.cross(u)*recip_length2;
		Self {
			indices,
			normal_recip: recip_length2.sqrt(),
			normals: [normal, n1, n2],
			offsets: [
				vertices[0].dot(normal),
				-vertices[0].dot(n1),
				-vertices[0].dot(n2),
			],
		}
	}
	pub fn ray(&self, start: F3, direction: F3, range: Option<f32>) -> Option<RayHit> {
		// This algorithm is built around these concepts:
		//
		// Planes can be described as a normal and an offset.
		// Point `p` is on the plane with normal `n` and offset `d` when:
		// p·n = d
		// 
		// Barycentric coordinates make it easy to tell positions inside a triangle.
		// They can be calculated by projecting (dot product) a cartesian point onto
		// directions perpendicular to the edge of the triangle and dividing by the
		// distance of the point opposite that face.
		//
		// Combining projecting the ray onto the triangle's normal and onto vectors
		// perpendicular to the edges of the triangle (but in its plane) allows
		// finding the point where the ray hits the triangle's plane, and from that
		// the barycentric coordinates of the hit.
		//
		// This is made more efficient by rearranging the calculations so that the
		// division needed in finding the time from its start that the ray hits the
		// plane happens as late as possible (it can often be skipped as a result).
		// All of the vectors used for projection (`self.normals`) are also scaled
		// so that the dot product does both the projection and division steps of
		// the conversion to barycentric coordinates.

		// Every `almost_<value>` is `<value>*determinant`, so we can check for
		// collision without dividing (which is expensive).
		let determinant = direction.dot(self.normals[0]);
		let limit = 0.000001; // complete guess
		if determinant.abs() < limit {
			// Ray parallel to triangle, likely to cause division by zero.
			return None;
		}
		let almost_time = self.offsets[0]-start.dot(self.normals[0]); // AKA t'
		if almost_time.is_sign_positive() != determinant.is_sign_positive() {
			// Hit before start of ray.
			return None;
		}
		debug_assert!(range.is_none() || range.unwrap() > 0.0);
		if range.is_some() && almost_time.abs() > range.unwrap()*determinant.abs() {
			// Hit after end of ray.
			return None;
		}
		let almost_point = start*determinant+direction*almost_time; // AKA P'
		let almost_u = almost_point.dot(self.normals[1])+determinant*self.offsets[1]; // AKA u'
		let almost_v = almost_point.dot(self.normals[2])+determinant*self.offsets[2]; // AKA v'
		if almost_u.is_sign_positive() != determinant.is_sign_positive() ||
			almost_v.is_sign_positive() != determinant.is_sign_positive() ||
			(almost_u+almost_v).abs() > determinant.abs()
		{
			// Hit outside of triangle.
			// `u` and `v` are barycentric coordinates, requiring the following for
			// inside:
			// - `u >= 0.0`
			// - `v >= 0.0`
			// - `u+v <= 1.0`
			// We can compensate for the multiplication by `determinant` by checking
			// signs for the first two cases, and multiplying the right hand side for
			// the second.
			return None;
		}
		// Remove the multiply by `determinant` now we know we hit the triangle.
		let determinant_recip = determinant.recip();
		let time = almost_time*determinant_recip;
		debug_assert!(time >= 0.0);
		debug_assert!(range.is_none() || time <= range.unwrap());
		let u = almost_u*determinant_recip;
		debug_assert!(u >= 0.0);
		let v = almost_v*determinant_recip;
		debug_assert!(v >= 0.0);
		debug_assert!(u+v <= 1.0);
		Some(RayHit {
			direction,
			indices: self.indices,
			normal: -self.normals[0]*self.normal_recip,
			start,
			time,
			u,
			v,
		})
	}
	pub fn sphere(&self, vertices: &[Vertex], centre: F3, radius: f32) -> Option<SphereHit> {
		// This is a modified verysion of `ray()` where we make the ray direction
		// the triangle's normal and choose a length that makes `determinant` equal
		// one.
		// This means we can skip quite a lot of operations.
		// This works because we're only interested in cases where the sphere touches
		// the face of the triangle (not the edges or vertices), and in that case the
		// contact point will always be on a line from the centre of the sphere in
		// the direction of the normal.

		// let direction = self.normals[0]/self.normals[0].length2();
		// let determinant = self.normals[0].dot(direction); assert!(determinant == 1.0);
		// let almost_time = self.offsets[0]-centre.dot(self.normals[0]);
		// let time = almost_time/determinant;
		let time = self.offsets[0]-centre.dot(self.normals[0]);
		let distance = time*self.normal_recip;
		if distance.abs() > radius {
			// Sphere not intersecting plane of triangle.
			return None;
		}
		// let almost_point = centre*determinant+self.normals[0]*almost_time;
		// let almost_u = almost_point.dot(self.normals[1])+determinant*self.offsets[1];
		// let almost_v = almost_point.dot(self.normals[2])+determinant*self.offsets[2];
		let point = centre+self.normals[0]*time;
		let u = point.dot(self.normals[1])+self.offsets[1];
		let v = point.dot(self.normals[2])+self.offsets[2];
		if u < 0.0 || v < 0.0 || u+v > 1.0 {
			// Sphere outside edges of triangle.
			return None;
		}
		// Interpolate surface normal.
		let a = vertices[self.indices[0] as usize].position;
		let b = vertices[self.indices[1] as usize].position;
		let c = vertices[self.indices[2] as usize].position;
		let smooth = (a+(b-a)*u+(c-a)*v).unit_zero();
		Some(SphereHit {
			direction: -self.normals[0]*self.normal_recip,
			distance: distance+radius,
			smooth,
		})
	}
}

/// Pre-calculated collision data for an edge.
/// This is used for sphere queries.
pub struct Edge {
	/// Vertex indices.
	indices: [Index; 2],
	/// Direction of edge, average of vertex normals, tangent.
	normals: [F3; 3],
	/// Offset of start of edge to match `normals`.
	offsets: [f32; 3],
}
impl Edge {
	pub fn new(vertices: [F3; 2], normals: [F3; 2], indices: [Index; 2]) -> Self {
		let normal = (normals[0]+normals[1]).unit_zero();
		debug_assert!(normal.length2() > 0.0, "edge with zero normal");
		debug_assert!(vertices[0] != vertices[1], "edge with length zero");
		let direction = (vertices[1]-vertices[0]).unit_zero();
		let tangent = normal.cross(direction).unit_zero();
		debug_assert!(tangent.length2() > 0.0, "edge with same normal as direction");
		Self {
			indices,
			normals: [direction, normal, tangent],
			offsets: [vertices[0].dot(direction), vertices[0].dot(normal), vertices[0].dot(tangent)],
		}
	}
	pub fn sphere(&self, vertices: &[Vertex], centre: F3, radius2: f32) -> Option<SphereHit> {
		// Push the sphere out along the edge normal.
		// This approximation is simpler than trying to reliably figure out the
		// exact shortest outward path, though we still need to push the sphere only
		// to the boundary of the intersection.
		// We do this by projecting into 2D looking along the edge.

		// Project centre onto edge to find position in range 0..1.
		let fraction = centre.dot(self.normals[0])-self.offsets[0];
		if fraction < 0.0 || fraction > 1.0 {
			// Beyond end of edge.
			return None;
		}
		// Position across tangent (start of 2D sub-problem).
		let u = centre.dot(self.normals[2])-self.offsets[2];
		let u2 = u*u;
		if u2 > radius2 {
			// Too far from edge.
			return None;
		}
		// Radius of sphere along normal.
		let slice2 = radius2-u2;
		// Position along normal.
		let v = centre.dot(self.normals[1])-self.offsets[1];
		let v2 = v*v;
		if v2 > slice2 {
			// Too far from edge.
			return None;
		}
		// Distance to separate.
		debug_assert!(u2+v2 <= radius2);
		eprintln!("TODO: why do we have to negate `v` to get the right output?");
		let distance = slice2.sqrt()+v;
		debug_assert!(distance >= 0.0);
		debug_assert!(distance*distance <= radius2*4.0);
		println!("f={} u={} v={} d={}", fraction, u, v, distance/radius2.sqrt());
		// Calculate interpolated surface normal.
		let smooth = vertices[self.indices[0] as usize].normal
			.lerp(vertices[self.indices[1] as usize].normal, fraction)
			.unit_zero();
		Some(SphereHit {
			direction: self.normals[1],
			distance,
			smooth,
		})
	}
}

/// Pre-calculated collision data for a vertex.
/// This is used for sphere queries.
pub struct Vertex {
	position: F3,
	normal: F3,
}
impl Vertex {
	pub fn sphere(&self, centre: F3, radius2: f32) -> Option<SphereHit> {
		// Separate along normal.

		// Check for contact.
		let delta = centre-self.position;
		let distance2 = delta.length2();
		if distance2 >= radius2 {
			return None;
		}
		// Find distance along normal.
		let v = delta.dot(self.normal);
		let v2 = v*v;
		debug_assert!(v2 <= radius2);
		// Use Pythagorous to find distance in plane described by normal and vertex.
		//           .centre
		// distance / |
		//         /  | v
		//   vertex---+
		//          u
		let u2 = distance2-v2;
		debug_assert!(u2 <= radius2);
		// Radius of sphere along normal.
		let slice = (radius2-u2).sqrt();
		// Distance to separate.
		let distance = slice-v;
		debug_assert!(distance >= 0.0);
		debug_assert!(distance*distance <= radius2*4.0);
		Some(SphereHit {
			direction: self.normal,
			distance,
			smooth: self.normal,
		})
	}
}

#[cfg(test)]
mod tests {
	use super::*;
	use crate::vector3::f3;

	fn equalish(a: f32, b: f32) -> bool {
		(a-b).abs() < 0.01
	}
	#[test]
	fn ray_triangle_intersect() {
		let triangles = [
			// (Looking along z+)
			// 2
			// |\
			// 0-1
			Triangle::new([f3(0.0, 0.0, 0.0), f3(1.0, 0.0, 0.0), f3(0.0, 1.0, 0.0)], [0, 0, 0]),
			//   2
			//  /|
			// 0-1
			Triangle::new([f3(0.0, 0.0, 0.0), f3(1.0, 0.0, 0.0), f3(1.0, 1.0, 0.0)], [0, 0, 0]),
			// 2-1
			//  \|
			//   0
			Triangle::new([f3(1.0, 0.0, 0.0), f3(1.0, 1.0, 0.0), f3(0.0, 1.0, 0.0)], [0, 0, 0]),
			// 2-1
			// |/
			// 0
			Triangle::new([f3(0.0, 0.0, 0.0), f3(1.0, 1.0, 0.0), f3(0.0, 1.0, 0.0)], [0, 0, 0]),
			//   0
			//  / \
			// 1---2
			Triangle::new([f3(0.5, 1.0, 0.0), f3(0.0, 0.0, 0.0), f3(1.0, 0.0, 0.0)], [0, 0, 0]),
		];
		let direction = f3(1.0, 1.0, 1.0);
		let point_uv = [
			// Vertices.
			(f3(0.0, 0.0, 0.0), [Some((0.0, 0.0)), Some((0.0, 0.0)), None, Some((0.0, 0.0)), Some((1.0, 0.0))]),
			(f3(1.0, 0.0, 0.0), [Some((1.0, 0.0)), Some((1.0, 0.0)), Some((0.0, 0.0)), None, Some((0.0, 1.0))]),
			(f3(1.0, 1.0, 0.0), [None, Some((0.0, 1.0)), Some((1.0, 0.0)), Some((1.0, 0.0)), None]),
			(f3(0.0, 1.0, 0.0), [Some((0.0, 1.0)), None, Some((0.0, 1.0)), Some((0.0, 1.0)), None]),

			// Just outside.
			(f3(-0.1, -0.1, 0.0), [None, None, None, None, None]),
			(f3(1.1, -0.1, 0.0), [None, None, None, None, None]),
			(f3(1.1, 1.1, 0.0), [None, None, None, None, None]),
			(f3(-0.1, 1.1, 0.0), [None, None, None, None, None]),

			// Horizontal and vertical edges.
			(f3(0.5, 0.0, 0.0), [Some((0.5, 0.0)), Some((0.5, 0.0)), None, None, Some((0.5, 0.5))]),
			(f3(1.0, 0.5, 0.0), [None, Some((0.5, 0.5)), Some((0.5, 0.0)), None, None]),
			(f3(0.5, 1.0, 0.0), [None, None, Some((0.5, 0.5)), Some((0.5, 0.5)), Some((0.0, 0.0))]),
			(f3(0.0, 0.5, 0.0), [Some((0.0, 0.5)), None, None, Some((0.0, 0.5)), None]),

			// Centre.
			(f3(0.5, 0.5, 0.0), [Some((0.5, 0.5)), Some((0.0, 0.5)), Some((0.0, 0.5)), Some((0.5, 0.0)), Some((0.25, 0.25))]),
		];

		for (test_index, (point, uvs)) in point_uv.iter().copied().enumerate() {
			let start = point-direction;
			for ((index, triangle), uv) in triangles.iter().enumerate().zip(uvs.iter().copied()) {
				if let Some(hit) = triangle.ray(start, direction, None) {
					assert!(equalish(hit.time, 1.0));
					assert!(uv.is_some());
					let uv = uv.unwrap();
					println!("{} {} {},{} == {},{}", test_index, index, hit.u, hit.v, uv.0, uv.1);
					assert!(equalish(hit.u, uv.0));
					assert!(equalish(hit.v, uv.1));
					assert!(hit.u >= 0.0);
					assert!(hit.v >= 0.0);
					assert!(hit.u+hit.v <= 1.0);
				} else {
					assert!(uv.is_none());
					println!("{} {} None == None", test_index, index);
				}
			}
		}
	}
}
