//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use super::vector3::{f3, F3};

/// A scale and rotate transform implemented as a 3x3 matrix.
#[derive(Clone, Debug, PartialEq)]
pub struct Matrix3 {
	///A 3x3 matrix, stored in column-major order.
	/// If we draw the array indices in conventional notation, we get this:
	/// [0 3 6]
	/// [1 4 7]
	/// [2 5 8]
	pub matrix: [f32; 9],
}
impl Matrix3 {
	/// Return a transform that does nothing.
	pub fn identity() -> Self {
		Self{matrix: [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0]}
	}
	/// Return a rotation of `angle` radians around a unit vector `axis`.
	/// If you want to concatenate rotations it is likely to be better to use
	/// `Rotation` because they're faster to concatenate and normalise.
	/// TODO: clockwise or anti-clockwise?
	pub fn rotate(axis: F3, angle: f32) -> Self {
		debug_assert!({
			let length2 = axis.length2();
			length2 > 0.9 && length2 < 1.1
		}, "axis should be a unit vector");
		let (s, c) = angle.sin_cos();
		Self{matrix: [
			// Left column.
			c+axis.x*axis.x*(1.0-c), axis.y*axis.x*(1.0-c)+axis.z*s, axis.z*axis.x*(1.0-c)-axis.y*s,
			// Centre column.
			axis.x*axis.y*(1.0-c)-axis.z*s, c+axis.y*axis.y*(1.0-c), axis.z*axis.y*(1.0-c)+axis.x*s,
			// Right column.
			axis.x*axis.z*(1.0-c)+axis.y*s, axis.y*axis.z*(1.0-c)-axis.x*s, c+axis.z*axis.z*(1.0-c),
		]}
	}
	/// Return an independent scale along the x, y, and z axes.
	pub fn scale(scale: F3) -> Self {
		Self{matrix: [scale.x, 0.0, 0.0, 0.0, scale.y, 0.0, 0.0, 0.0, scale.z]}
	}
	/// Transform a vector.
	pub fn apply(&self, vector: F3) -> F3 {
		// A matrix is just a compact way to describe projecting a vector onto a new
		// set of axes.
		f3(
			vector.dot(f3(self.matrix[0], self.matrix[3], self.matrix[6])),
			vector.dot(f3(self.matrix[1], self.matrix[4], self.matrix[7])),
			vector.dot(f3(self.matrix[2], self.matrix[5], self.matrix[8])),
		)
	}
	/// Concatenate two matrices, the result will perform `self` then `next` when
	/// applied to a vector.
	pub fn then(&self, next: &Self) -> Self {
		// a[0]*b[0] + a[3]*b[1] + a[6]*b[2], a[0]*b[3] + a[3]*b[4] + a[6]*b[5], a[0]*b[6] + a[3]*b[7] + a[6]*b[8]
		// a[1]*b[0] + a[4]*b[1] + a[7]*b[2], a[1]*b[3] + a[4]*b[4] + a[7]*b[5], a[1]*b[6] + a[4]*b[7] + a[7]*b[8]
		// a[2]*b[0] + a[5]*b[1] + a[8]*b[2], a[2]*b[3] + a[5]*b[4] + a[8]*b[5], a[2]*b[6] + a[5]*b[7] + a[8]*b[8]
		Self{matrix: [
			next.matrix[0]*self.matrix[0] + next.matrix[3]*self.matrix[1] + next.matrix[6]*self.matrix[2],
			next.matrix[1]*self.matrix[0] + next.matrix[4]*self.matrix[1] + next.matrix[7]*self.matrix[2],
			next.matrix[2]*self.matrix[0] + next.matrix[5]*self.matrix[1] + next.matrix[8]*self.matrix[2],
			next.matrix[0]*self.matrix[3] + next.matrix[3]*self.matrix[4] + next.matrix[6]*self.matrix[5],
			next.matrix[1]*self.matrix[3] + next.matrix[4]*self.matrix[4] + next.matrix[7]*self.matrix[5],
			next.matrix[2]*self.matrix[3] + next.matrix[5]*self.matrix[4] + next.matrix[8]*self.matrix[5],
			next.matrix[0]*self.matrix[6] + next.matrix[3]*self.matrix[7] + next.matrix[6]*self.matrix[8],
			next.matrix[1]*self.matrix[6] + next.matrix[4]*self.matrix[7] + next.matrix[7]*self.matrix[8],
			next.matrix[2]*self.matrix[6] + next.matrix[5]*self.matrix[7] + next.matrix[8]*self.matrix[8],
		]}
	}
}
