//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use std::io::{BufRead, Error, ErrorKind, Read, Result, Write};
use std::convert::TryFrom;
use super::mesh::StaticMesh;
use super::vector3::{F3, f3};
use super::vertex::{Index, Primitive, Vertex};
use super::window::Window;

/// A container for a mesh in a standard-ish internal format.
#[derive(Clone)]
pub struct MeshData {
	pub indices: Box<[Index]>,
	pub primitive: Primitive,
	pub vertices: Box<[Vertex]>,
}
impl MeshData {
	/// Create a test mesh containing a plane.
	pub fn test() -> Self {
		let vertices = vec![ // clockwise
			Vertex{position: [-1.0, -1.0, 0.0], texcoord: [0.0, 0.0], normal: [0.0, 0.0, -1.0]},
			Vertex{position: [-1.0, 1.0, 0.0], texcoord: [0.0, 1.0], normal: [0.0, 0.0, -1.0]},
			Vertex{position: [1.0, 1.0, 0.0], texcoord: [1.0, 1.0], normal: [0.0, 0.0, -1.0]},
			Vertex{position: [1.0, -1.0, 0.0], texcoord: [1.0, 0.0], normal: [0.0, 0.0, -1.0]},
		].into_boxed_slice();
		let indices = vec![ // counter-clockwise winding
			0, 2, 1,
			0, 3, 2,
		].into_boxed_slice();
		Self{indices, primitive: Primitive::Triangles, vertices}
	}
	/// Load an embedded file, meant for use with `include_bytes!()`.
	pub fn embedded(mut data: &[u8]) -> Self {
		Self::read(&mut data).unwrap()
	}
	/// Load a mesh from a file.
	/// Expected format:
	/// - vertices: u32 - number of vertices
	/// - [(F3, F2, F3)] - vertex data
	/// - indices: u32 - number of indices
	/// - indices: [u32] - vertex indices
	pub fn read<R: Read>(from: &mut R) -> Result<Self> {
		let vertex_count = read_u32(from)?;
		let mut vertices = Vec::with_capacity(vertex_count as usize);
		for _ in 0..vertex_count {
			let position = read_vec3(from)?;
			let texcoord = read_vec2(from)?;
			vertices.push(Vertex {
				position,
				texcoord,
				normal: [0.0; 3],
			});
		}
		let vertices = vertices.into_boxed_slice();

		let index_count = read_u32(from)?;
		let mut indices = Vec::with_capacity(index_count as usize);
		for _ in 0..index_count {
			let index = read_u32(from)?;
			if index >= vertex_count {
				return Err(Error::new(ErrorKind::InvalidData, "index beyond vertices"));
			}
			if index as usize > Index::MAX as usize {
				return Err(Error::new(ErrorKind::InvalidData, "index beyond maximum index value"));
			}
			indices.push(index as Index);
		}

		let indices = indices.into_boxed_slice();
		let mut out = Self{indices, primitive: Primitive::Triangles, vertices};
		out.calculate_normals();
		Ok(out)
	}
	/// Load an STL (see: https://en.wikipedia.org/wiki/STL_(file_format)) file.
	pub fn read_stl<R: Read>(from: &mut R) -> Result<Self> {
		// Skip header.
		from.read_exact(&mut [0 as u8; 80])?;
		// Number of triangles.
		let count = read_u32(from)?;
		// Triangle vertices.
		let vertex_count_guess = (count*2) as usize;
		let mut vertices = Vec::<Vertex>::with_capacity(vertex_count_guess);
		let mut indices = Vec::with_capacity(vertex_count_guess);
		for _ in 0..count {
			// Read triangle.
			let _normal = read_vec3(from)?;
			'vertex_loop: for _ in 0..3 {
				let position = read_vec3(from)?;
				// Remap axes.
				let position = [position[0], position[2], position[1]];
				// Look for previous use of vertex and convert to index.
				for index in 0..vertices.len() {
					if vertices[index].position == position {
						indices.push(Index::try_from(index).unwrap());
						continue 'vertex_loop;
					}
				}
				// Not used before, add to vertex buffer.
				if let Ok(index) = Index::try_from(vertices.len()) {
					indices.push(index);
				} else {
					return Err(Error::new(ErrorKind::InvalidData, "index beyond maximum index value"));
				}
				vertices.push(Vertex {
					position,
					texcoord: [0.0; 2],
					normal: [0.0; 3],
				});
			}
			let mut attributes = [0 as u8; 2];
			from.read_exact(&mut attributes)?;
		}
		// Clean up.
		let vertices = vertices.into_boxed_slice();
		let indices = indices.into_boxed_slice();
		let mut out = Self{indices, primitive: Primitive::Triangles, vertices};
		out.calculate_normals();
		Ok(out)
	}
	/// Load an obj (see: https://en.wikipedia.org/wiki/Wavefront_.obj_file) file.
	pub fn read_obj<R: BufRead>(from: &mut R) -> Result<Self> {
		let mut positions = Vec::new();
		let mut texcoords = Vec::new();
		positions.push([0.0 as f32; 3]); // fallback value, obj doesn't use index 0
		texcoords.push([0.0 as f32; 2]); // fallback value, obj doesn't use index 0
		let mut vertices = Vec::new();
		let mut indices = Vec::new();
		// Parse each line separately.
		for line in from.split(b'\n') {
			let line = line?;
			// Skip comments.
			if let Some(b'#') = line.first() { continue; }
			// Differentiate data by first word.
			let mut words = line.split(|byte| *byte == b' ').filter(|word| word.len() > 0);
			let command = if let Some(word) = words.next() { word } else { continue };
			match command {
				b"v" => {
					let mut position = [0.0 as f32; 3];
					parse_list(words, parse_f32, &mut position)?;
					position = [position[0], position[1], -position[2]];
					positions.push(position);
				},
				b"vt" => {
					let mut texcoord = [0.0 as f32; 2];
					parse_list(words, parse_f32, &mut texcoord)?;
					texcoords.push(texcoord);
				},
				b"f" => match (words.next(), words.next(), words.next(), words.next()) {
					(Some(a), Some(b), Some(c), None) => for word in [a, b, c].iter().copied() {
						// Parse list of data indices.
						let mut data_indices = [0 as u32; 2]; // position, texcoord
						let parse_index = |index: &[u8]| if index.len() == 0 { Ok(0) } else { parse_u32(index) };
						parse_list(word.split(|byte| *byte == b'/'), parse_index, &mut data_indices)?;
						// Lookup data.
						let vertex = Vertex {
							position: positions[data_indices[0] as usize],
							texcoord: texcoords[data_indices[1] as usize],
							normal: [0.0; 3],
						};
						// Deduplicate vertices.
						// We have to index into a list of complete vertices, so we can't use
						// the file's separate indices for position and texcoord.
						let index = if let Some(index) = vertices.iter().position(|check| *check == vertex) {
							index
						} else {
							vertices.push(vertex);
							vertices.len()-1
						};
						indices.push(Index::try_from(index)
							.map_err(|_| Error::new(ErrorKind::InvalidData, "index beyond maximum index value"))?);
					},
					_ => return Err(Error::new(ErrorKind::InvalidData, "non-triangle faces are not supported")),
				},
				b"vn" | b"vp" => (), // ignore
				command => eprintln!("WARNING: unsupported .obj command {}, ignoring", std::str::from_utf8(command).unwrap_or("<invalid UTF8>")),
			}
		}
		// Clean up.
		let vertices = vertices.into_boxed_slice();
		let indices = indices.into_boxed_slice();
		let mut out = Self{indices, primitive: Primitive::Triangles, vertices};
		out.calculate_normals();
		Ok(out)
	}
	/// Write to a file in the same format as `read()`.
	pub fn write<W: Write>(&self, to: &mut W) -> Result<()> {
		if self.primitive != Primitive::Triangles {
			return Err(Error::new(ErrorKind::InvalidData, "can only write triangle meshes"));
		}
		to.write_all(&u32::try_from(self.vertices.len()).unwrap().to_le_bytes())?;
		for vertex in self.vertices.iter() {
			to.write_all(&vertex.position[0].to_le_bytes())?;
			to.write_all(&vertex.position[1].to_le_bytes())?;
			to.write_all(&vertex.position[2].to_le_bytes())?;
			to.write_all(&vertex.texcoord[0].to_le_bytes())?;
			to.write_all(&vertex.texcoord[1].to_le_bytes())?;
		}
		to.write_all(&u32::try_from(self.indices.len()).unwrap().to_le_bytes())?;
		for index in self.indices.iter().copied() {
			to.write_all(&u32::from(index).to_le_bytes())?;
		}
		Ok(())
	}
	/// Average face windings to calculate vertex normals.
	pub fn calculate_normals(&mut self) {
		assert!(self.primitive == Primitive::Triangles);
		let mut normals = vec![f3(0.0, 0.0, 0.0); self.vertices.len()];
		for face in self.indices.chunks_exact(3) {
			let vertices = [
				F3::from_array(self.vertices[face[0] as usize].position),
				F3::from_array(self.vertices[face[1] as usize].position),
				F3::from_array(self.vertices[face[2] as usize].position),
			];
			// Assume triangles are counter-clockwise when viewed from outside.
			// `y.cross(x) == -z`
			if let Some(normal) = (vertices[2]-vertices[0]).cross(vertices[1]-vertices[0]).unit() {
				normals[face[0] as usize] += normal;
				normals[face[1] as usize] += normal;
				normals[face[2] as usize] += normal;
			} else {
				eprintln!("WARNING: triangle with zero-length edge");
			}
		}
		for (vertex, normal) in self.vertices.iter_mut().zip(normals.iter().copied()) {
			let normal = normal.unit_zero();
			if normal == f3(0.0, 0.0, 0.0) { eprintln!("WARNING: zero normal in mesh"); }
			vertex.normal = normal.to_array();
		}
	}
	pub fn to_mesh(&self, window: &mut Window) -> StaticMesh<Vertex> {
		StaticMesh::new(window, self.primitive, &self.indices, &self.vertices)
	}
}

/// Parse each item from an iterator using `f()` and populate `out` with the
/// results.
/// Stop at first parse error.
fn parse_list<'a, F: Fn(&'a [u8]) -> Result<T>, I: Iterator<Item=&'a [u8]>, T>(mut words: I, f: F, out: &mut [T]) -> Result<()> {
	for out in out.iter_mut() {
		if let Some(word) = words.next() {
			*out = f(word)?;
		} else {
			break;
		}
	}
	for word in words {
		f(word)?;
	}
	Ok(())
}
fn parse_f32(number: &[u8]) -> Result<f32> {
	let negative = number.first().copied() == Some(b'-');
	let number = if negative { &number[1..] } else { number };
	let dot = number.iter().position(|byte| *byte == b'.').unwrap_or(number.len());
	let mut value = parse_u32(&number[..dot])? as f32;
	if dot < number.len() {
		let fraction = &number[dot+1..];
		value += (parse_u32(fraction)? as f32)/(10.0 as f32).powi(fraction.len() as i32);
	}
	if negative { value = -value; }
	Ok(value)
}
fn parse_u32(number: &[u8]) -> Result<u32> {
	if number.first().copied() == Some(b'-') {
		return Err(Error::new(ErrorKind::InvalidData, "negative sign not supported here"));
	}
	let mut value = 0 as u32;
	if number.len() == 0 { return Err(Error::new(ErrorKind::InvalidData, "missing digits")); }
	for byte in number.iter().copied() {
		if byte >= b'0' && byte <= b'9' {
			value = value.checked_mul(10)
				.and_then(|value| value.checked_add((byte-b'0') as u32))
				.ok_or(Error::new(ErrorKind::InvalidData, "number too large"))?;
		} else {
			return Err(Error::new(ErrorKind::InvalidData, "invalid digit in number"));
		}
	}
	Ok(value)
}
fn read_u32<R: Read>(from: &mut R) -> Result<u32> {
	let mut buffer = [0 as u8; 4];
	from.read_exact(&mut buffer)?;
	Ok(u32::from_le_bytes(buffer))
}
fn read_f32<R: Read>(from: &mut R) -> Result<f32> {
	let mut bytes = [0 as u8; 4];
	from.read_exact(&mut bytes)?;
	Ok(f32::from_le_bytes(bytes))
}
fn read_vec3<R: Read>(from: &mut R) -> Result<[f32; 3]> {
	Ok([
		read_f32(from)?,
		read_f32(from)?,
		read_f32(from)?,
	])
}
fn read_vec2<R: Read>(from: &mut R) -> Result<[f32; 2]> {
	Ok([
		read_f32(from)?,
		read_f32(from)?,
	])
}
