//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use super::matrix3::Matrix3;
use super::vector3::{f3, F3};

/// A 3D rotation transform.
/// This is represented internally as a unit quaternion for compactness,
/// and efficiency of most operations.
#[derive(Clone, Copy, Debug)]
pub struct Rotation {
	// The real (scalar) part.
	pub a: f32,
	// The vector (imaginary) part, these three are respectively multiplied by i,
	// j, and k when written in mathematical notation.
	pub b: f32,
	pub c: f32,
	pub d: f32,
}
impl Rotation {
	/// Return a rotation of `angle` radians anti-clockwise looking in the
	/// direction of the unit vector `axis`.
	pub fn new(axis: F3, angle: f32) -> Self {
		debug_assert!({
			let length2 = axis.length2();
			length2 > 0.9 && length2 < 1.1
		}, "axis should be a unit vector");
		let (s, c) = (angle*0.5).sin_cos();
		Self{a: c, b: s*axis.x, c: s*axis.y, d: s*axis.z}
	}
	/// Return a rotation that does nothing.
	pub fn identity() -> Self {
		Self{a: 1.0, b: 0.0, c: 0.0, d: 0.0}
	}
	/// Return a rotation that points object's z axis along `forward`, and it's
	/// y points as near as possible to `up`.
	pub fn look(forward: F3, up: F3) -> Self {
		debug_assert!((forward.length2()-1.0).abs() < 0.01, "forward should be a unit vector");
		debug_assert!((up.length2()-1.0).abs() < 0.01, "up should be a unit vector");
		let z = f3(0.0, 0.0, 1.0);
		// Rotate to point local z along `forward`.
		// We need to rotate around an axis perpendicular to both our initial
		// direction (z) and our desired direction (`forward`).
		// If those are the same (or exactly opposite directions) we need to either
		// do nothing or rotate 180° around an axis perpendicular to `forward`.
		let axis = z.cross(forward);
		let length2 = axis.length2();
		// Choose a limit to make the cone within which cameras behave strangely to
		// ~0.1 degrees (I think?)
		// Hopefully this is a good compromise between not being visible to a user
		// and loss of precision artefacts at tiny angles.
		let limit2 = (0.002 as f32).powi(2);
		let mut out = if length2 < limit2 {
			// `forward` is very close to z or -z, do nothing or rotate 180 degrees
			// around an arbitrary axis.
			if z.dot(forward) < 0.0 {
				Self::new(up, std::f32::consts::PI)
			} else {
				Self::identity()
			}
		} else {
			let axis = axis/length2.sqrt();
			let angle = z.dot(forward).acos();
			Self::new(axis, angle)
		};
		// Rotate around `forward` to make local y match `up`.
		// Convert `up` to local coordinates, and project onto z plane to calculate
		// rotation.
		let local_up = out.inverse().apply(up);
		let length2 = local_up.x*local_up.x+local_up.y*local_up.y;
		if length2 > limit2 {
			// Not looking along `up`.
			let mut angle = (local_up.y/length2.sqrt()).acos();
			if local_up.x > 0.0 { angle = -angle; }
			out = out.then(Self::new(forward, angle));
		}
		out.unit().unwrap()
	}
	/// Rotate a vector around the origin.
	/// Note that if doing two or more of the same rotation it is more efficient
	/// to convert to a 3x3 matrix and perform the transformation using that.
	pub fn apply(&self, vector: F3) -> F3 {
		// See: https://blog.molecular-matters.com/2013/05/24/a-faster-quaternion-vector-multiplication/
		// See: https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation#Used_methods
		debug_assert!(self.is_unit(), "unit rotation needed for correct output");
		let imaginary = f3(self.b, self.c, self.d);
		vector+(imaginary*2.0).cross(imaginary.cross(vector)+vector*self.a)
	}
	/// Concatenate two rotations, the result will perform `self` then `next` when applied to a vector.
	/// Remember to normalise occasionally to avoid loss of precision (`unit()`).
	pub fn then(self, next: Self) -> Self {
		// AKA Hamilton product next*self
		Self {
			a: next.a*self.a - next.b*self.b - next.c*self.c - next.d*self.d,
			b: next.a*self.b + next.b*self.a + next.c*self.d - next.d*self.c,
			c: next.a*self.c - next.b*self.d + next.c*self.a + next.d*self.b,
			d: next.a*self.d + next.b*self.c - next.c*self.b + next.d*self.a,
		}
	}
	/// Return the opposite (i.e. negative) of a rotation.
	pub fn inverse(self) -> Self {
		// In other words, calculate the reciprocal of the quaternion.
		// This is generally q*/length(q)^2, but for a unit quaternion length(q) == 1.
		debug_assert!(self.is_unit(), "unit rotation needed for correct output");
		Self{a: self.a, b: -self.b, c: -self.c, d: -self.d}
	}
	/// Return true if this rotation appears not to have lost precision and is ok
	/// for use.
	pub fn is_unit(self) -> bool {
		let length2 = self.a*self.a+self.b*self.b+self.c*self.c+self.d*self.d;
		(length2-1.0).abs() < 0.0001
	}
	/// Normalise rotation to avoid loss of precision.
	pub fn unit(self) -> Option<Self> {
		let length2 = self.a*self.a+self.b*self.b+self.c*self.c+self.d*self.d;
		if length2 == 0.0 {
			None
		} else {
			let scale = length2.sqrt().recip();
			Some(Self{a: self.a*scale, b: self.b*scale, c: self.c*scale, d: self.d*scale})
		}
	}
	pub fn unit_identity(self) -> Self {
		self.unit().unwrap_or(Self::identity())
	}
	/// Linearly interpolate from `self` to `other` by the amount `factor` (0..1)
	/// using a crude but efficient algorithm (nlerp).
	/// Unlike spherical linear interpolate (slerp) this lerp does *not* maintain
	/// a constant velocity.
	pub fn fast_lerp(&self, mut other: Self, factor: f32) -> Self {
		// Negated quaternions represent the same rotation as their positive
		// counterpart, but interpolating between the two produces a long path.
		// If we're interpolating between quaternions that are too different, invert
		// one to avoid taking a ridiculously long path between the two.
		let dot = self.a*other.a + self.b*other.b + self.c*other.c + self.d*other.d;
		if dot < 0.0 {
			other = Self{a: -other.a, b: -other.b, c: -other.c, d: -other.d};
		}
		// Interpolate and normalise.
		Self {
			a: self.a+(other.a-self.a)*factor,
			b: self.b+(other.b-self.b)*factor,
			c: self.c+(other.c-self.c)*factor,
			d: self.d+(other.d-self.d)*factor,
		}.unit_identity()
	}
	/// Scale the amount of a rotation by `value`.
	/// For example, scaling a 90 degree rotation by 0.5 will result in a 45
	/// degree rotation around the same axis.
	/// Remember to normalise occasionally to avoid loss of precision (`unit()`).
	///
	/// This seems to behave badly with small scale values, using it for damping
	/// angular velocity and similar is not advised.
	pub fn scale(self, value: f32) -> Self {
		// While it is apprently possible to implement this as a power operation
		// `exp(log(quaternion)*scale)`
		// the implementation of this seems to end up being complicated and error
		// prone.
		// Doing it this way is short, simpler, easier to understand, and looks
		// like it costs a similar amount of compute.
		let axis = self.to_axis();
		let angle = self.to_angle();
		Self::new(axis, angle*value)
		/*
		// This is implemented as a quaterion to the power of `value`.
		// The power operation is in turn decomposed into `exp(ln(quaterion)*power)`.
		// We take advantage of `self` being a unit quaterion to skip some
		// operations.
		// See: https://users.aalto.fi/~ssarkka/pub/quat.pdf
		debug_assert!(self.is_unit(), "unit rotation needed for correct output");
		if value.abs() < 0.0001 {
			// Special case because `scale` (below) would become all zero.
			return Self::identity();
		}
		println!("self={},{},{},{}", self.a, self.b, self.c, self.d);
		let acos_a = self.a.acos(); // `length(self) == 1.0`
		let length_bcd = (self.b*self.b+self.c*self.c+self.d*self.d).sqrt();
		println!("length_bcd={}", length_bcd);
		if length_bcd < 0.0001 {
			// Special case because all of `bcd` would become infinity.
			return Self::identity();
		}
		let length_bcd_recip = length_bcd.recip();
		let logarithm = Self {
			a: 0.0, // `length(self) == 1.0` so `ln(length(self)) == 0.0`
			b: self.b*length_bcd_recip*acos_a,
			c: self.c*length_bcd_recip*acos_a,
			d: self.d*length_bcd_recip*acos_a,
		};
		println!("logarithm={},{},{},{}", logarithm.a, logarithm.b, logarithm.c, logarithm.d);
		// Note that this is likely to *not* be a unit quaternion.
		let scale = Self {
			a: logarithm.a*value,
			b: logarithm.b*value,
			c: logarithm.c*value,
			d: logarithm.d*value,
		};
		println!("scale={},{},{},{}", scale.a, scale.b, scale.c, scale.d);
		let exp_a = scale.a.exp();
		let length_bcd = (scale.b*scale.b+scale.c*scale.c+scale.d*scale.d).sqrt();
		println!("length_bcd={}", length_bcd);
		if length_bcd < 0.0001 {
			// Special case because otherwise we divide by zero.
			return Self{a: exp_a, b: 0.0, c: 0.0, d: 0.0};
		}
		let multiplier = length_bcd.recip()*length_bcd.sin();
		let exponential = Self {
			a: exp_a*length_bcd.cos(),
			b: exp_a*scale.b*multiplier,
			c: exp_a*scale.c*multiplier,
			d: exp_a*scale.d*multiplier,
		};
		debug_assert!(exponential.a.is_finite());
		debug_assert!(exponential.b.is_finite());
		debug_assert!(exponential.c.is_finite());
		debug_assert!(exponential.d.is_finite());
		exponential
		*/
	}
	/// Return axis rotation is around as a unit vector.
	pub fn to_axis(self) -> F3 {
		debug_assert!(self.is_unit(), "unit rotation needed for correct output");
		if false {
			let length2 = self.b*self.b+self.c*self.c+self.d*self.d;
			if length2.sqrt() < 0.0001 { return f3(0.0, 0.0, 1.0); }
			return f3(self.b, self.c, self.d)/length2.sqrt();
		}
		// Handle identities (imaginary part becomes zeroes).
		if self.a.abs() > 1.0-0.0001 { return f3(0.0, 0.0, 1.0); }
		// Normalise imaginary part to get axis.
		// Because we use unit quaternions, we can calculate the length of the
		// imaginary part more efficiently:
		// length = sqrt(x^2+y^2+z^2) = sqrt(1-real^2)
		// (because by definition `a^2+b^2+c^2+d^2 == 1.0`)
		// See: https://github.com/godotengine/godot/blob/master/core/math/quaternion.cpp
		let length = (1.0-self.a*self.a).sqrt();
		f3(self.b, self.c, self.d)/length
	}
	/// Return angle of rotation in radians.
	pub fn to_angle(self) -> f32 {
		// See: https://github.com/godotengine/godot/blob/master/core/math/quaternion.cpp
		// See: https://stackoverflow.com/questions/3825571/how-to-convert-quaternion-to-angle
		debug_assert!(self.is_unit(), "unit rotation needed for correct output");
		self.a.acos()*2.0
	}
	/// Return a matrix that performs the same rotation.
	/// Building a matrix and applying it instead is more efficient when rotating
	/// more than a couple of vectors.
	pub fn to_matrix3(self) -> Matrix3 {
		// See: https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation
		debug_assert!(self.is_unit(), "unit rotation needed for correct output");
		Matrix3{matrix: [
			// Left column
			1.0-2.0*(self.c*self.c+self.d*self.d),
			2.0*(self.b*self.c+self.d*self.a),
			2.0*(self.b*self.d-self.c*self.a),
			// Centre column
			2.0*(self.b*self.c-self.d*self.a),
			1.0-2.0*(self.b*self.b+self.d*self.d),
			2.0*(self.c*self.d+self.b*self.a),
			// Right column
			2.0*(self.b*self.d+self.c*self.a),
			2.0*(self.c*self.d-self.b*self.a),
			1.0-2.0*(self.b*self.b+self.c*self.c),
		]}
	}
	/// Equivalent to `Matrix3::scale(scale).then(self.to_matrix3())`, but skips
	/// multiplies where the scale matrix will always be zero.
	pub fn to_matrix3_scale(self, scale: F3) -> Matrix3 {
		let mut matrix = self.to_matrix3();
		// Left column.
		matrix.matrix[0] *= scale.x;
		matrix.matrix[1] *= scale.x;
		matrix.matrix[2] *= scale.x;
		// Middle column.
		matrix.matrix[3] *= scale.y;
		matrix.matrix[4] *= scale.y;
		matrix.matrix[5] *= scale.y;
		// Right column.
		matrix.matrix[6] *= scale.z;
		matrix.matrix[7] *= scale.z;
		matrix.matrix[8] *= scale.z;
		matrix
	}
}

#[cfg(test)]
mod tests {
	use super::*;
	use crate::matrix3::Matrix3;

	fn equal_ish(a: F3, b: F3) -> bool {
		let limit = 0.0001;
		(a.x-b.x).abs() < limit && (a.y-b.y).abs() < limit && (a.z-b.z).abs() < limit
	}
	fn equal_rotations(a: Rotation, b: Rotation) -> bool {
		println!("{},{},{},{} == {},{},{},{}", a.a, a.b, a.c, a.d, b.a, b.b, b.c, b.d);
		let limit = 0.01;
		(a.a-b.a).abs() < limit
			&& (a.b-b.b).abs() < limit
			&& (a.c-b.c).abs() < limit
			&& (a.d-b.d).abs() < limit
	}
	fn apply_hamilton(rotation: Rotation, vector: F3) -> F3 {
		// Crude and inefficient version of `apply()`.
		let inverse = rotation.inverse();
		let vector = Rotation{a: 0.0, b: vector.x, c: vector.y, d: vector.z};
		let result = inverse.then(vector.then(rotation));
		f3(result.b, result.c, result.d)
	}

	#[test]
	fn apply() {
		let zero = f3(0.0, 0.0, 0.0);
		let x = f3(1.0, 0.0, 0.0);
		let y = f3(0.0, 1.0, 0.0);
		let z = f3(0.0, 0.0, 1.0);

		let rotation = Rotation::identity();
		assert!(equal_ish(rotation.apply(zero), zero));
		assert!(equal_ish(rotation.apply(x), x));
		assert!(equal_ish(rotation.apply(y), y));
		assert!(equal_ish(rotation.apply(z), z));

		let rotation = Rotation::new(x, std::f32::consts::PI*0.5);
		assert!(equal_ish(rotation.apply(zero), zero));
		assert!(equal_ish(rotation.apply(x), x));
		assert!(equal_ish(rotation.apply(y), z));
		assert!(equal_ish(rotation.apply(z), -y));

		let rotation = Rotation::new(x, -std::f32::consts::PI*0.5);
		assert!(equal_ish(rotation.apply(zero), zero));
		assert!(equal_ish(rotation.apply(x), x));
		assert!(equal_ish(rotation.apply(y), -z));
		assert!(equal_ish(rotation.apply(z), y));

		let rotation = Rotation::new(y, std::f32::consts::PI*0.5);
		assert!(equal_ish(rotation.apply(zero), zero));
		assert!(equal_ish(rotation.apply(x), -z));
		assert!(equal_ish(rotation.apply(y), y));
		assert!(equal_ish(rotation.apply(z), x));

		let rotation = Rotation::new(y, -std::f32::consts::PI*0.5);
		assert!(equal_ish(rotation.apply(zero), zero));
		assert!(equal_ish(rotation.apply(x), z));
		assert!(equal_ish(rotation.apply(y), y));
		assert!(equal_ish(rotation.apply(z), -x));

		let rotation = Rotation::new(z, std::f32::consts::PI*0.5);
		assert!(equal_ish(rotation.apply(zero), zero));
		assert!(equal_ish(rotation.apply(x), y));
		assert!(equal_ish(rotation.apply(y), -x));
		assert!(equal_ish(rotation.apply(z), z));

		let rotation = Rotation::new(z, -std::f32::consts::PI*0.5);
		assert!(equal_ish(rotation.apply(zero), zero));
		assert!(equal_ish(rotation.apply(x), -y));
		assert!(equal_ish(rotation.apply(y), x));
		assert!(equal_ish(rotation.apply(z), z));

		for rotation in [
			Rotation::identity(),
			Rotation::new(x, std::f32::consts::PI*0.5),
			Rotation::new(x, -std::f32::consts::PI*0.5),
			Rotation::new(y, std::f32::consts::PI*0.5),
			Rotation::new(y, -std::f32::consts::PI*0.5),
			Rotation::new(z, std::f32::consts::PI*0.5),
			Rotation::new(z, -std::f32::consts::PI*0.5),
		].iter() {
			assert!(equal_ish(rotation.apply(zero), apply_hamilton(*rotation, zero)));
			assert!(equal_ish(rotation.apply(x), apply_hamilton(*rotation, x)));
			assert!(equal_ish(rotation.apply(y), apply_hamilton(*rotation, y)));
			assert!(equal_ish(rotation.apply(z), apply_hamilton(*rotation, z)));
		}
	}

	#[test]
	fn inverse() {
		let zero = f3(0.0, 0.0, 0.0);
		let x = f3(1.0, 0.0, 0.0);
		let y = f3(0.0, 1.0, 0.0);
		let z = f3(0.0, 0.0, 1.0);
		
		for rotation in [
			Rotation::identity(),
			Rotation::new(x, std::f32::consts::PI*0.5),
			Rotation::new(x, -std::f32::consts::PI*0.5),
			Rotation::new(y, std::f32::consts::PI*0.5),
			Rotation::new(y, -std::f32::consts::PI*0.5),
			Rotation::new(z, std::f32::consts::PI*0.5),
			Rotation::new(z, -std::f32::consts::PI*0.5),
		].iter() {
			let inverse = rotation.inverse();
			assert!(equal_ish(inverse.apply(rotation.apply(zero)), zero));
			assert!(equal_ish(inverse.apply(rotation.apply(x)), x));
			assert!(equal_ish(inverse.apply(rotation.apply(y)), y));
			assert!(equal_ish(inverse.apply(rotation.apply(z)), z));

			let identity = rotation.then(inverse);
			let limit = 0.0001;
			assert!((identity.a-1.0).abs() < limit);
			assert!((identity.b-0.0).abs() < limit);
			assert!((identity.c-0.0).abs() < limit);
			assert!((identity.d-0.0).abs() < limit);
		}
	}

	#[test]
	fn look() {
		let x = f3(1.0, 0.0, 0.0);
		let y = f3(0.0, 1.0, 0.0);
		let z = f3(0.0, 0.0, 1.0);

		let rotation = Rotation::look(x, y);
		assert!(equal_ish(rotation.apply(x), -z));
		assert!(equal_ish(rotation.apply(y), y));
		assert!(equal_ish(rotation.apply(z), x));

		let rotation = Rotation::look(-x, y);
		assert!(equal_ish(rotation.apply(x), z));
		assert!(equal_ish(rotation.apply(y), y));
		assert!(equal_ish(rotation.apply(z), -x));

		let rotation = Rotation::look(y, y);
		assert!(equal_ish(rotation.apply(z), y));

		let rotation = Rotation::look(-y, y);
		assert!(equal_ish(rotation.apply(z), -y));

		let rotation = Rotation::look(z, y);
		assert!(equal_ish(rotation.apply(x), x));
		assert!(equal_ish(rotation.apply(y), y));
		assert!(equal_ish(rotation.apply(z), z));

		let rotation = Rotation::look(-z, y);
		assert!(equal_ish(rotation.apply(x), -x));
		assert!(equal_ish(rotation.apply(y), y));
		assert!(equal_ish(rotation.apply(z), -z));

		let cant = Rotation::new(x, 0.2);
		let cant_y = cant.apply(y);
		let cant_z = cant.apply(z);

		let rotation = Rotation::look(cant_z, y);
		assert!(equal_ish(rotation.apply(x), x));
		assert!(equal_ish(rotation.apply(y), cant_y));
		assert!(equal_ish(rotation.apply(z), cant_z));

		let rotation = Rotation::look(-cant_z, y);
		assert!(equal_ish(rotation.apply(x), -x));
		assert!(equal_ish(rotation.apply(y), cant_y));
		assert!(equal_ish(rotation.apply(z), -cant_z));

		let rotation = Rotation::look(cant_y, y);
		assert!(equal_ish(rotation.apply(x), x));
		assert!(equal_ish(rotation.apply(y), -cant_z));
		assert!(equal_ish(rotation.apply(z), cant_y));

		let rotation = Rotation::look(-cant_y, y);
		assert!(equal_ish(rotation.apply(x), -x));
		assert!(equal_ish(rotation.apply(y), -cant_z));
		assert!(equal_ish(rotation.apply(z), -cant_y));

		let cant = Rotation::new(z, 0.2);
		let cant_x = cant.apply(x);
		let cant_y = cant.apply(y);

		let rotation = Rotation::look(cant_x, y);
		assert!(equal_ish(rotation.apply(x), -z));
		assert!(equal_ish(rotation.apply(y), cant_y));
		assert!(equal_ish(rotation.apply(z), cant_x));

		let rotation = Rotation::look(-cant_x, y);
		assert!(equal_ish(rotation.apply(x), z));
		assert!(equal_ish(rotation.apply(y), cant_y));
		assert!(equal_ish(rotation.apply(z), -cant_x));

		let rotation = Rotation::look(cant_y, y);
		println!("{:?} {:?}", rotation.apply(x), z);
		assert!(equal_ish(rotation.apply(x), z));
		assert!(equal_ish(rotation.apply(y), cant_x));
		assert!(equal_ish(rotation.apply(z), cant_y));

		let rotation = Rotation::look(-cant_y, y);
		assert!(equal_ish(rotation.apply(x), -z));
		assert!(equal_ish(rotation.apply(y), cant_x));
		assert!(equal_ish(rotation.apply(z), -cant_y));
	}

	#[test]
	fn scale() {
		for (axis, angle) in [
			(f3(1.0, 0.0, 0.0), std::f32::consts::PI*0.5),
			(f3(1.0, 0.0, 0.0), std::f32::consts::PI*0.3),
			(f3(0.0, 1.0, 0.0), std::f32::consts::PI*0.5),
			(f3(0.0, 1.0, 0.0), std::f32::consts::PI*0.3),
			(f3(0.0, 0.0, 1.0), std::f32::consts::PI*0.5),
			(f3(0.0, 0.0, 1.0), std::f32::consts::PI*0.3),
			(f3(1.0, 1.0, 0.0).unit().unwrap(), std::f32::consts::PI*0.5),
			(f3(1.0, 1.0, 0.0).unit().unwrap(), std::f32::consts::PI*0.3),
			(f3(0.0, 1.0, 1.0).unit().unwrap(), std::f32::consts::PI*0.5),
			(f3(0.0, 1.0, 1.0).unit().unwrap(), std::f32::consts::PI*0.3),
			(f3(1.0, 0.0, 1.0).unit().unwrap(), std::f32::consts::PI*0.5),
			(f3(1.0, 0.0, 1.0).unit().unwrap(), std::f32::consts::PI*0.3),
		].iter().copied() {
			let original = Rotation::new(axis, angle);
			for scale in [2.0, 1.0, 0.3, -0.3, -1.0, -2.0].iter().copied() {
				let value = original.scale(scale);
				let mut check_axis = value.to_axis();
				let mut check_angle = value.to_angle();
				if check_axis.dot(axis) < 0.0 { // allow inverted axis
					println!("(inverted)");
					check_axis = -check_axis;
					check_angle = -check_angle;
				}
				if check_angle <= -std::f32::consts::PI && angle*scale > 0.0 { // alow -pi == pi
					println!("(wrapped)");
					check_angle += std::f32::consts::PI*2.0;
				}
				println!("{} {} == {} {}", check_axis, check_angle, axis, angle*scale);
				assert!(equal_ish(check_axis, axis));
				assert!((check_angle-angle*scale).abs() < 0.01);
			}

			assert!(equal_rotations(original.scale(1.0), original));
			assert!(equal_rotations(original.scale(0.0), Rotation::identity()));
			assert!(equal_rotations(original.scale(-1.0), original.inverse()));
			assert!(equal_rotations(original.scale(2.0), original.then(original)));
		}

		for scale in [2.0, 1.0, 0.3, 0.0, -0.3, -1.0, 2.0].iter().copied() {
			let rotation = Rotation::identity().scale(scale);
			println!("{}", scale);
			assert!(equal_rotations(rotation, Rotation::identity()));
		}
	}

	#[test]
	fn to_axis() {
		let x = f3(1.0, 0.0, 0.0);
		let y = f3(0.0, 1.0, 0.0);
		let z = f3(0.0, 0.0, 1.0);

		for (axis, angle) in [
			(z, 0.0),
			(x, std::f32::consts::PI*0.5),
			(-x, std::f32::consts::PI*0.5),
			(y, std::f32::consts::PI*0.5),
			(-y, std::f32::consts::PI*0.5),
			(z, std::f32::consts::PI*0.5),
			(-z, std::f32::consts::PI*0.5),
		].iter().copied() {
			let rotation = Rotation::new(axis, angle);
			assert!(equal_ish(rotation.to_axis(), axis));
			assert!((rotation.to_angle()-angle).abs() < 0.0001);
		}
	}

	#[test]
	fn to_matrix3() {
		let zero = f3(0.0, 0.0, 0.0);
		let x = f3(1.0, 0.0, 0.0);
		let y = f3(0.0, 1.0, 0.0);
		let z = f3(0.0, 0.0, 1.0);

		assert!(Rotation::identity().to_matrix3() == Matrix3::identity());

		for rotation in [
			Rotation::identity(),
			Rotation::new(x, std::f32::consts::PI*0.5),
			Rotation::new(x, -std::f32::consts::PI*0.5),
			Rotation::new(y, std::f32::consts::PI*0.5),
			Rotation::new(y, -std::f32::consts::PI*0.5),
			Rotation::new(z, std::f32::consts::PI*0.5),
			Rotation::new(z, -std::f32::consts::PI*0.5),
		].iter() {
			assert!(equal_ish(rotation.apply(zero), rotation.to_matrix3().apply(zero)));
			assert!(equal_ish(rotation.apply(x), rotation.to_matrix3().apply(x)));
			assert!(equal_ish(rotation.apply(y), rotation.to_matrix3().apply(y)));
			assert!(equal_ish(rotation.apply(z), rotation.to_matrix3().apply(z)));

			let scale = f3(1.0, 2.0, -2.5);
			assert!(equal_ish(rotation.apply(zero*scale), rotation.to_matrix3_scale(scale).apply(zero)));
			assert!(equal_ish(rotation.apply(x*scale), rotation.to_matrix3_scale(scale).apply(x)));
			assert!(equal_ish(rotation.apply(y*scale), rotation.to_matrix3_scale(scale).apply(y)));
			assert!(equal_ish(rotation.apply(z*scale), rotation.to_matrix3_scale(scale).apply(z)));
		}
	}
}
