//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use std::convert::TryFrom;
use std::marker::PhantomData;
use std::rc::Weak;
use super::blend::Blend;
use super::ffi_gles::*;
use super::matrix3::Matrix3;
use super::mesh::StaticMesh;
use super::metrics::MetricUse;
use super::rotation::Rotation;
use super::shader::{ActiveShader, Uniform};
use super::vector3::{F3, f3};
use super::vertex::{INDEX_TYPE, Index, Primitive, VertexBinding};
use super::window::Window;

/// Trait to provide extra methods for a normal vertex struct (containing
/// attribute data) allowing operations needed for building scenery meshes.
pub trait SceneryVertex: VertexBinding {
	fn position(&self) -> F3;
	fn transform(&self, matrix: &Matrix3, translate: F3) -> Self;
}
impl SceneryVertex for super::vertex::Vertex {
	fn position(&self) -> F3 {
		F3::from_array(self.position)
	}
	fn transform(&self, matrix: &Matrix3, translate: F3) -> Self {
		Self {
			position: (matrix.apply(F3::from_array(self.position))+translate).to_array(),
			normal: matrix.apply(F3::from_array(self.normal)).unit_zero().to_array(),
			texcoord: self.texcoord,
		}
	}
}

/// Utility to concatenate mesh data together to build a scenery mesh.
pub struct SceneryBuilder<Vertex: SceneryVertex> {
	pub indices: Vec<Index>,
	pub vertices: Vec<Vertex>,
}
impl<Vertex: SceneryVertex> SceneryBuilder<Vertex> {
	pub fn new() -> Self {
		Self {
			indices: Vec::new(),
			vertices: Vec::new(),
		}
	}
	/// Append a mesh to the scenery with the given transformation pre-applied.
	pub fn push(&mut self, indices: &[Index], vertices: &[Vertex], scale: F3, rotate: Rotation, translate: F3) {
		let offset = self.vertices.len();
		self.vertices.reserve(vertices.len());
		let matrix = rotate.to_matrix3_scale(scale);
		for vertex in vertices.iter() {
			self.vertices.push(vertex.transform(&matrix, translate));
		}
		debug_assert!(self.vertices.len() <= Index::MAX as usize, "ran out of vertex indices");

		debug_assert!(indices.len()%3 == 0);
		self.indices.reserve(indices.len());
		for index in indices.iter().copied() {
			self.indices.push(Index::try_from(index as usize+offset).unwrap());
		}
	}
	/// Return the finished scenery mesh.
	/// `grid_size` chooses the size of the grid the mesh is divided into.
	/// `visible` is the distance at which chunks of mesh are fully opaque.
	/// `fade` is the distance beyond `visible` over which chunks fade in.
	pub fn build(self, window: &mut Window, grid_size: [usize; 3], visible: f32, fade: f32) -> SceneryMesh<Vertex> {
		let geometry_bytes = window.metrics().geometry_bytes.take(
			std::mem::size_of::<Index>()*self.indices.len()+
			std::mem::size_of::<Vertex>()*self.vertices.len());
		// Bounding box.
		let mut grid_min = self.vertices.first()
			.map(|vertex| vertex.position())
			.unwrap_or(f3(0.0, 0.0, 0.0));
		let mut grid_max = grid_min;
		for vertex in self.vertices.iter() {
			let position = vertex.position();
			grid_min = position.min(grid_min);
			grid_max = position.max(grid_max);
		}
		// Divide triangle indices by cell.
		let length = grid_size[0]*grid_size[1]*grid_size[2];
		let mut cell_indices = vec![Vec::new(); length].into_boxed_slice();
		let cell_size = f3(
			(grid_max.x-grid_min.x)/grid_size[0] as f32,
			(grid_max.y-grid_min.y)/grid_size[1] as f32,
			(grid_max.z-grid_min.z)/grid_size[2] as f32,
		);
		for indices in self.indices.chunks_exact(3) {
			let mut min = self.vertices[indices[0] as usize].position();
			let mut max = self.vertices[indices[0] as usize].position();
			for index in indices[1..].iter().copied() {
				let position = self.vertices[index as usize].position();
				min = position.min(min);
				max = position.max(max);
			}
			debug_assert!(max.x-min.x <= cell_size.x,
				"triangle larger than grid cell (x, {} vs {})", max.x-min.x, cell_size.x);
			debug_assert!(max.y-min.y <= cell_size.y,
				"triangle larger than grid cell (y, {} vs {})", max.y-min.y, cell_size.y);
			debug_assert!(max.z-min.z <= cell_size.z,
				"triangle larger than grid cell (z, {} vs {})", max.z-min.z, cell_size.z);
			let centre = (min+max)*0.5;
			let cell = [
				((centre.x-grid_min.x)/cell_size.x).floor() as usize,
				((centre.y-grid_min.y)/cell_size.y).floor() as usize,
				((centre.z-grid_min.z)/cell_size.z).floor() as usize,
			];
			let index = cell[0]+cell[1]*grid_size[0]+cell[2]*grid_size[0]*grid_size[1];
			cell_indices[index].extend_from_slice(indices);
		}
		// Build new buffers where indices for triangles in each cell are contiguous.
		let mut cells = Vec::with_capacity(cell_indices.len());
		let mut indices = Vec::with_capacity(self.indices.len());
		for cell in cell_indices.iter() {
			debug_assert!(cell.len()%3 == 0);
			cells.push(Cell {
				start: indices.len(),
				count: Index::try_from(cell.len()).unwrap(),
			});
			indices.extend_from_slice(cell);
		}
		// Upload data onto GPU.
		unsafe {
			let gl = window.gl();
			let vertices = StaticMesh::<Vertex>::upload_buffer(gl, GL_ARRAY_BUFFER, &self.vertices);
			let indices = StaticMesh::<Vertex>::upload_buffer(gl, GL_ELEMENT_ARRAY_BUFFER, &indices);
			SceneryMesh {
				cell_size,
				cells: cells.into_boxed_slice(),
				fade,
				fade_queue: Vec::new(),
				geometry_bytes,
				gl: window.weak_gl(),
				grid_origin: grid_min,
				grid_size,
				indices,
				phantom: PhantomData,
				vertices,
				visible,
			}
		}
	}
}

/// A specialised static mesh intended for drawing large unchanging, unmoving
/// objects like landscapes, building, and immovable props.
/// This is done by combining all of the elements of the mesh together,
/// pre-transforming the vertices, and dividing it into a grid.
/// Each cell of the grid is automatically turned on and off with a fade
/// in/out animation when it gets to given distance from the camera (although
/// the visible area is a cube instead of a sphere to save compute).
/// With appropriately chosen grid cell sizes and draw distance this should
/// reduce GPU load for large maps.
/// Pre-transforming vertices means that only a camera transformation matrix
/// is needed - which can be shared with the rest of the scene - instead of
/// concatenating it with a model transform (or many depending on the scene)
/// every frame.
pub struct SceneryMesh<Vertex: SceneryVertex> {
	/// Size of a grid cell in world coordinates.
	cell_size: F3,
	/// Contents of each grid cell.
	cells: Box<[Cell]>,
	/// Distance over which to fade cells in/out.
	/// Note that this describes the radius of a cube, not a sphere.
	fade: f32,
	/// Data about fading cells, recycled to save allocations.
	/// See `draw()`.
	fade_queue: Vec<(usize, f32)>,
	/// Bytes used on GPU.
	#[allow(dead_code)]
	geometry_bytes: MetricUse,
	/// Reference to the OpenGL context.
	gl: Weak<GL>,
	/// Position of corner of first cell in world coordinates.
	grid_origin: F3,
	/// Dimensions of grid.
	grid_size: [usize; 3],
	/// Handle of index buffer.
	indices: GLuint,
	phantom: PhantomData<Vertex>,
	/// Handle of vertex buffer.
	vertices: GLuint,
	/// Distance at which cells are visible and completely opaque.
	/// Note that this describes the radius of a cube, not a sphere.
	visible: f32,
}
impl<Vertex: SceneryVertex> SceneryMesh<Vertex> {
	/// Draw mesh using currently active shader.
	/// The shader's view matrix must be set to the camera's projection matrix
	/// before calling this function!
	/// This function overrides the active blend mode (see `Blend`).
	///
	/// `camera_centre` is the position of the camera in world coordinates, for
	/// calculation of cell visibility.
	/// `alpha_uniform` is a float shader uniform which will be set to the opacity
	/// of each cell during rendering.
	pub fn draw(&mut self, window: &mut Window, camera_centre: F3, shader: &ActiveShader<Vertex>, alpha_uniform: &Option<Uniform>) {
		Blend::Replace.draw(window);
		shader.set(window, alpha_uniform, 1.0);

		{
			let gl = window.get_gl(&self.gl);

			// Activate buffers.
			unsafe {
				(gl.BindBuffer)(GL_ARRAY_BUFFER, self.vertices);
				(gl.BindBuffer)(GL_ELEMENT_ARRAY_BUFFER, self.indices);
				Vertex::load(gl);
			}

			// Totally visible cells.
			let max_cell_size = self.cell_size.x.max(self.cell_size.y).max(self.cell_size.z);
			let radius = self.visible+self.fade;
			let radius = f3(radius, radius, radius);
			let min = self.to_cell(camera_centre-radius);
			let max = self.to_cell(camera_centre+radius+self.cell_size); // add one, iteration exludes max
			let fade_recip = self.fade.recip();
			let mut offset_z = self.grid_origin.z+
				min[2] as f32*self.cell_size.z+
				self.cell_size.z*0.5-camera_centre.z;
			debug_assert!(self.fade_queue.len() == 0);
			for z in min[2]..max[2] {
				let mut offset_y = self.grid_origin.y+
					min[1] as f32*self.cell_size.y+
					self.cell_size.y*0.5-camera_centre.y;
				for y in min[1]..max[1] {
					let mut offset_x = self.grid_origin.x+
						min[0] as f32*self.cell_size.x+
						self.cell_size.x*0.5-camera_centre.x;
					for x in min[0]..max[0] {
						let distance = offset_x.abs().max(offset_y.abs()).max(offset_z.abs());
						let index = x+y*self.grid_size[0]+z*self.grid_size[0]*self.grid_size[1];
						if distance <= self.visible {
							let cell = &self.cells[index];
							let start = (cell.start*std::mem::size_of::<Index>()) as *const GLvoid;
							unsafe {
								(gl.DrawElements)(
									Primitive::Triangles.raw(), cell.count as GLint, INDEX_TYPE, start);
							}
						} else if distance < self.visible+self.fade {
							let alpha = 1.0-(distance-self.visible)*fade_recip;
							self.fade_queue.push((index, alpha));
						} else {
							debug_assert!(distance <= self.visible+self.fade+max_cell_size);
						}
						offset_x += self.cell_size.x;
					}
					offset_y += self.cell_size.y;
				}
				offset_z += self.cell_size.z;
			}
		}

		// Partially visible cells.
		Blend::Alpha.draw(window);
		let mut last_alpha = 0.0; // force change of uniform
		for (index, alpha) in self.fade_queue.drain(..) {
			debug_assert!(alpha > 0.0);
			if alpha != last_alpha {
				shader.set(window, alpha_uniform, alpha);
				last_alpha = alpha;
			}
			let cell = &self.cells[index];
			let start = (cell.start*std::mem::size_of::<Index>()) as *const GLvoid;
			unsafe {
				let gl = window.gl();
				(gl.DrawElements)(Primitive::Triangles.raw(), cell.count as GLint, INDEX_TYPE, start);
			}
		}

		// Tidy up.
		if cfg!(debug_assertions) { unsafe {
			let gl = window.gl();
			(gl.BindBuffer)(GL_ARRAY_BUFFER, 0);
			(gl.BindBuffer)(GL_ELEMENT_ARRAY_BUFFER, 0);
			gl.check();
		} }
	}

	fn to_cell(&self, point: F3) -> [usize; 3] {
		[
			(((point.x-self.grid_origin.x)/self.cell_size.x).floor()
				.max(0.0) as usize).min(self.grid_size[0]),
			(((point.y-self.grid_origin.y)/self.cell_size.y).floor()
				.max(0.0) as usize).min(self.grid_size[1]),
			(((point.z-self.grid_origin.z)/self.cell_size.z).floor()
				.max(0.0) as usize).min(self.grid_size[2]),
		]
	}
}
impl<Vertex: SceneryVertex> Drop for SceneryMesh<Vertex> {
	fn drop(&mut self) { unsafe {
		if let Some(gl) = self.gl.upgrade() {
			let buffers = [self.indices, self.vertices];
			(gl.DeleteBuffers)(buffers.len() as GLsizei, buffers.as_ptr());
		}
	} }
}

struct Cell {
	/// Start of slice of index buffer to draw (triangles within this cell).
	pub start: usize,
	/// Length of slice of index buffer to draw (triangles within this cell).
	pub count: Index,
}
