//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use std::ffi::CString;
use std::marker::PhantomData;
use std::rc::Weak;
use super::ffi_gles::*;
use super::matrix3::Matrix3;
use super::texture::Texture;
use super::vector2::F2;
use super::vector3::F3;
use super::vertex::VertexBinding;
use super::window::Window;

	/// Return a matrix that transforms from model vectors to world vectors
	/// without including translation.
	/// This is useful for normals and lighting information (thus light matrix).
	// fn light_matrix(scale: F3, rotate: Rotation) -> [f32; 9] {
	// 	// Special casing this is a lot more efficient than multiplying two matrices.
	// 	let mut light_matrix = rotate.to_transform();
	// 	// Left column.
	// 	light_matrix.matrix[0] *= scale.x;
	// 	light_matrix.matrix[1] *= scale.x;
	// 	light_matrix.matrix[2] *= scale.x;
	// 	// Middle column.
	// 	light_matrix.matrix[3] *= scale.y;
	// 	light_matrix.matrix[4] *= scale.y;
	// 	light_matrix.matrix[5] *= scale.y;
	// 	// Right column.
	// 	light_matrix.matrix[6] *= scale.z;
	// 	light_matrix.matrix[7] *= scale.z;
	// 	light_matrix.matrix[8] *= scale.z;
	// 	light_matrix.matrix
	// }

/// Implemented for any type that can be passed to a shader as a uniform.
pub trait UniformValue {
	/// Set the uniform at `index` to this value.
	/// This requires interacting with OpenGL directly, so is `unsafe`.
	unsafe fn set(&self, gl: &GL, index: GLint);
}
impl UniformValue for f32 {
	unsafe fn set(&self, gl: &GL, index: GLint) { (gl.Uniform1f)(index, *self); }
}
impl UniformValue for [f32; 2] {
	unsafe fn set(&self, gl: &GL, index: GLint) { (gl.Uniform2f)(index, self[0], self[1]); }
}
impl UniformValue for [f32; 3] {
	unsafe fn set(&self, gl: &GL, index: GLint) { (gl.Uniform3f)(index, self[0], self[1], self[2]); }
}
impl UniformValue for [f32; 4] {
	unsafe fn set(&self, gl: &GL, index: GLint) { (gl.Uniform4f)(index, self[0], self[1], self[2], self[3]); }
}
impl UniformValue for F2 {
	unsafe fn set(&self, gl: &GL, index: GLint) { (gl.Uniform2f)(index, self.x, self.y); }
}
impl UniformValue for F3 {
	unsafe fn set(&self, gl: &GL, index: GLint) { (gl.Uniform3f)(index, self.x, self.y, self.z); }
}
impl UniformValue for &Matrix3 {
	unsafe fn set(&self, gl: &GL, index: GLint) { (gl.UniformMatrix3fv)(index, 1, GL_FALSE, self.matrix.as_ptr()); }
}
impl UniformValue for (u32, &Texture) { // texture slot index and texture to bind
	unsafe fn set(&self, gl: &GL, index: GLint) {
		debug_assert!(self.0 < 8, "only eight texture units are guaranteed to be available");
		(gl.ActiveTexture)(GL_TEXTURE0+self.0 as GLenum);
		(gl.BindTexture)(GL_TEXTURE_2D, self.1.raw());
		(gl.Uniform1i)(index, self.0 as GLint);
	}
}

/// Reference to a uniform in a loaded shader.
#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub struct Uniform {
	index: GLint,
	program: GLuint,
}

/// Handle to a shader that is currently active.
/// All this really does is make it clear that drawing operation require a
/// shader to be setup first.
pub struct ActiveShader<'a, Vertex: VertexBinding> {
	shader: &'a Shader<Vertex>,
}
impl<'a, Vertex: VertexBinding> ActiveShader<'a, Vertex> {
	/// Set a uniform attached to this shader for drawing operations until this
	/// function is called again.
	/// See `Shader::find()` to get a reference to a uniform to set.
	pub fn set<U: UniformValue>(&self, window: &mut Window, uniform: &Option<Uniform>, value: U) {
		if let Some(uniform) = uniform {
			debug_assert!(uniform.program == self.shader.program, "uniform must be from this shader");
			let gl = window.get_gl(&self.shader.gl);
			if cfg!(debug_assertions) {
				let mut active = 0;
				unsafe{(gl.GetIntegerv)(GL_CURRENT_PROGRAM, &mut active);}
				debug_assert!(active as GLuint == self.shader.program, "shader must be active");
			}
			unsafe{value.set(gl, uniform.index)}
		}
	}
}

/// A combined vertex and fragment shader pair.
/// Shaders are programs which run on the GPU and are used to draw every kind
/// of geometry.
/// The vertex part of the shader performs transformations to a mesh's
/// vertices that need to happen every frame (view transforms and similar).
/// The fragment part of the shader is run for every pixel on screen in which
/// the mesh is visible, and calculates the colour of the pixel (shading,
/// texture lookups, fancy effects passes).
pub struct Shader<Vertex: VertexBinding> {
	/// Used to drop the shader.
	gl: Weak<GL>,
	/// Name of this shader for debugging output.
	name: &'static str,
	phantom: PhantomData<Vertex>,
	/// Handle of the compiled shader program.
	program: GLuint,
}
impl<Vertex: VertexBinding> Shader<Vertex> {
	/// Compile a shader from the sources `vertex` and `fragment`.
	/// Any errors during compilation are returned.
	pub fn new(window: &mut Window, name: &'static str, vertex: &str, fragment: &str) -> Result<Self, String> {
		let gl = unsafe{window.gl()};
		// Compile and link shader.
		let vertex = ShaderObject::new(gl, name, vertex, GL_VERTEX_SHADER)?;
		let fragment = ShaderObject::new(gl, name, fragment, GL_FRAGMENT_SHADER)?;
		let program = ProgramObject::new::<Vertex>(gl, name, vertex, fragment)?;
		// let uniforms = Uniforms {
		// 	preclip_matrix: program.get_uniform(gl, name, "preclip_matrix", &allow),
		// 	translation: program.get_uniform(gl, name, "translation", &allow),
		// 	z_offset: program.get_uniform(gl, name, "z_offset", &allow),
		// 	z_scale: program.get_uniform(gl, name, "z_scale", &allow),
		// 	light_matrix: program.get_uniform(gl, name, "light_matrix", &allow),
		// 	rgba: program.get_uniform(gl, name, "rgba", &allow),
		// 	texture0: program.get_uniform(gl, name, "texture0", &allow),
		// 	texture1: program.get_uniform(gl, name, "texture1", &allow),
		// };
		// Tidy up.
		gl.check();
		Ok(Self {
			gl: window.weak_gl(),
			name,
			phantom: PhantomData,
			program: unsafe{program.to_raw()},
		})
	}
	/// Load a shader from a pair of files, `path+".vert"` and `path+".frag".
	/// Return any filesystem and compilation errors.
	pub fn load(window: &mut Window, name: &'static str, path: &str) -> Result<Self, String> {
		let load = |extension| {
			let path = format!("{}{}", path, extension);
			let data = std::fs::read(&path)
				.map_err(|error| format!("failed to load {}, {}", &path, error))?;
			String::from_utf8(data)
				.map_err(|error| format!("failed to load {}, {}", &path, error))
		};
		let vertex = load(".vert")?;
		let fragment = load(".frag")?;
		Self::new(window, name, &vertex, &fragment)
	}
	/// Look up a uniform by name for future use.
	/// For efficiency, I'd recommend  looking up all uniforms when a shader is
	/// loaded to avoid needing to look for them every frame.
	pub fn uniform(&self, window: &mut Window, uniform: &str) -> Option<Uniform> { unsafe {
		let gl = window.get_gl(&self.gl);
		let c_uniform = CString::new(uniform).unwrap();
		let index = (gl.GetUniformLocation)(self.program, c_uniform.as_ptr());
		if index < 0 {
			eprintln!("WARNING: shader {} is missing uniform {}", self.name, uniform);
			None
		} else {
			Some(Uniform{index, program: self.program})
		}
	} }
	/// Make the shader active and return a handle for drawing operations to use.
	pub fn draw(&self, window: &mut Window) -> ActiveShader<Vertex> { unsafe {
		let gl = window.get_gl(&self.gl);
		(gl.UseProgram)(self.program);
		if cfg!(debug_assertions) {
			(gl.ValidateProgram)(self.program);
			let mut length = 0 as GLint;
			(gl.GetProgramiv)(self.program, GL_INFO_LOG_LENGTH, &mut length);
			if length > 0 {
				let mut log = vec![0 as u8; length as usize];
				(gl.GetProgramInfoLog)(self.program, log.len() as GLsizei, std::ptr::null_mut(), log.as_mut_ptr() as *mut GLchar);
				assert!(log.pop().unwrap() == 0); // drop null terminator
				eprintln!("{}", std::str::from_utf8(&log).unwrap());
			}
		}
		ActiveShader{shader: self}
	} }
	/// Attach a named vertex attribute to a named variable in the shader.
	/// This is for use by implementations of `VertexBinding`.
	/// `shader` is the name of the shader used for debugging output.
	/// `program` is the OpenGL program object to bind the attribute to.
	pub unsafe fn bind_attribute(gl: &GL, shader: &str, program: GLuint, index: GLuint, attribute: &str) {
		let c_attribute = CString::new(attribute).unwrap();
		if cfg!(debug_assertions) && (gl.GetAttribLocation)(program, c_attribute.as_ptr()) < 0 {
			eprintln!("WARNING: shader {} is missing attribute {}", shader, attribute);
		}
		(gl.BindAttribLocation)(program, index, c_attribute.as_ptr());
	}
}
impl<Vertex: VertexBinding> Drop for Shader<Vertex> {
	fn drop(&mut self) {
		if let Some(gl) = self.gl.upgrade() {
			unsafe{(gl.DeleteProgram)(self.program)};
		}
	}
}

/// Part of a shader after it is loaded and compiled but before the two pieces
/// are linked into a program.
struct ShaderObject<'a> {
	gl: &'a GL,
	shader: GLuint,
}
impl<'a> ShaderObject<'a> {
	pub fn new(gl: &'a GL, name: &str, source: &str, mode: GLenum) -> Result<Self, String> { unsafe {
		// Create shader object.
		assert!(mode == GL_VERTEX_SHADER || mode == GL_FRAGMENT_SHADER);
		let shader = (gl.CreateShader)(mode);
		if shader == 0 { panic!("GLES: failed to create shader"); }
		let out = Self{gl, shader};

		// Attach source code.
		let pointer = source.as_bytes().as_ptr() as *const GLchar;
		let length = source.as_bytes().len() as GLint;
		(gl.ShaderSource)(out.shader, 1, &pointer, &length);

		// Compile.
		(gl.CompileShader)(out.shader);
		let mut compiled = GL_FALSE as GLint;
		(gl.GetShaderiv)(shader, GL_COMPILE_STATUS, &mut compiled);
		if compiled != GL_TRUE as GLint {
			if cfg!(debug_assertions) {
				let mut length = 0;
				(gl.GetShaderiv)(out.shader, GL_INFO_LOG_LENGTH, &mut length);
				let mut log = vec![0 as u8; length as usize];
				(gl.GetShaderInfoLog)(out.shader, log.len() as GLsizei, std::ptr::null_mut(), log.as_mut_ptr() as *mut GLchar);
				assert!(log.pop().unwrap() == 0); // drop null terminator
				eprintln!("{}", std::str::from_utf8(&log).unwrap());
			}
			let mode = if mode == GL_VERTEX_SHADER { "vertex" } else { "fragment" };
			return Err(format!("failed to compile {} shader {}", mode, name));
		}

		Ok(out)
	} }
}
impl<'a> Drop for ShaderObject<'a> {
	fn drop(&mut self) { unsafe {
		(self.gl.DeleteShader)(self.shader);
	} }
}

/// A compiled and linked shader during shader setup.
struct ProgramObject<'a> {
	gl: &'a GL,
	program: GLuint,
}
impl<'a> ProgramObject<'a> {
	pub fn new<V: VertexBinding>(gl: &'a GL, name: &str, vertex: ShaderObject, fragment: ShaderObject) -> Result<Self, String> { unsafe {
		// Create program object.
		let program = (gl.CreateProgram)();
		if program == 0 { panic!("GLES: failed to create shader program"); }
		let out = Self{gl, program};

		// Attach shader objects.
		(gl.AttachShader)(out.program, vertex.shader);
		(gl.AttachShader)(out.program, fragment.shader);

		// Link to get errors and find out which attributes are used.
		(gl.LinkProgram)(out.program);
		let mut linked = GL_FALSE as GLint;
		(gl.GetProgramiv)(out.program, GL_LINK_STATUS, &mut linked);
		if linked != GL_TRUE as GLint {
			if cfg!(debug_assertions) {
				let mut length = 0 as GLint;
				(gl.GetProgramiv)(out.program, GL_INFO_LOG_LENGTH, &mut length);
				let mut log = vec![0 as u8; length as usize];
				(gl.GetProgramInfoLog)(out.program, log.len() as GLsizei, std::ptr::null_mut(), log.as_mut_ptr() as *mut GLchar);
				assert!(log.pop().unwrap() == 0); // drop null terminator
				eprintln!("{}", std::str::from_utf8(&log).unwrap());
			}
			return Err(format!("failed to link shader {}", name));
		}

		// Bind vertex attributes and re-link to apply the bindings.
		V::bind(gl, name, program);
		(gl.LinkProgram)(out.program);
		let mut linked = GL_FALSE as GLint;
		(gl.GetProgramiv)(out.program, GL_LINK_STATUS, &mut linked);
		assert!(linked == GL_TRUE as GLint);

		Ok(out)
	} }
	pub unsafe fn to_raw(self) -> GLuint {
		let out = self.program;
		std::mem::forget(self);
		out
	}
}
impl<'a> Drop for ProgramObject<'a> {
	fn drop(&mut self) { unsafe {
		(self.gl.DeleteProgram)(self.program);
	} }
}
