//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use std::io::{Error, ErrorKind, Read};
use std::rc::Weak;
use super::ffi_gles::*;
use super::metrics::MetricUse;
use super::window::Window;

#[derive(Clone, Copy, Debug, Eq, PartialEq)]
pub enum Repeat { None, Copy, Mirror }
impl Repeat {
	fn to_raw(&self) -> GLint {
		match self {
			Self::None => GL_CLAMP_TO_EDGE,
			Self::Copy => GL_REPEAT,
			Self::Mirror => GL_MIRRORED_REPEAT,
		}
	}
}

pub struct Texture {
	gl: Weak<GL>,
	height: f32,
	texture: GLuint,
	#[allow(dead_code)]
	texture_bytes: MetricUse,
	width: f32,
}
impl Texture {
	/// Upload an RGBA image to the GPU.
	/// `rgba` is pixel data, one byte per channel, four bytes per pixel.
	/// Pixels are packed one after another, so the array contains
	/// `[r0, g0, b0, a0, r1, g1, b1, a1, r2...]`.
	/// Pixels are in rows, bottom to top, so the first pixel given will appear
	/// at texture coordinate 0,0 in the bottom left of the image.
	/// (this order comes from OpenGL)
	/// `width` and `height are the dimensions of the image, and must be powers of
	/// two to meet OpenGL's requirements.
	/// `interpolate` enables linear inerpolation (i.e. smooth up-scaling) if
	/// `true`, otherwise nearest neighbour is used.
	pub fn new(window: &mut Window, rgba: &[u8], width: u32, height: u32, interpolate: bool, repeat: Repeat) -> Self { unsafe {
		let texture_bytes = window.metrics().texture_bytes.take(rgba.len());

		let gl = window.gl();

		debug_assert!(width.is_power_of_two(), "texture width must be a power of two");
		debug_assert!(height.is_power_of_two(), "texture heigh must be a power of two");
		debug_assert!(rgba.len() == width as usize*height as usize*4);

		let mut texture = 0;
		(gl.GenTextures)(1, &mut texture);
		assert!(texture != 0);
		(gl.BindTexture)(GL_TEXTURE_2D, texture);
		(gl.TexImage2D)(GL_TEXTURE_2D, 0, GL_RGBA as GLint, width as GLsizei, height as GLsizei, 0, GL_RGBA, GL_UNSIGNED_BYTE, rgba.as_ptr() as *const GLvoid);
		(gl.GenerateMipmap)(GL_TEXTURE_2D);
		let interpolation = if interpolate { GL_LINEAR } else { GL_NEAREST };
		(gl.TexParameteri)(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, interpolation);
		(gl.TexParameteri)(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR as GLint);
		let repeat = repeat.to_raw();
		(gl.TexParameteri)(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, repeat);
		(gl.TexParameteri)(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, repeat);

		gl.check();
		Self {
			gl: window.weak_gl(),
			height: height as f32,
			texture,
			texture_bytes,
			width: width as f32,
		}
	} }
	/// Load a test image with squares in this pattern:
	/// `red   | yellow`
	/// `--------------`
	/// `green | blue  `
	/// The blue square is 50% transparent.
	pub fn test(window: &mut Window) -> Self {
		Self::new(window, &[
			0x00, 0xff, 0x00, 0xff,
			0x00, 0x00, 0xff, 0x80,
			0xff, 0x00, 0x00, 0xff,
			0xff, 0xff, 0x00, 0xff,
		], 2, 2, false, Repeat::None)
	}
	/// Read texture from file embedded in program using `include_bytes!()`.
	pub fn embedded(window: &mut Window, mut data: &[u8], interpolate: bool, repeat: Repeat) -> Self {
		Self::read(window, &mut data, interpolate, repeat).unwrap()
	}
	/// Read texture from file.
	/// Expected format is:
	/// - width: u8 - stored as a power of two
	/// - height: u8 - stored as a power of two
	/// - red: [u8] - stored as run length encoded bytes
	/// - green: [u8] - stored as run length encoded bytes
	/// - blue: [u8] - stored as run length encoded bytes
	/// - alpha: [u8] - stored as run length encoded bytes
	pub fn read<R: Read>(window: &mut Window, mut from: R, interpolate: bool, repeat: Repeat) -> Result<Self, Error> {
		// Read packed width and height.
		let mut exponents = [0 as u8; 2];
		from.read_exact(&mut exponents)?;
		assert!(exponents.iter().all(|value| *value <= 12), "unreasonably large texture (>4,000 pixels)");
		let width = 1<<exponents[0] as i32;
		let height = 1<<exponents[1] as i32;
		// Unpack run length encoded colour planes to build completed RGBA image.
		let planes = 4; // RGBA
		let mut rgba = vec![0 as u8; width as usize*height as usize*planes].into_boxed_slice();
		for offset in 0..planes {
			let mut index = offset;
			let mut buffer = [0 as u8; 1];
			while index < rgba.len() {
				from.read_exact(&mut buffer)?;
				if buffer[0]&0b11000000 == 0b10000000 {
					// Encoded repeating value (a "run"), extract number of repeats and read
					// value.
					let run = ((buffer[0]&0b00111111)+1) as usize;
					if index+run*planes > rgba.len()+offset {
						return Err(Error::new(ErrorKind::InvalidData, "overlong image plane"));
					}
					from.read_exact(&mut buffer)?;
					// Add repeated value within the current colour plane.
					for _ in 0..run {
						rgba[index] = buffer[0];
						index += planes;
					}
				} else {
					// Raw pixel.
					rgba[index] = buffer[0];
					index += planes;
				}
			}
		}
		// Load texture.
		Ok(Self::new(window, &rgba, width, height, interpolate, repeat))
	}
	pub fn height(&self) -> f32 { self.height }
	pub fn width(&self) -> f32 { self.width }
	pub unsafe fn raw(&self) -> GLuint { self.texture }
}
impl Drop for Texture {
	fn drop(&mut self) { unsafe {
		if let Some(gl) = self.gl.upgrade() {
			(gl.DeleteTextures)(1, &self.texture);
		}
	} }
}
