//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use std::collections::HashMap;
use std::fs::File;
use std::io::{BufRead, BufReader, Error, ErrorKind, Write};
use std::time::{Duration, Instant};
use super::vector3::{F3, f3};

/// Loads variables from a file (`/tmp/tweak.cfg`) once a second to make
/// tweaking things easier.
/// The file is automatically created if it does not exist.
pub struct Tweak {
	last: Instant,
	values: HashMap<&'static str, Value>,
}
impl Tweak {
	const PATH: &'static str = "/tmp/tweak.cfg";
	pub fn new() -> Self {
		Self {
			last: Instant::now(),
			values: HashMap::new(),
		}
	}
	/// Poll file for new values if enough time has passed.
	pub fn tick(&mut self) {
		let now = Instant::now();
		if now.duration_since(self.last) > Duration::new(1, 0) {
			self.last = now;
			match File::open(Self::PATH) {
				Ok(file) => if let Err(error) = self.read(file) {
					eprintln!("WARNING: failed to read tweak file, {}", error);
				},
				Err(error) if error.kind() == ErrorKind::NotFound => if let Err(error) = self.write() {
					eprintln!("WARNING: failed to write tweak file, {}", error);
				},
				Err(error) => eprintln!("WARNING: failed to open tweak file, {}", error),
			}
		}
	}
	/// Get boolean value.
	pub fn bool(&mut self, name: &'static str, value: &mut bool) {
		*value = match *self.values.entry(name).or_insert(Value::Bool(*value)) {
			Value::Bool(value) => value,
			_ => panic!("{} has varying type", name),
		};
	}
	/// Get vector value.
	pub fn f3(&mut self, name: &'static str, value: &mut F3) {
		*value = match *self.values.entry(name).or_insert(Value::F3(*value)) {
			Value::F3(value) => value,
			_ => panic!("{} has varying type", name),
		};
	}
	/// Get floating point value.
	pub fn f32(&mut self, name: &'static str, value: &mut f32) {
		*value = match *self.values.entry(name).or_insert(Value::F32(*value)) {
			Value::F32(value) => value,
			_ => panic!("{} has varying type", name),
		};
	}

	fn read(&mut self, file: File) -> Result<(), Error> {
		let reader = BufReader::new(file);
		for (index, line) in reader.lines().enumerate() {
			let count = index+1;
			let line = line?;
			if line.starts_with("#") { continue; }
			let mut iter = line.split('=');
			let (left, right) = match (iter.next(), iter.next(), iter.next()) {
				(Some(left), Some(right), None) => (left, right),
				(_, _, Some(_)) => return Err(Error::new(ErrorKind::InvalidData,
					format!("{}: too many equals signs", count))),
				_ => return Err(Error::new(ErrorKind::InvalidData,
					format!("{}: missing equals sign", count))),
			};
			let name = left.trim();
			let value = right.trim();
			if let Some(out) = self.values.get_mut(name) {
				match out {
					Value::Bool(out) => *out = Self::parse_bool(value, count)?,
					Value::F3(out) => *out = Self::parse_f3(value, count)?,
					Value::F32(out) => *out = Self::parse_f32(value, count)?,
				}
			}
		}
		Ok(())
	}
	fn write(&self) -> Result<(), Error> {
		let mut file = File::create(Self::PATH)?;
		for (name, value) in self.values.iter() {
			write!(file, "{} = ", name)?;
			match value {
				Value::Bool(value) => writeln!(file, "{}", if *value { "true" } else { "false" })?,
				Value::F3(value) => writeln!(file, "{},{},{}", value.x, value.y, value.z)?,
				Value::F32(value) => writeln!(file, "{}", value)?,
			}
		}
		Ok(())
	}
	fn parse_bool(value: &str, count: usize) -> Result<bool, Error> {
		match value {
			"true" => Ok(true),
			"false" => Ok(false),
			value => Err(Error::new(ErrorKind::InvalidData,
				format!("{}: invalid bool, {} is not true or false", count, value))),
		}
	}
	fn parse_f3(value: &str, count: usize) -> Result<F3, Error> {
		let mut iter = value.split(',')
			.map(|value: &str| -> Result<f32, Error> {
				Self::parse_f32(value.trim(), count)
			});
		match (iter.next(), iter.next(), iter.next(), iter.next()) {
			(Some(x), Some(y), Some(z), None) => Ok(f3(x?, y?, z?)),
			_ => Err(Error::new(ErrorKind::InvalidData, format!("{}: expected three components", count))),
		}
	}
	fn parse_f32(value: &str, count: usize) -> Result<f32, Error> {
		value.parse::<f32>()
			.map_err(|error| Error::new(ErrorKind::InvalidData,
				format!("{}: failed to parse float, {}", count, error)))
	}
}

#[derive(Clone, Copy, Debug, PartialEq)]
enum Value {
	Bool(bool),
	F3(F3),
	F32(f32),
}
