//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use super::ffi_gles::*;
use super::offset_of;
use super::shader::Shader;

/// Type used to represent vertex indices.
pub type Index = u16;
/// Description of `Index` as an OpenGL type.
pub const INDEX_TYPE: GLenum = GL_UNSIGNED_SHORT;

/// Ways to use vertices for drawing.
#[derive(Clone, Copy, Debug, Eq, PartialEq)]
pub enum Primitive { Lines, Triangles }
impl Primitive {
	/// Convert to OpenGL value.
	pub fn raw(&self) -> GLenum {
		match self {
			Primitive::Lines => GL_LINES,
			Primitive::Triangles => GL_TRIANGLES,
		}
	}
	/// Return number of vertices needed for each shape.
	pub fn vertices(&self) -> usize {
		match self {
			Primitive::Lines => 2,
			Primitive::Triangles => 3,
		}
	}
}

/// Trait for structures used to pass arrays of vertex attribute data to
/// OpenGL to be loaded into a vertex buffer object.
/// The layout of these structures must conform to the OpenGL specification so
/// this trait is `unsafe`.
///
/// In normal use, each field of a structure this trait is implemented for
/// provides the value for one vertex attribute.
/// Each attribute will have a fixed index used to identify it to OpenGL.
/// When a shader is loaded, the attribute indices for the vertex type it will
/// be used with will be setup (so that they only need to be looked up once).
/// Then, each time an array of that type of vertex is drawn the offsets of
/// the fields will be setup just before drawing.
/// (ideally we would only do this once, but OpenGLES 2 doesn't support VAOs)
pub unsafe trait VertexBinding {
	/// Set vertex attribute indices for whatever shader is active.
	/// This is called just after compiling a shader to setup its attributes.
	/// `shader` is the name of the shader for debugging output.
	/// `program` is the OpenGL handle for the shader program.
	unsafe fn bind(gl: &GL, shader: &str, program: GLuint);
	/// Set vertex attribute fields offsets and enable provided attributes.
	/// This is called just before drawing with a vertex buffer of this type of
	/// vertex.
	unsafe fn load(gl: &GL);
}

/// A standard vertex type that should cover many use cases.
#[derive(Clone, Debug, PartialEq)]
pub struct Vertex {
	/// Location in model coordinates.
	pub position: [GLfloat; 3],
	/// where to map this vertex in a texture (for UV mapping).
	pub texcoord: [GLfloat; 2],
	/// Unit vector pointing out of the mesh (for lighting).
	pub normal: [GLfloat; 3],
}
impl Vertex {
	const INDEX_POSITION: GLuint = 0;
	const INDEX_TEXCOORD: GLuint = 1;
	const INDEX_NORMAL: GLuint = 2;
}
unsafe impl VertexBinding for Vertex {
	unsafe fn bind(gl: &GL, shader: &str, program: GLuint) {
		Shader::<Self>::bind_attribute(gl, shader, program, Self::INDEX_POSITION, "position");
		Shader::<Self>::bind_attribute(gl, shader, program, Self::INDEX_TEXCOORD, "texcoord");
		Shader::<Self>::bind_attribute(gl, shader, program, Self::INDEX_NORMAL, "normal");
	}
	unsafe fn load(gl: &GL) {
		let stride = std::mem::size_of::<Self>() as GLsizei;

		(gl.EnableVertexAttribArray)(Self::INDEX_POSITION);
		(gl.VertexAttribPointer)(Self::INDEX_POSITION, 3, GL_FLOAT, GL_FALSE, stride,
			offset_of!(Self, position) as *const GLvoid);

		(gl.EnableVertexAttribArray)(Self::INDEX_TEXCOORD);
		(gl.VertexAttribPointer)(Self::INDEX_TEXCOORD, 3, GL_FLOAT, GL_FALSE, stride,
			offset_of!(Self, texcoord) as *const GLvoid);

		(gl.EnableVertexAttribArray)(Self::INDEX_NORMAL);
		(gl.VertexAttribPointer)(Self::INDEX_NORMAL, 3, GL_FLOAT, GL_FALSE, stride,
			offset_of!(Self, normal) as *const GLvoid);
	}
}
