//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use std::ffi::{CStr, CString};
use std::rc::{Rc, Weak};
use super::event::{Button, ControllerAxis, ControllerButton, Event, Modifiers};
use super::ffi_gles::*;
use super::ffi_sdl::*;
use super::fps::FPS;
use super::metrics::Metrics;
use super::rgba::rgba;
use super::vector2::{f2, F2};

pub struct Window {
	benchmark: bool,
	context: *mut SDL_GLContext,
	controllers: Vec<(u32, *mut SDL_GameController)>,
	fps: FPS,
	gl: Rc<GL>,
	metrics: Metrics,
	window: *mut SDL_Window,
}
impl Window {
	pub const FPS: f32 = 60.0;
	pub const FRAME_SECONDS: f32 = 1.0/Self::FPS;

	pub fn new(title: &str) -> Result<Self, String> { unsafe {
		// Initialise SDL.
		if SDL_Init(SDL_INIT_VIDEO|SDL_INIT_GAMECONTROLLER) != 0 { return Self::sdl_error(); }
		let flags = if cfg!(debug_assertions) { SDL_GL_CONTEXT_DEBUG_FLAG } else { 0 };
		if SDL_GL_SetAttribute(SDL_GLattr::SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_ES) != 0 ||
			SDL_GL_SetAttribute(SDL_GLattr::SDL_GL_CONTEXT_MAJOR_VERSION, 2) != 0 ||
			SDL_GL_SetAttribute(SDL_GLattr::SDL_GL_CONTEXT_MINOR_VERSION, 0) != 0 ||
			SDL_GL_SetAttribute(SDL_GLattr::SDL_GL_CONTEXT_FLAGS, flags) != 0 {
			return Self::sdl_error();
		}
		SDL_GL_SetAttribute(SDL_GLattr::SDL_GL_STENCIL_SIZE, 1);

		// Create window.
		let c_title = CString::new(title).unwrap();
		let x = SDL_WINDOWPOS_UNDEFINED;
		let y = SDL_WINDOWPOS_UNDEFINED;
		let width = 640;
		let height = 480;
		let flags = SDL_WINDOW_OPENGL|SDL_WINDOW_MAXIMIZED|SDL_WINDOW_RESIZABLE;
		let window = SDL_CreateWindow(c_title.as_ptr(), x, y, width, height, flags);
		if window.is_null() { return Self::sdl_error(); }
		if SDL_SetRelativeMouseMode(true) != 0 { return Self::sdl_error(); }

		// Create OpenGL context.
		let context = SDL_GL_CreateContext(window);
		if context.is_null() { return Self::sdl_error(); }
		let gl = Rc::new(GL::load(context)?);
		Self::resize(window, &gl);

		let fps = FPS::new(Self::FPS as u32);
		Ok(Self{
			benchmark: false,
			context,
			controllers: Vec::new(),
			fps,
			gl,
			metrics: Metrics::new(),
			window,
		})
	} }
	/// Return window size in pixels (this is cheap to call).
	pub fn size(&self) -> F2 { unsafe {
		let mut width = 0;
		let mut height = 0;
		SDL_GetWindowSize(self.window, &mut width, &mut height);
		f2(width as f32, height as f32)
	} }
	/// Disable vsync and framerate regulation to test performance.
	pub fn benchmark(&mut self) -> Result<(), String> { unsafe {
		if SDL_GL_SetSwapInterval(0) != 0 { return Self::sdl_error(); }
		self.benchmark = true;
		Ok(())
	} }
	/// Return the next queued input event.
	pub fn poll(&mut self) -> Option<Event> { unsafe {
		fn button_from_raw(raw: u8) -> Option<Button> {
			match raw {
				SDL_BUTTON_LEFT => Some(Button::Left),
				SDL_BUTTON_MIDDLE => Some(Button::Middle),
				SDL_BUTTON_RIGHT => Some(Button::Right),
				_ => None,
			}
		}
		fn controller_axis_from_raw(raw: u8) -> ControllerAxis {
			match raw {
				SDL_CONTROLLER_AXIS_LEFTX => ControllerAxis::LeftX,
				SDL_CONTROLLER_AXIS_LEFTY => ControllerAxis::LeftY,
				SDL_CONTROLLER_AXIS_RIGHTX => ControllerAxis::RightX,
				SDL_CONTROLLER_AXIS_RIGHTY => ControllerAxis::RightY,
				SDL_CONTROLLER_AXIS_TRIGGERLEFT => ControllerAxis::LeftTrigger,
				SDL_CONTROLLER_AXIS_TRIGGERRIGHT => ControllerAxis::RightTrigger,
				_ => unreachable!(),
			}
		}
		fn controller_button_from_raw(raw: u8) -> Option<ControllerButton> {
			match raw {
				SDL_CONTROLLER_BUTTON_A => Some(ControllerButton::A),
				SDL_CONTROLLER_BUTTON_B => Some(ControllerButton::B),
				SDL_CONTROLLER_BUTTON_X => Some(ControllerButton::X),
				SDL_CONTROLLER_BUTTON_Y => Some(ControllerButton::Y),
				SDL_CONTROLLER_BUTTON_BACK => Some(ControllerButton::Back),
				SDL_CONTROLLER_BUTTON_START => Some(ControllerButton::Start),
				SDL_CONTROLLER_BUTTON_LEFTSTICK => Some(ControllerButton::LeftStick),
				SDL_CONTROLLER_BUTTON_RIGHTSTICK => Some(ControllerButton::RightStick),
				SDL_CONTROLLER_BUTTON_LEFTSHOULDER => Some(ControllerButton::LeftShoulder),
				SDL_CONTROLLER_BUTTON_RIGHTSHOULDER => Some(ControllerButton::RightShoulder),
				SDL_CONTROLLER_BUTTON_DPAD_UP => Some(ControllerButton::DpadUp),
				SDL_CONTROLLER_BUTTON_DPAD_DOWN => Some(ControllerButton::DpadDown),
				SDL_CONTROLLER_BUTTON_DPAD_LEFT => Some(ControllerButton::DpadLeft),
				SDL_CONTROLLER_BUTTON_DPAD_RIGHT => Some(ControllerButton::DpadRight),
				_ => None,
			}
		}
		fn pixel_to_gl(window: *mut SDL_Window, x: i32, y: i32) -> F2 {
			let mut width = 0;
			let mut height = 0;
			unsafe{SDL_GetWindowSize(window, &mut width, &mut height)};
			f2(x as f32*2.0/width as f32-1.0, y as f32*2.0/height as f32-1.0)
		}

		loop {
			let mut event = std::mem::MaybeUninit::<SDL_Event>::uninit();
			if SDL_PollEvent(event.as_mut_ptr()) == 0 { break; }
			let event = event.assume_init();
			match event.tag {
				SDL_CONTROLLERAXISMOTION => return Some(Event::ControllerAxis {
					id: event.caxis.which as u32,
					axis: controller_axis_from_raw(event.caxis.axis),
					value: event.caxis.value as f32/-(i16::MIN as f32),
				}),
				SDL_CONTROLLERBUTTONDOWN | SDL_CONTROLLERBUTTONUP => if let Some(button) = controller_button_from_raw(event.cbutton.button) {
					return Some(Event::ControllerButton {
						id: event.cbutton.which as u32,
						button,
						down: event.cbutton.state == 1,
					});
				},
				SDL_CONTROLLERDEVICEADDED => {
					let controller = SDL_GameControllerOpen(event.cdevice.which);
					if controller.is_null() {
						eprintln!("WARNING: failed to add controller");
						continue;
					}
					let joystick = SDL_GameControllerGetJoystick(controller);
					assert!(!joystick.is_null());
					let id = SDL_JoystickInstanceID(joystick);
					assert!(id >= 0);
					let id = id as u32;
					debug_assert!(self.controllers.iter().copied().all(|(used_id, _)| used_id != id));
					self.controllers.push((id, controller));
					return Some(Event::ControllerAdded{id});
				},
				SDL_CONTROLLERDEVICEREMOVED => {
					debug_assert!(event.cdevice.which >= 0);
					let id = event.cdevice.which as u32;
					let mut index = 0;
					while index < self.controllers.len() {
						if self.controllers[index].0 == id {
							SDL_GameControllerClose(self.controllers[index].1);
							self.controllers.swap_remove(index);
						} else {
							index += 1;
						}
					}
					return Some(Event::ControllerLost{id});
				},
				SDL_KEYDOWN | SDL_KEYUP => return Some(Event::Key {
					key: event.key.keysym.sym,
					down: event.key.state != 0,
					modifiers: Modifiers {
						alt: event.key.keysym.modifiers & KMOD_ALT != 0,
						ctrl: event.key.keysym.modifiers & KMOD_CTRL != 0,
						shift: event.key.keysym.modifiers & KMOD_SHIFT != 0,
					},
					repeat: event.key.repeat != 0,
				}),
				SDL_MOUSEBUTTONDOWN | SDL_MOUSEBUTTONUP => if let Some(button) = button_from_raw(event.button.button) {
					return Some(Event::Button{
						button,
						down: event.button.state != 0,
						position: pixel_to_gl(self.window, event.button.x, event.button.y),
					});
				},
				SDL_MOUSEMOTION => return Some(Event::Motion {
					position: pixel_to_gl(self.window, event.motion.x, event.motion.y),
					relative: f2(event.motion.xrel as f32, event.motion.yrel as f32),
				}),
				SDL_MOUSEWHEEL => return Some(Event::Wheel{x: event.wheel.x, y: event.wheel.y}),
				SDL_QUIT => return Some(Event::Quit),
				SDL_WINDOWEVENT if event.window.event == WindowEvent::Resized => {
					Self::resize(self.window, &self.gl);
					return Some(Event::Resize);
				},
				_ => (),
			}
		}
		None
	} }
	/// Enable or disable depth sorted rendering.
	/// When enabled, faces pointing away from the camera are discarded* and faces
	/// nearer the camera are drawn over faces further away (using the depth
	/// buffer).
	/// The depth buffer is cleared every time depth mode is enabled.
	pub fn depth(&mut self, enable: bool) { unsafe {
		if enable {
			// We also clear the stencil buffer here incase it's used.
			// Ideally, we'd do this when it's used later, but it appears that the
			// Raspberry Pi Zero W (revision 1.1) messes up the depth buffer if asked
			// to clear only the stencil buffer (with `glClear(GL_STENCIL_BUFFER_BIT)`).
			(self.gl.Clear)(GL_DEPTH_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);
			(self.gl.Enable)(GL_DEPTH_TEST);
			(self.gl.Enable)(GL_CULL_FACE);
		} else {
			(self.gl.Disable)(GL_CULL_FACE);
			(self.gl.Disable)(GL_DEPTH_TEST);
		}
		self.gl.check();
	} }
	/// Flush drawing operation to the screen and wait for the end of the current
	/// frame.
	/// In debug builds, the screen will be cleared to magenta to help dianose
	/// missing background bugs.
	pub fn frame(&mut self) {
		self.gl.check();
		unsafe{SDL_GL_SwapWindow(self.window)};
		if cfg!(debug_assertions) {
			rgba(0xff00ffff).draw(self);
		}
		if !self.benchmark {
			self.fps.tick();
		}
	}
	/// Return a reference to OpenGL functions.
	/// This is primarily for internal use as OpenGL keeps all sorts of state
	/// hanging around.
	/// This handle will expire automatically when the context is destroyed
	/// (hence using `Weak`). Automatic expiration is useful to ensure no
	/// `drop()` methods try to free OpenGL resources after the context has been
	//// destroyed.
	pub fn weak_gl(&self) -> Weak<GL> {
		Rc::downgrade(&self.gl)
	}
	/// Return a reference to OpenGL functions via checking a weak reference came from this window.
	pub fn get_gl(&self, gl: &Weak<GL>) -> &GL {
		let same_gl = gl.upgrade().map(|rc| Rc::ptr_eq(&rc, &self.gl)).unwrap_or(false);
		debug_assert!(same_gl);
		&self.gl
	}
	/// Return a reference to OpenGL functions.
	/// This is primarily for internal use, and marked unsafe because using it to
	/// skip the checks in `get_gl()` could cause accidental use of resources from
	/// the wrong OpenGL context.
	/// Most code (internal included) should be using `weak_gl()` and `get_gl()`
	/// so they can make sure they don't try to release OpenGL resources after
	/// the OpenGL context and `Window` have been destroyed.
	pub unsafe fn gl(&self) -> &GL { &self.gl }
	/// Convert an SDL error to `Result::<_, String>::Err(_)`.
	pub unsafe fn sdl_error<T>() -> Result<T, String> {
		let message = SDL_GetError();
		assert!(!message.is_null());
		let message = CStr::from_ptr(message).to_str().unwrap().to_string();
		assert!(message.len() > 0);
		Err(message)
	}
	/// Return a reference to statistics tracked over the program's lifecycle.
	/// This is intended mostly for internal use, but allows any caller to add to
	/// the tracked resource totals.
	pub fn metrics(&mut self) -> &mut Metrics {
		&mut self.metrics
	}

	fn resize(window: *mut SDL_Window, gl: &GL) { unsafe {
		let mut width = 0;
		let mut height = 0;
		SDL_GetWindowSize(window, &mut width, &mut height);
		(gl.Viewport)(0, 0, width as GLsizei, height as GLsizei);
	} }
}
impl Drop for Window {
	fn drop(&mut self) { unsafe {
		self.controllers.iter().copied().for_each(|(_, controller)| SDL_GameControllerClose(controller));
		SDL_GL_DeleteContext(self.context);
		SDL_DestroyWindow(self.window);
		SDL_Quit();
		self.metrics.dump();
	} }
}
