//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

#![allow(dead_code)]

use std::io::{Error, ErrorKind, Read};

pub struct Bitmap {
	height: usize,
	pub rgba: Box<[u8]>,
	width: usize,
}
impl Bitmap {
	pub const PLANES: usize = 4;
	pub fn new(width: usize, height: usize) -> Self {
		Self {
			height,
			rgba: vec![0 as u8; width*height*Self::PLANES].into_boxed_slice(),
			width,
		}
	}
	pub fn embedded(mut data: &[u8]) -> Self {
		Self::read(&mut data).unwrap()
	}
	pub fn read<R: Read>(mut from: R) -> Result<Self, Error> {
		// Read packed width and height.
		let mut exponents = [0 as u8; 2];
		from.read_exact(&mut exponents)?;
		assert!(exponents.iter().all(|value| *value <= 12), "unreasonably large texture (>4,000 pixels)");
		let width = 1<<exponents[0] as usize;
		let height = 1<<exponents[1] as usize;
		// Unpack run length encoded colour planes to build completed RGBA image.
		let mut rgba = vec![0 as u8; width*height*Self::PLANES].into_boxed_slice();
		for offset in 0..Self::PLANES {
			let mut index = offset;
			let mut buffer = [0 as u8; 1];
			while index < rgba.len() {
				from.read_exact(&mut buffer)?;
				if buffer[0]&0b11000000 == 0b10000000 {
					// Encoded repeating value (a "run"), extract number of repeats and read
					// value.
					let run = ((buffer[0]&0b00111111)+1) as usize;
					if index+run*Self::PLANES > rgba.len()+offset {
						return Err(Error::new(ErrorKind::InvalidData, "overlong image plane"));
					}
					from.read_exact(&mut buffer)?;
					// Add repeated value within the current colour plane.
					for _ in 0..run {
						rgba[index] = buffer[0];
						index += Self::PLANES;
					}
				} else {
					// Raw pixel.
					rgba[index] = buffer[0];
					index += Self::PLANES;
				}
			}
		}
		Ok(Self{height, rgba, width})
	}
	pub fn width(&self) -> usize { self.width }
	pub fn height(&self) -> usize { self.height }
	pub fn get(&self, x: i32, y: i32) -> Option<[u8; 4]> {
		self.index(x, y).map(|start| [
			self.rgba[start],
			self.rgba[start+1],
			self.rgba[start+2],
			self.rgba[start+3],
		])
	}
	pub fn set(&mut self, x: i32, y: i32, rgba: [u8; 4]) -> Option<[u8; 4]> {
		self.index(x, y).map(|start| {
			let value = [
				self.rgba[start],
				self.rgba[start+1],
				self.rgba[start+2],
				self.rgba[start+3],
			];
			self.rgba[start..start+Self::PLANES].copy_from_slice(&rgba);
			value
		})
	}

	fn index(&self, x: i32, y: i32) -> Option<usize> {
		if x < 0 || y < 0 || x as usize >= self.width || y as usize >= self.height {
			None
		} else {
			Some((y as usize*self.width+x as usize)*Self::PLANES)
		}
	}
}
