//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use engine::{Camera, F2, f2, F3, f3, Rotation};

pub struct FPSCamera {
	pub camera: Camera,
	pub centre: F3,
	pub rotate_x: f32,
	pub rotate_y: f32,
}
impl FPSCamera {
	pub fn new(near: f32, far: f32, fov: f32, viewport: F2, centre: F3) -> Self {
		Self {
			camera: Camera::new(near, far, fov, f2(0.0, 0.0), viewport),
			centre,
			rotate_x: 0.0,
			rotate_y: 0.0,
		}
	}
	pub fn tick(&mut self) {
		if self.rotate_x < -std::f32::consts::PI*0.5 { self.rotate_x = -std::f32::consts::PI*0.5; }
		if self.rotate_x > std::f32::consts::PI*0.5 { self.rotate_x = std::f32::consts::PI*0.5; }
		if self.rotate_y < 0.0 { self.rotate_y += std::f32::consts::PI*2.0; }
		if self.rotate_y > std::f32::consts::PI*2.0 { self.rotate_y -= std::f32::consts::PI*2.0; }
		let rotate = Rotation::new(f3(1.0, 0.0, 0.0), -self.rotate_x)
			.then(Rotation::new(f3(0.0, 1.0, 0.0), self.rotate_y));
		self.camera.place(self.centre, rotate);
	}
}

pub struct FollowCamera {
	pub camera: Camera,
	pub centre: F3,
	last_look: F3,
	last_up: F3,
	pub offset: F3,
	up: F3,
}
impl FollowCamera {
	pub fn new(near: f32, far: f32, fov: f32, viewport: F2, centre: F3) -> Self {
		Self {
			camera: Camera::new(near, far, fov, f2(0.0, 0.0), viewport),
			centre,
			last_look: f3(0.0, 0.0, 1.0),
			last_up: f3(0.0, 1.0, 0.0),
			offset: f3(0.0, 0.5, -2.0),
			up: f3(0.0, 1.0, 0.0),
		}
	}
	pub fn tick(&mut self, target: F3, rotation: Rotation) {
		self.centre = self.centre.lerp(target+rotation.apply(self.offset), 0.1);
		let up = rotation.apply(f3(0.0, 1.0, 0.0));
		self.up = self.up.lerp(up, 0.1);
		if let Some(look) = (target-self.centre).unit() {
			self.last_look = look;
		}
		if let Some(up) = self.up.unit() {
			self.last_up = up;
		}
		self.camera.place(self.centre, Rotation::look(self.last_look, self.last_up));
	}
}
