//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

#![allow(non_camel_case_types)]
#![allow(non_snake_case)]
#![allow(dead_code)]

use std::os::raw::{c_char, c_int, c_void};

pub type size_t = usize;

pub type WrenReallocateFn = extern "C" fn(*mut c_void, size_t, *mut c_void) -> *mut c_void;
pub type WrenForeignMethodFn = extern "C" fn(*mut WrenVM);
pub type WrenFinalizerFn = extern "C" fn (*mut c_void);
pub type WrenResolveModuleFn = extern "C" fn (*mut WrenVM, *const c_char, *const c_char) -> *const c_char;
pub type WrenLoadModuleCompleteFn = extern "C" fn (*mut WrenVM, *const c_char, WrenLoadModuleResult);
pub type WrenLoadModuleFn = extern "C" fn (*mut WrenVM, *const c_char) -> WrenLoadModuleResult;
pub type WrenBindForeignMethodFn = extern "C" fn (*mut WrenVM, *const c_char, *const c_char, bool, *const c_char) -> Option<WrenForeignMethodFn>;
pub type WrenWriteFn = extern "C" fn (*mut WrenVM, *const c_char);
pub type WrenErrorFn = extern "C" fn (*mut WrenVM, WrenErrorType, *const c_char, c_int, *const c_char);
pub type WrenBindForeignClassFn = extern "C" fn (*mut WrenVM, *const c_char, *const c_char) -> WrenForeignClassMethods;

#[derive(Clone, Copy, Debug, Eq, PartialEq)]
#[repr(C)]
pub enum WrenErrorType {
	Compile, Runtime, StackTrace
	// WREN_ERROR_COMPILE,
	// WREN_ERROR_RUNTIME,
	// WREN_ERROR_STACK_TRACE
}
#[derive(Clone, Copy, Debug, Eq, PartialEq)]
#[repr(C)]
pub enum WrenInterpretResult {
	Success, CompileError, RuntimeError
	// WREN_RESULT_SUCCESS,
	// WREN_RESULT_COMPILE_ERROR,
	// WREN_RESULT_RUNTIME_ERROR
}
#[derive(Clone, Copy, Debug, Eq, PartialEq)]
#[repr(C)]
pub enum WrenType {
	Bool, Number, Foreign, List, Map, Null, String, Unknown
	// WREN_TYPE_BOOL,
	// WREN_TYPE_NUM,
	// WREN_TYPE_FOREIGN,
	// WREN_TYPE_LIST,
	// WREN_TYPE_MAP,
	// WREN_TYPE_NULL,
	// WREN_TYPE_STRING,
	// WREN_TYPE_UNKNOWN
}
impl WrenType {
	pub fn to_str(&self) -> &'static str {
		match self {
			WrenType::Bool => "bool",
			WrenType::Number => "number",
			WrenType::Foreign => "foreign",
			WrenType::List => "list",
			WrenType::Map => "map",
			WrenType::Null => "null",
			WrenType::String => "string",
			WrenType::Unknown => "unknown",
		}
	}
}

#[repr(C)] pub struct WrenVM { _private: [u8; 0] }
#[repr(C)] pub struct WrenHandle { _private: [u8; 0] }

#[repr(C)] pub struct WrenLoadModuleResult {
	pub source: *const c_char,
	pub onComplete: WrenLoadModuleCompleteFn,
	pub userData: *mut c_void,
}
#[repr(C)] pub struct WrenForeignClassMethods {
	pub allocate: WrenForeignMethodFn,
	pub finalize: Option<WrenFinalizerFn>,
}
#[repr(C)] pub struct WrenConfiguration {
	pub reallocateFn: WrenReallocateFn,
	pub resolveModuleFn: WrenResolveModuleFn,
	pub loadModuleFn: WrenLoadModuleFn,
	pub bindForeignMethodFn: WrenBindForeignMethodFn,
	pub bindForeignClassFn: WrenBindForeignClassFn,
	pub writeFn: WrenWriteFn,
	pub errorFn: WrenErrorFn,
	pub initialHeapSize: size_t,
	pub minHeapSize: size_t,
	pub heapGrowthPercent: c_int,
	pub userData: *mut c_void,
}

#[link(name="wren", kind="static")]
extern "C" {
	pub fn wrenGetVersionNumber() -> c_int;
	pub fn wrenInitConfiguration(configuration: *mut WrenConfiguration);
	pub fn wrenNewVM(configuration: *mut WrenConfiguration) -> *mut WrenVM;
	pub fn wrenFreeVM(vm: *mut WrenVM);
	pub fn wrenCollectGarbage(vm: *mut WrenVM);
	pub fn wrenInterpret(vm: *mut WrenVM, module: *const c_char, source: *const c_char) -> WrenInterpretResult;
	pub fn wrenMakeCallHandle(vm: *mut WrenVM, signature: *const c_char) -> *mut WrenHandle;
	pub fn wrenCall(vm: *mut WrenVM, method: *mut WrenHandle) -> WrenInterpretResult;
	pub fn wrenReleaseHandle(vm: *mut WrenVM, handle: *mut WrenHandle);
	pub fn wrenGetSlotCount(vm: *mut WrenVM) -> c_int;
	pub fn wrenEnsureSlots(vm: *mut WrenVM, numSlots: c_int);
	pub fn wrenGetSlotType(vm: *mut WrenVM, slot: c_int) -> WrenType;
	pub fn wrenGetSlotBool(vm: *mut WrenVM, slot: c_int) -> bool;
	pub fn wrenGetSlotBytes(vm: *mut WrenVM, slot: c_int, length: *mut c_int) -> *const c_char;
	pub fn wrenGetSlotDouble(vm: *mut WrenVM, slot: c_int) -> f64;
	pub fn wrenGetSlotForeign(vm: *mut WrenVM, slot: c_int) -> *mut c_void;
	pub fn wrenGetSlotString(vm: *mut WrenVM, slot: c_int) -> *const c_char;
	pub fn wrenGetSlotHandle(vm: *mut WrenVM, slot: c_int) -> *mut WrenHandle;
	pub fn wrenSetSlotBool(vm: *mut WrenVM, slot: c_int, value: bool);
	pub fn wrenSetSlotBytes(vm: *mut WrenVM, slot: c_int, bytes: *const c_char, length: size_t);
	pub fn wrenSetSlotDouble(vm: *mut WrenVM, slot: c_int, value: f64);
	pub fn wrenSetSlotNewForeign(vm: *mut WrenVM, slot: c_int, classSlot: c_int, size: size_t) -> *mut c_void;
	pub fn wrenSetSlotNewList(vm: *mut WrenVM, slot: c_int);
	pub fn wrenSetSlotNewMap(vm: *mut WrenVM, slot: c_int);
	pub fn wrenSetSlotNull(vm: *mut WrenVM, slot: c_int);
	pub fn wrenSetSlotString(vm: *mut WrenVM, slot: c_int, text: *const c_char);
	pub fn wrenSetSlotHandle(vm: *mut WrenVM, slot: c_int, handle: *mut WrenHandle);
	pub fn wrenGetListCount(vm: *mut WrenVM, slot: c_int) -> c_int;
	pub fn wrenGetListElement(vm: *mut WrenVM, listSlot: c_int, index: c_int, elementSlot: c_int);
	pub fn wrenSetListElement(vm: *mut WrenVM, listSlot: c_int, index: c_int, elementSlot: c_int);
	pub fn wrenInsertInList(vm: *mut WrenVM, listSlot: c_int, index: c_int, elementSlot: c_int);
	pub fn wrenGetMapCount(vm: *mut WrenVM, slot: c_int) -> c_int;
	pub fn wrenGetMapContainsKey(vm: *mut WrenVM, mapSlot: c_int, keySlot: c_int) -> bool;
	pub fn wrenGetMapValue(vm: *mut WrenVM, mapSlot: c_int, keySlot: c_int, valueSlot: c_int);
	pub fn wrenSetMapValue(vm: *mut WrenVM, mapSlot: c_int, keySlot: c_int, valueSlot: c_int);
	pub fn wrenRemoveMapValue(vm: *mut WrenVM, mapSlot: c_int, keySlot: c_int, removedValueSlot: c_int);
	pub fn wrenGetVariable(vm: *mut WrenVM, module: *const c_char, name: *const c_char, slot: c_int);
	pub fn wrenHasVariable(vm: *mut WrenVM, module: *const c_char, name: *const c_char) -> bool;
	pub fn wrenHasModule(vm: *mut WrenVM, module: *const c_char) -> bool;
	pub fn wrenAbortFiber(vm: *mut WrenVM, slot: c_int);
	pub fn wrenGetUserData(vm: *mut WrenVM) -> *mut c_void;
	pub fn wrenSetUserData(vm: *mut WrenVM, userData: *mut c_void);
}
