//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use crate::bitmap::Bitmap;
use engine::{F2, f2, F3, f3, Index, MeshData, Primitive, Vertex};
use std::convert::TryFrom;

#[derive(Clone, Debug, PartialEq)]
pub struct MapPoint {
	/// Vertex in world coordinates.
	pub position: F3,
	/// Original colour in heightmap (0..1 for each channel).
	pub rgba: [f32; 4],
	/// Direction toward axis of cylinder.
	pub up: F3,
	/// Coordinates on cylinder (0..1 for each channel).
	pub uv: F2,
}

#[derive(Clone, Debug, PartialEq)]
pub struct MapCell<'a> {
	/// Point positions in heightmap's cache.
	pub indices: [usize; 3],
	/// Heightmap vertex information.
	pub points: [&'a MapPoint; 3],
}
impl<'a> MapCell<'a> {
	pub fn height(&self, u: f32, v: f32) -> f32 {
		Self::interpolate_f32([
			self.points[0].rgba[3],
			self.points[1].rgba[3],
			self.points[2].rgba[3],
		], u, v)
	}
	pub fn interpolate_f32(points: [f32; 3], u: f32, v: f32) -> f32 {
		debug_assert!(u >= 0.0);
		debug_assert!(v >= 0.0);
		debug_assert!(u+v <= 1.0);
		points[0]+(points[1]-points[0])*u+(points[2]-points[0])*v
	}
	pub fn interpolate_f2(points: [F2; 3], u: f32, v: f32) -> F2 {
		f2(
			Self::interpolate_f32([points[0].x, points[1].x, points[2].x], u, v),
			Self::interpolate_f32([points[0].y, points[1].y, points[2].y], u, v),
		)
	}
	pub fn interpolate_f3(points: [F3; 3], u: f32, v: f32) -> F3 {
		f3(
			Self::interpolate_f32([points[0].x, points[1].x, points[2].x], u, v),
			Self::interpolate_f32([points[0].y, points[1].y, points[2].y], u, v),
			Self::interpolate_f32([points[0].z, points[1].z, points[2].z], u, v),
		)
	}
	pub fn position(&self, u: f32, v: f32) -> F3 {
		Self::interpolate_f3([
			self.points[0].position,
			self.points[1].position,
			self.points[2].position,
		], u, v)
	}
}

pub struct ONeilHeightmap {
	height: usize,
	length: f32,
	radius: f32,
	vertices: Box<[MapPoint]>,
	width: usize,
}
impl ONeilHeightmap {
	/// Create new cylindrical terrain from a heightmap.
	/// The red channel is used for height data, zero height is considered sea.
	/// The green channel is used to colour variation.
	/// Blue and alpha are left for whatever other uses turn up.
	/// `radius` chooses the size of the cylinder in world coordinates.
	/// `height_scale` is used to scale heights, and should be 0..1.
	/// The cylinder will be centred on zero, and have it's axis along world x.
	pub fn new(heights: &Bitmap, radius: f32, height_scale: f32) -> Self {
		// Build cylinder with bitmap x/width mapped to the height (i.e. axis) of the
		// cylinder, and bitmap y/height mapped to the circumference.
		// We add vertices in the order of the values in the heightmap, left to
		// right, then top to bottom.
		// The cylinder is oriented with it's axis along world x, and is centred on
		// 0,0,0.
		//
		// The terrain around the circumference of the cylilnder is a grid, but rows
		// are skewed so that the quads become tessalated equalateral triangles.
		// This makes the interpolation between points of the heightmap look nicer,
		// and stops the choice of which diagonal to divide quads along from
		// appearing as a visual artefact.
		//   u →
		// v o---o---o---o even
		// ↓  \ / \ / \ / \
		//     o---o---o---o odd (offset →)
		//    / \ / \ / \ /
		//   o---o---o---o even
		//    \ / \ / \ / \
		//     o---o---o---o odd (offset →)
		// The ragged ends of rows are squashed in.
		//     u=0       u=1
		// v=0 o-o---o---o even
		// 0.1 |/ \ / \ /|
		//     o---o---o-o odd
		//     |\ / \ / \|
		//     o-o---o---o even
		// 0.9 |/ \ / \ /|
		// v=0 o---o---o-o odd

		debug_assert!(height_scale >= 0.0);
		debug_assert!(height_scale <= 1.0);
		let width = heights.width();
		let height = heights.height();
		assert!(width > 0);
		assert!(height > 0);
		assert!(width <= 256); // this already takes us into megabytes of cache!
		assert!(height <= 256);
		let triangle_height = std::f32::consts::PI*radius*2.0/height as f32;
		let triangle_width = 2.0*triangle_height/(60.0/180.0*std::f32::consts::PI).tan();
		let length = triangle_width*(width as f32-1.5);

		let u_step = length/(width as f32-1.5); // we loose half a step to squashing
		let v_step = std::f32::consts::PI*2.0/height as f32;
		let mut vertices = Vec::with_capacity(width*(height+1));
		let mut iter = heights.rgba.chunks_exact(4).cycle();
		let mut v = 0.0 as f32; // radians, 0.0 points down/y-
		let mut even = true;
		for _ in 0..height+1 { // generate a last row where the first is to avoid uv issues
			let (s, c) = v.sin_cos();
			let mut u = -length*0.5;
			for u_index in 0..width {
				let rgba = iter.next().unwrap();
				let rgba = [
					rgba[0] as f32/255.0,
					rgba[1] as f32/255.0,
					rgba[2] as f32/255.0,
					rgba[3] as f32/255.0,
				];
				let outset = (1.0-rgba[3]*height_scale)*radius;
				vertices.push(MapPoint {
					position: f3(u, -c*outset, -s*outset),
					rgba,
					up: f3(0.0, c, s),
					uv: f2((u+length*0.5)/length, v/std::f32::consts::PI/2.0),
				});
				if (even && u_index == 0) || (!even && u_index == width-2) {
					u += u_step*0.5;
				} else {
					u += u_step;
				}
			}
			v += v_step;
			even = !even;
		}

		Self {
			height,
			length,
			radius,
			vertices: vertices.into_boxed_slice(),
			width,
		}
	}
	/// Iterate over the triangles in the terrain (land and sea).
	pub fn for_triangles<F: FnMut(MapCell)>(&self, mut f: F) {
		let vertices = &self.vertices;
		let mut this_row = 0;
		let mut next_row = self.width;
		let mut even = true;
		for _ in 0..self.height {
			for _ in 0..self.width-1 {
				if even {
					//   u →
					// v +---+
					// ↓ |A /|
					//   |/ B|
					//   +---+
					f(MapCell{
						points: [&vertices[this_row], &vertices[next_row], &vertices[this_row+1]],
						indices: [this_row, next_row, this_row+1],
					});
					f(MapCell{
						points: [&vertices[next_row+1], &vertices[this_row+1], &vertices[next_row]],
						indices: [next_row+1, this_row+1, next_row],
					});
				} else {
					//   u →
					// v +---+
					// ↓ |\ C|
					//   |D \|
					//   +---+
					f(MapCell{
						points: [&vertices[this_row+1], &vertices[this_row], &vertices[next_row+1]],
						indices: [this_row+1, this_row, next_row+1],
					});
					f(MapCell{
						points: [&vertices[next_row], &vertices[next_row+1], &vertices[this_row]],
						indices: [next_row, next_row+1, this_row],
					});
				}
				this_row += 1;
				next_row += 1;
			}
			this_row += 1;
			next_row += 1;
			debug_assert!(this_row%self.width == 0);
			debug_assert!(next_row%self.width == 0);
			even = !even;
		}
		debug_assert!(this_row == self.vertices.len()-self.width);
		debug_assert!(next_row == self.vertices.len());
	}
	/// Return mesh of the land, end caps, and sea.
	pub fn to_mesh(&self) -> MeshData {
		// Vertex data.
		let mut indices = Vec::with_capacity(self.vertices.len());
		let mut vertices = Vec::with_capacity(self.vertices.len());
		// Map from heightmap point index to vertex data index.
		// This allow compute efficient deduplication of vertex data.
		let mut lookup = vec![None; self.vertices.len()];
		// Build mesh.
		self.for_triangles(|cell| for point in 0..3 {
			// Only add to vertex list if the vertex isn't already in it.
			if let Some(index) = lookup[cell.indices[point]] {
				indices.push(index);
			} else {
				let index = Index::try_from(vertices.len()).unwrap();
				vertices.push(Vertex {
					position: cell.points[point].position.to_array(),
					normal: cell.points[point].up.to_array(),
					texcoord: cell.points[point].uv.to_array(),
				});
				lookup[cell.indices[point]] = Some(index);
				indices.push(index);
			}
		});
		// Add centre vertex for end caps to make them easier to build.
		let left_cap = Index::try_from(vertices.len()).unwrap();
		vertices.push(Vertex {
			position: [-self.length*0.5, 0.0, 0.0],
			normal: [1.0, 0.0, 0.0],
			texcoord: [0.0, 1.0],
		});
		let right_cap = Index::try_from(vertices.len()).unwrap();
		vertices.push(Vertex {
			position: [self.length*0.5, 0.0, 0.0],
			normal: [-1.0, 0.0, 0.0],
			texcoord: [1.0, 1.0],
		});
		// Make sure vertices for end caps have been added.
		let mut left = 0;
		let mut right = self.width-1;
		for _ in 0..self.height+1 {
			debug_assert!(lookup[left].is_some());
			if lookup[left].is_none() {
				let vertex = &self.vertices[left];
				let index = Index::try_from(vertices.len()).unwrap();
				vertices.push(Vertex {
					position: vertex.position.to_array(),
					normal: [1.0, 0.0, 0.0],
					texcoord: [0.0, 1.0],
				});
				lookup[left] = Some(index);
			}
			debug_assert!(lookup[right].is_some());
			if lookup[right].is_none() {
				let vertex = &self.vertices[right];
				let index = Index::try_from(vertices.len()).unwrap();
				vertices.push(Vertex {
					position: vertex.position.to_array(),
					normal: [-1.0, 0.0, 0.0],
					texcoord: [1.0, 1.0],
				});
				lookup[right] = Some(index);
			}
			left += self.width;
			right += self.width;
			if left == self.vertices.len() {
				// Wrap around cylilnder.
				debug_assert!(right > self.vertices.len()); 
				left = 0;
				right = self.width-1;
			}
		}
		// Add end caps.
		let mut left = 0;
		let mut right = self.width-1;
		for _ in 0..self.height {
			let last_left = left;
			let last_right = right;
			left += self.width;
			right += self.width;
			indices.push(left_cap);
			indices.push(lookup[left].unwrap());
			indices.push(lookup[last_left].unwrap());
			indices.push(right_cap);
			indices.push(lookup[last_right].unwrap());
			indices.push(lookup[right].unwrap());
		}
		// Finalise meshes.
		let mut mesh = MeshData {
			indices: indices.into_boxed_slice(),
			primitive: Primitive::Triangles,
			vertices: vertices.into_boxed_slice(),
		};
		mesh.calculate_normals();
		mesh
	}
	pub fn to_sea_mesh(&self, height: f32, texcoord_scale: f32) -> MeshData {
		// Subdivide so that per-vertex fog will work.
		let mut vertices = Vec::with_capacity(self.width*self.height);
		let outset = self.radius*(1.0-height);
		let u_step = self.length/(self.width-1) as f32;
		let v_step = std::f32::consts::PI*2.0/self.height as f32;
		let mut v = 0.0 as f32; // radians, 0.0 points down/y-
		for _ in 0..self.height+1 {
			let (s, c) = v.sin_cos();
			let mut u = -self.length*0.5;
			for _ in 0..self.width {
				vertices.push(Vertex {
					position: [u, -c*outset, -s*outset],
					normal: [0.0, c, s],
					texcoord: [
						(u+self.length*0.5)*texcoord_scale/self.length,
						v*texcoord_scale/std::f32::consts::PI/2.0,
					],
				});
				u += u_step;
			}
			debug_assert!(u > self.length*0.5);
			v += v_step;
		}
		let mut indices = Vec::with_capacity(self.width*self.height);
		let mut this_row = 0 as usize;
		let mut next_row = self.width;
		for _ in 0..self.height {
			for _ in 0..self.width-1 {
				indices.push(Index::try_from(this_row).unwrap());
				indices.push(Index::try_from(next_row).unwrap());
				indices.push(Index::try_from(this_row+1).unwrap());
				indices.push(Index::try_from(this_row+1).unwrap());
				indices.push(Index::try_from(next_row).unwrap());
				indices.push(Index::try_from(next_row+1).unwrap());
				this_row += 1;
				next_row += 1;
			}
			this_row += 1;
			next_row += 1;
			debug_assert!(this_row%self.width == 0);
			debug_assert!(next_row%self.width == 0);
		}
		MeshData {
			indices: indices.into_boxed_slice(),
			primitive: Primitive::Triangles,
			vertices: vertices.into_boxed_slice(),
		}
	}
	/*
	pub fn to_sea_mesh(&self, height: f32, texcoord_scale: f32) -> MeshData {
		let mut vertices = Vec::with_capacity((self.height-1)*2);
		let mut indices = Vec::with_capacity(vertices.len()*2);
		let radius = self.radius*(1.0-height);
		let mut last_left = Index::try_from(vertices.len()).unwrap();
		vertices.push(Vertex {
			position: [-self.length*0.5, -radius, 0.0],
			normal: [0.0, 1.0, 0.0],
			texcoord: [0.0, 0.0],
		});
		let mut last_right = Index::try_from(vertices.len()).unwrap();
		vertices.push(Vertex {
			position: [self.length*0.5, -radius, 0.0],
			normal: [0.0, 1.0, 0.0],
			texcoord: [texcoord_scale, 0.0],
		});
		let v_step = std::f32::consts::PI*2.0/self.height as f32;
		let mut v = v_step; // radians, 0.0 points down/y-
		for _ in 0..self.height {
			let (s, c) = v.sin_cos();
			let left = Index::try_from(vertices.len()).unwrap();
			vertices.push(Vertex {
				position: [-self.length*0.5, -c*radius, -s*radius],
				normal: [0.0, c, s],
				texcoord: [0.0, v*texcoord_scale/std::f32::consts::PI/2.0],
			});
			let right = Index::try_from(vertices.len()).unwrap();
			vertices.push(Vertex {
				position: [self.length*0.5, -c*radius, -s*radius],
				normal: [0.0, c, s],
				texcoord: [texcoord_scale, v*texcoord_scale/std::f32::consts::PI/2.0],
			});
			indices.push(last_left);
			indices.push(left);
			indices.push(last_right);
			indices.push(last_right);
			indices.push(left);
			indices.push(right);
			last_left = left;
			last_right = right;
			v += v_step;
		}
		MeshData {
			indices: indices.into_boxed_slice(),
			primitive: Primitive::Triangles,
			vertices: vertices.into_boxed_slice(),
		}
	}
	*/
}
