//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use engine::*;

#[derive(Clone, Copy, Debug, Eq, PartialEq)]
#[repr(u32)]
pub enum LogicalAxis {
	Pitch,
	Roll,
	Throttle,
	Yaw,
}
pub const LOGIC_AXIS_COUNT: usize = LogicalAxis::Yaw as usize+1;

pub struct Seat<T> {
	pub axes: [f32; LOGIC_AXIS_COUNT],
	/// Which controller drives this seat.
	/// `None` means keyboard.
	/// If the seat is in the `lost` list then this value is meaningless.
	controller_id: Option<u32>,
	/// Game data attached to this seat.
	pub data: T,
}
impl<T> Seat<T> {
	pub fn new(controller_id: Option<u32>, data: T) -> Self {
		Self {
			axes: [0.0; LOGIC_AXIS_COUNT],
			controller_id,
			data,
		}
	}
	pub fn axis(&self, axis: LogicalAxis) -> f32 { self.axes[axis as usize] }
}

pub struct Input<T> {
	/// Maximum number of players.
	capacity: usize,
	/// Seats whos controller has been disconnected.
	/// This is a queue for new controllers to be handed out to.
	lost: Vec<Seat<T>>,
	/// Active seats.
	seats: Vec<Seat<T>>,
}
impl<T> Input<T> {
	pub fn new(capacity: usize) -> Self {
		Self {
			capacity,
			lost: Vec::new(),
			seats: Vec::new(),
		}
	}
	pub fn event<AddFn: FnMut() -> T, RemoveFn: FnMut(T)>(&mut self, event: &Event, mut add_fn: AddFn, mut remove_fn: RemoveFn) {
		match event {
			&Event::ControllerAdded{id} => {
				debug_assert!(self.seats.iter().all(|seat| seat.controller_id != Some(id)));
				if let Some(mut seat) = self.lost.pop() {
					seat.controller_id = Some(id);
					self.seats.push(seat);
				}
			},
			&Event::ControllerAxis{id, axis, value} => match axis {
				ControllerAxis::LeftX => self.set_axis(Some(id), LogicalAxis::Roll, value),
				ControllerAxis::LeftY => self.set_axis(Some(id), LogicalAxis::Pitch, -value),
				ControllerAxis::LeftTrigger => self.set_axis(Some(id), LogicalAxis::Throttle, -value),
				ControllerAxis::RightTrigger => self.set_axis(Some(id), LogicalAxis::Throttle, value),
				_ => (),
			},
			&Event::ControllerButton{id, button, down} => if button == ControllerButton::Start && self.lost.len()+self.seats.len() < self.capacity && self.seats.iter().all(|seat| seat.controller_id != Some(id)) {
				let data = add_fn();
				self.seats.push(Seat::new(Some(id), data));
			} else if button == ControllerButton::Back {
				self.remove_seats(Some(id), &mut remove_fn);
			} else {
				let value = if down { 1.0 } else { -1.0 };
				match button {
					ControllerButton::LeftShoulder => self.adjust_axis(Some(id), LogicalAxis::Yaw, -value),
					ControllerButton::RightShoulder => self.adjust_axis(Some(id), LogicalAxis::Yaw, value),
					_ => (),
				}
			},
			&Event::ControllerLost{id} => {
				let mut index = 0;
				while index < self.seats.len() {
					if self.seats[index].controller_id == Some(id) {
						self.lost.push(self.seats.swap_remove(index));
					} else {
						index += 1;
					}
				}
			},
			&Event::Key{key, down, ..} => if key == Key::Return && self.lost.len()+self.seats.len() < self.capacity && self.seats.iter().all(|seat| seat.controller_id != None) {
				let data = add_fn();
				self.seats.push(Seat::new(None, data));
			} else if key == Key::Escape {
				self.remove_seats(None, &mut remove_fn);
			} else {
				let value = if down { 1.0 } else { -1.0 };
				match key {
					Key::A => self.adjust_axis(None, LogicalAxis::Roll, -value),
					Key::D => self.adjust_axis(None, LogicalAxis::Roll, value),
					Key::S => self.adjust_axis(None, LogicalAxis::Pitch, -value),
					Key::W => self.adjust_axis(None, LogicalAxis::Pitch, value),
					Key::Q => self.adjust_axis(None, LogicalAxis::Yaw, -value),
					Key::E => self.adjust_axis(None, LogicalAxis::Yaw, value),
					Key::LCtrl => self.adjust_axis(None, LogicalAxis::Throttle, -value),
					Key::LShift => self.adjust_axis(None, LogicalAxis::Throttle, value),
					_ => (),
				}
			},
			_ => (),
		}
	}
	pub fn seats(&self) -> &[Seat<T>] { &self.seats }
	pub fn seats_mut(&mut self) -> &mut [Seat<T>] { &mut self.seats }

	fn remove_seats<F: FnMut(T)>(&mut self, controller_id: Option<u32>, mut remove_fn: F) {
		let mut index = 0;
		while index < self.seats.len() {
			if self.seats[index].controller_id == controller_id {
				let seat = self.seats.swap_remove(index);
				remove_fn(seat.data);
			} else {
				index += 1;
			}
		}
	}
	fn set_axis(&mut self, controller_id: Option<u32>, axis: LogicalAxis, value: f32) {
		for seat in self.seats.iter_mut() {
			if seat.controller_id == controller_id {
				seat.axes[axis as usize] = value;
			}
		}
	}
	fn adjust_axis(&mut self, controller_id: Option<u32>, axis: LogicalAxis, delta: f32) {
		for seat in self.seats.iter_mut() {
			if seat.controller_id == controller_id {
				seat.axes[axis as usize] = (seat.axes[axis as usize]+delta).max(-1.0).min(1.0);
			}
		}
	}
}
