//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use engine::{F3, f3, Rotation};
use entities::Handle;
use ffi_wren::*;
use game::{GameData, SeatData};
use std::ffi::{CStr, CString};
use std::mem::MaybeUninit;
use std::os::raw::{c_char, c_int, c_void};

pub struct Scripting {
	data: Box<Data>,
	vm: *mut WrenVM,
}
impl Scripting {
	pub fn new() -> Self { unsafe {
		// Setup VM.
		let mut config = MaybeUninit::<WrenConfiguration>::uninit();
		wrenInitConfiguration(config.as_mut_ptr());
		let mut config = config.assume_init();
		config.writeFn = write_callback;
		config.errorFn = error_callback;
		config.bindForeignMethodFn = bind_method_callback;
		config.bindForeignClassFn = bind_class_callback;
		let vm = wrenNewVM(&mut config);
		debug_assert!(!vm.is_null());

		// Load builtin functions.
		let module = CString::new("main").unwrap();
		let code = CString::new(include_str!("scripting.wren")).unwrap();
		let result = wrenInterpret(vm, module.as_ptr(), code.as_ptr());
		assert!(result == WrenInterpretResult::Success);

		// Get global handles.
		let signature_call0 = CString::new("call()").unwrap();
		let signature_transfer = CString::new("transfer()").unwrap();
		let mut data = Box::new(Data {
			class_entity: Self::get_class(vm, "main", "Entity"),
			class_vector3: Self::get_class(vm, "main", "Vector3"),
			delayed: Vec::new(),
			game: std::ptr::null_mut(),
			method_call0: wrenMakeCallHandle(vm, signature_call0.as_ptr()),
			method_transfer: wrenMakeCallHandle(vm, signature_transfer.as_ptr()),
		});
		wrenSetUserData(vm, &mut *data as *mut Data as *mut c_void);

		Self{data, vm}
	} }
	pub fn run(&mut self, code: &str, game: &mut GameData) { unsafe {
		self.data.game = game;
		let module = CString::new("main").unwrap();
		let code = CString::new(code).unwrap();
		match wrenInterpret(self.vm, module.as_ptr(), code.as_ptr()) {
			WrenInterpretResult::CompileError => panic!("script compilation error"),
			WrenInterpretResult::RuntimeError => panic!("script runtime error"),
			WrenInterpretResult::Success => (),
		}
		self.data.game = std::ptr::null_mut();
	} }
	pub fn tick(&mut self, game: &mut GameData) { unsafe {
		self.data.game = game;
		let stop = self.data.delayed.iter().position(|delayed| delayed.frames > 0)
			.unwrap_or(self.data.delayed.len());
		for mut delayed in self.data.delayed.drain(0..stop) {
			// TODO: share code with `call_closure()`, borrowing is making a mess
			wrenEnsureSlots(self.vm, 1);
			wrenSetSlotHandle(self.vm, 0, delayed.fiber);
			wrenCall(self.vm, self.data.method_transfer);
			wrenReleaseHandle(self.vm, delayed.fiber);
			delayed.fiber = std::ptr::null_mut();
		}
		if let Some(first) = self.data.delayed.first_mut() {
			first.frames -= 1;
		}
		self.data.game = std::ptr::null_mut();
	} }
	pub unsafe fn call_closure(&mut self, game: &mut GameData, closure_receiver: *mut WrenHandle) {
		debug_assert!(!closure_receiver.is_null());
		self.data.game = game;
		wrenEnsureSlots(self.vm, 1);
		wrenSetSlotHandle(self.vm, 0, closure_receiver);
		wrenCall(self.vm, self.data.method_call0);
		self.data.game = std::ptr::null_mut();
	}
	pub unsafe fn vm(&self) -> *mut WrenVM { self.vm }

	unsafe fn get_class(vm: *mut WrenVM, module: &str, name: &str) -> *mut WrenHandle {
		let module = CString::new(module).unwrap();
		let name = CString::new(name).unwrap();
		wrenEnsureSlots(vm, 1);
		assert!(wrenHasVariable(vm, module.as_ptr(), name.as_ptr()));
		wrenGetVariable(vm, module.as_ptr(), name.as_ptr(), 0);
		let handle = wrenGetSlotHandle(vm, 0);
		debug_assert!(!handle.is_null());
		handle
	}
}
impl Drop for Scripting {
	fn drop(&mut self) { unsafe {
		for delayed in self.data.delayed.iter_mut() {
			wrenReleaseHandle(self.vm, delayed.fiber);
			delayed.fiber = std::ptr::null_mut();
		}
		wrenReleaseHandle(self.vm, self.data.class_entity);
		wrenReleaseHandle(self.vm, self.data.class_vector3);
		wrenReleaseHandle(self.vm, self.data.method_call0);
		wrenReleaseHandle(self.vm, self.data.method_transfer);
		wrenFreeVM(self.vm);
	} }
}

struct Data {
	class_entity: *mut WrenHandle,
	class_vector3: *mut WrenHandle,
	delayed: Vec<Delayed>,
	game: *mut GameData, // only valid during `Scripting::run()` and `Scripting::tick()`
	method_call0: *mut WrenHandle,
	method_transfer: *mut WrenHandle,
}

struct Delayed {
	fiber: *mut WrenHandle,
	frames: u32,
}
#[cfg(debug_assertions)]
impl Drop for Delayed {
	fn drop(&mut self) {
		debug_assert!(self.fiber.is_null());
	}
}

extern "C" fn bind_method_callback(_vm: *mut WrenVM, _module: *const c_char, class: *const c_char, is_static: bool, signature: *const c_char) -> Option<WrenForeignMethodFn> { unsafe {
	let class = CStr::from_ptr(class).to_str().unwrap();
	let signature = CStr::from_ptr(signature).to_str().unwrap();
	match (class, is_static, signature) {
		("Game", true, "count_players()") => Some(api::game_count_players),
		("Game", true, "player_position(_)") => Some(api::game_player_position),
		("Game", true, "set_player_position(_,_)") => Some(api::game_set_player_position),
		("Game", true, "player_velocity(_)") => Some(api::game_player_velocity),
		("Game", true, "set_player_velocity(_,_)") => Some(api::game_set_player_velocity),
		("Game", true, "player_direction(_)") => Some(api::game_player_direction),
		("Game", true, "set_player_direction(_,_)") => Some(api::game_set_player_direction),
		("Game", true, "add(_,_,_,_)") => Some(api::game_add),
		// ("Game", true, "delay(_,_)") => Some(api::game_delay),
		("Game", true, "get_x(_)") => Some(api::game_get_x),
		("Game", true, "get_y(_)") => Some(api::game_get_y),
		("Game", true, "get_z(_)") => Some(api::game_get_z),
		("Game", true, "random()") => Some(api::game_random),
		("Game", true, "ray(_,_)") => Some(api::game_ray),
		("Game", true, "ray(_,_,_)") => Some(api::game_ray),
		("Game", true, "wait_internal(_,_)") => Some(api::game_wait_internal),
		("Entity", false, "remove()") => Some(api::entity_remove),
		("Entity", false, "direction") => Some(api::entity_get_direction),
		("Entity", false, "direction=(_)") => Some(api::entity_set_direction),
		("Entity", false, "mesh") => Some(api::entity_get_mesh),
		("Entity", false, "mesh=(_)") => Some(api::entity_set_mesh),
		("Entity", false, "position") => Some(api::entity_get_position),
		("Entity", false, "position=(_)") => Some(api::entity_set_position),
		("Entity", false, "scale") => Some(api::entity_get_scale),
		("Entity", false, "scale=(_)") => Some(api::entity_set_scale),
		("Entity", false, "tick=(_)") => Some(api::entity_set_tick),
		("Vector3", true, "new()") => Some(api::vector3_new_0),
		("Vector3", true, "new(_)") => Some(api::vector3_new_1),
		("Vector3", true, "new(_,_,_)") => Some(api::vector3_new_3),
		("Vector3", false, "x") => Some(api::vector3_get_x),
		("Vector3", false, "y") => Some(api::vector3_get_y),
		("Vector3", false, "z") => Some(api::vector3_get_z),
		("Vector3", false, "x=(_)") => Some(api::vector3_set_x),
		("Vector3", false, "y=(_)") => Some(api::vector3_set_y),
		("Vector3", false, "z=(_)") => Some(api::vector3_set_z),
		("Vector3", false, "-") => Some(api::vector3_negate),
		("Vector3", false, "+(_)") => Some(api::vector3_add),
		("Vector3", false, "-(_)") => Some(api::vector3_subtract),
		("Vector3", false, "*(_)") => Some(api::vector3_multiply),
		("Vector3", false, "/(_)") => Some(api::vector3_divide),
		("Vector3", false, "cross(_)") => Some(api::vector3_cross),
		("Vector3", false, "dot(_)") => Some(api::vector3_dot),
		("Vector3", false, "length()") => Some(api::vector3_length),
		("Vector3", false, "length2()") => Some(api::vector3_length2),
		("Vector3", false, "lerp(_,_)") => Some(api::vector3_lerp),
		("Vector3", false, "unitZero()") => Some(api::vector3_unit_zero),
		("Vector3", false, "toString()") => Some(api::vector3_to_string),
		_ => panic!("script: unknown foreign method {} {}.{}",
			if is_static { "static" } else { "" }, class, signature),
	}
} }

extern "C" fn bind_class_callback(_vm: *mut WrenVM, _module: *const c_char, class: *const c_char) -> WrenForeignClassMethods { unsafe {
	let class = CStr::from_ptr(class).to_str().unwrap();
	match class {
		"Entity" => WrenForeignClassMethods {
			allocate: api::entity_allocate,
			finalize: Some(api::entity_finalise),
		},
		"Vector3" => WrenForeignClassMethods {
			allocate: api::vector3_allocate,
			finalize: None,
		},
		_ => panic!("script: unknown foreign class"),
	}
} }

extern "C" fn write_callback(_vm: *mut WrenVM, text: *const c_char) { unsafe {
	if cfg!(debug_assertions) {
		let text = CStr::from_ptr(text).to_str().unwrap_or("<invalid UTF8>");
		println!("script: {}", text);
	}
} }

extern "C" fn error_callback(_vm: *mut WrenVM, error_type: WrenErrorType, module: *const c_char, line: c_int, message: *const c_char) { unsafe {
	if cfg!(debug_assertions) {
		let message = CStr::from_ptr(message).to_str().unwrap_or("<invalid UTF8>");
		let module = if module.is_null() { "<NULL>" } else {
			CStr::from_ptr(module).to_str().unwrap_or("<invalid UTF8>")
		};
		match error_type {
			WrenErrorType::Compile => eprintln!("script: {}:{}: error: {}", module, line, message),
			WrenErrorType::Runtime => eprintln!("script: error: {}", message),
			WrenErrorType::StackTrace => eprintln!("script: {}:{}: in {}", module, line, message),
		}
	}
	panic!("script error");
} }

mod api {
	use super::*;

	unsafe fn game_get(vm: *mut WrenVM) -> &'static mut GameData {
		let data = wrenGetUserData(vm) as *const Data;
		debug_assert!(!data.is_null());
		let data = &*data;
		debug_assert!(!data.game.is_null());
		&mut *data.game
	}
	unsafe fn seat_get_index(vm: *mut WrenVM, slot: c_int) -> &'static mut SeatData {
		let game = game_get(vm);
		let index = wrenGetSlotDouble(vm, slot);
		debug_assert!(index >= 0.0);
		debug_assert!((index as usize) < game.input.seats().len());
		&mut game.input.seats_mut()[index as usize].data
	}
	pub extern "C" fn game_count_players(vm: *mut WrenVM) { unsafe {
		let game = game_get(vm);
		wrenSetSlotDouble(vm, 0, game.input.seats().len() as f64);
	} }
	pub extern "C" fn game_player_position(vm: *mut WrenVM) { unsafe {
		let seat = seat_get_index(vm, 1);
		vector3_new_real(vm, seat.aircraft.centre);
	} }
	pub extern "C" fn game_set_player_position(vm: *mut WrenVM) { unsafe {
		let seat = seat_get_index(vm, 1);
		seat.aircraft.centre = *vector3_get(vm, 2);
	} }
	pub extern "C" fn game_player_velocity(vm: *mut WrenVM) { unsafe {
		let seat = seat_get_index(vm, 1);
		vector3_new_real(vm, seat.aircraft.velocity);
	} }
	pub extern "C" fn game_set_player_velocity(vm: *mut WrenVM) { unsafe {
		let seat = seat_get_index(vm, 1);
		seat.aircraft.velocity = *vector3_get(vm, 2);
	} }
	pub extern "C" fn game_player_direction(vm: *mut WrenVM) { unsafe {
		let seat = seat_get_index(vm, 1);
		let direction = seat.aircraft.rotation.apply(f3(0.0, 0.0, 1.0));
		vector3_new_real(vm, direction);
	} }
	pub extern "C" fn game_set_player_direction(vm: *mut WrenVM) { unsafe {
		let seat = seat_get_index(vm, 1);
		let direction = vector3_get(vm, 2);
		seat.aircraft.rotation = Rotation::look(direction.unit_zero(), f3(0.0, 1.0, 0.0));
	} }
	pub extern "C" fn game_add(vm: *mut WrenVM) { unsafe {
		let game = game_get(vm);
		let mesh = wrenGetSlotDouble(vm, 1) as usize;
		debug_assert!(mesh < 7); // TODO: better way to get number of built in meshes
		let scale = vector3_get(vm, 2);
		let direction = vector3_get(vm, 3).unit().unwrap_or(f3(0.0, 0.0, 1.0));
		let translate = vector3_get(vm, 4);
		let up = f3(0.0, -translate.y, -translate.z).unit().unwrap_or(f3(0.0, 1.0, 0.0));
		let rotate = Rotation::look(direction, up);
		let handle = game.entities.push(mesh, *scale, rotate, *translate);
		entity_new_real(vm, handle);
	} }
	/*
	TODO: remove
	pub extern "C" fn game_delay(vm: *mut WrenVM) { unsafe {
		let data = wrenGetUserData(vm) as *mut Data;
		debug_assert!(!data.is_null());
		let data = &mut *data;
		let delay = wrenGetSlotDouble(vm, 1) as f32;
		debug_assert!(delay >= 0.0);
		let receiver = wrenGetSlotHandle(vm, 2);
		let mut delay = ((delay*engine::Window::FPS).round() as u32).max(1);
		let mut index = 0;
		while index < data.delayed.len() {
			if delay < data.delayed[index].delay {
				data.delayed[index].delay -= delay;
				data.delayed.insert(index, Delayed{delay, receiver});
				return;
			}
			delay -= data.delayed[index].delay;
			index += 1;
		}
		data.delayed.push(Delayed{delay, receiver});
	} }
	*/
	pub extern "C" fn game_get_x(vm: *mut WrenVM) { unsafe {
		let _position = vector3_get(vm, 1);
		vector3_new_real(vm, f3(1.0, 0.0, 0.0));
	} }
	pub extern "C" fn game_get_y(vm: *mut WrenVM) { unsafe {
		let position = vector3_get(vm, 1);
		vector3_new_real(vm, f3(0.0, -position.y, -position.z).unit().unwrap_or(f3(0.0, 1.0, 0.0)));
	} }
	pub extern "C" fn game_get_z(vm: *mut WrenVM) { unsafe {
		let position = vector3_get(vm, 1);
		vector3_new_real(vm, f3(0.0, position.z, -position.y).unit().unwrap_or(f3(0.0, 0.0, 1.0)));
	} }
	pub extern "C" fn game_random(vm: *mut WrenVM) { unsafe {
		let game = game_get(vm);
		wrenSetSlotDouble(vm, 0, game.rng.f32() as f64);
	} }
	pub extern "C" fn game_ray(vm: *mut WrenVM) { unsafe {
		let game = game_get(vm);
		let start = vector3_get(vm, 1);
		let direction = vector3_get(vm, 2);
		let slot_count = wrenGetSlotCount(vm);
		debug_assert!(slot_count == 3 || slot_count == 4);
		let range = if slot_count == 3 { None } else {
			Some(wrenGetSlotDouble(vm, 3) as f32)
		};
		if let Some(hit) = game.floor_collision.ray(*start, *direction, range, None) {
			vector3_new_real(vm, hit.point());
		} else {
			wrenSetSlotNull(vm, 0);
		}
	} }
	pub extern "C" fn game_wait_internal(vm: *mut WrenVM) { unsafe {
		let data = wrenGetUserData(vm) as *mut Data;
		debug_assert!(!data.is_null());
		let data = &mut *data;
		let seconds = wrenGetSlotDouble(vm, 1) as f32;
		let fiber = wrenGetSlotHandle(vm, 2);
		debug_assert!(seconds >= 0.0);
		let mut frames = ((seconds*engine::Window::FPS).round() as u32).max(1);
		let mut index = 0;
		while index < data.delayed.len() {
			if frames < data.delayed[index].frames {
				data.delayed[index].frames -= frames;
				data.delayed.insert(index, Delayed{fiber, frames});
				return;
			}
			frames -= data.delayed[index].frames;
			index += 1;
		}
		data.delayed.push(Delayed{fiber, frames});
	} }

	unsafe fn entity_new_real(vm: *mut WrenVM, handle: Handle) {
		let data = wrenGetUserData(vm) as *const Data;
		debug_assert!(!data.is_null());
		let data = &*data;
		wrenEnsureSlots(vm, 1);
		wrenSetSlotHandle(vm, 0, data.class_entity);
		let allocation = wrenSetSlotNewForeign(vm, 0, 0, std::mem::size_of::<Handle>()) as *mut Handle;
		debug_assert!(!allocation.is_null());
		std::ptr::write(allocation, handle);
	}
	unsafe fn entity_get(vm: *mut WrenVM, slot: c_int) -> &'static mut Handle {
		let allocation = wrenGetSlotForeign(vm, slot) as *mut Handle;
		debug_assert!(!allocation.is_null());
		&mut *allocation
	}
	pub extern "C" fn entity_allocate(_vm: *mut WrenVM) { panic!("script: use Game.add() instead") }
	pub extern "C" fn entity_finalise(allocation: *mut c_void) { unsafe {
		debug_assert!(!allocation.is_null());
		std::ptr::drop_in_place(allocation as *mut Handle);
	} }
	pub extern "C" fn entity_remove(vm: *mut WrenVM) { unsafe {
		let game = game_get(vm);
		let handle = entity_get(vm, 0);
		game.entities.remove(vm, handle);
	} }
	pub extern "C" fn entity_get_direction(vm: *mut WrenVM) { unsafe {
		let handle = entity_get(vm, 0);
		let direction = if let Some(entity) = handle.upgrade() {
			entity.borrow().next_rotate.apply(f3(0.0, 0.0, 1.0))
		} else {
			f3(0.0, 0.0, 1.0)
		};
		vector3_new_real(vm, direction);
	} }
	pub extern "C" fn entity_set_direction(vm: *mut WrenVM) { unsafe {
		let handle = entity_get(vm, 0);
		let direction = vector3_get(vm, 1).unit().unwrap_or(f3(0.0, 0.0, 1.0));
		if let Some(entity) = handle.upgrade() {
			let mut entity = entity.borrow_mut();
			let up = f3(0.0, -entity.next_translate.y, -entity.next_translate.z)
				.unit().unwrap_or(f3(0.0, 1.0, 0.0));
			entity.next_rotate = Rotation::look(direction, up);
		}
	} }
	pub extern "C" fn entity_get_mesh(vm: *mut WrenVM) { unsafe {
		let handle = entity_get(vm, 0);
		let mesh = if let Some(entity) = handle.upgrade() {
			entity.borrow().mesh
		} else {
			0
		};
		wrenSetSlotDouble(vm, 0, mesh as f64);
	} }
	pub extern "C" fn entity_set_mesh(vm: *mut WrenVM) { unsafe {
		let handle = entity_get(vm, 0);
		let mesh = wrenGetSlotDouble(vm, 1) as usize;
		debug_assert!(mesh < 7); // TODO: better way to get number of built in meshes
		if let Some(entity) = handle.upgrade() {
			entity.borrow_mut().mesh = mesh;
		}
	} }
	pub extern "C" fn entity_get_position(vm: *mut WrenVM) { unsafe {
		let handle = entity_get(vm, 0);
		let position = if let Some(entity) = handle.upgrade() {
			entity.borrow().next_translate
		} else {
			f3(0.0, 0.0, 0.0)
		};
		vector3_new_real(vm, position);
	} }
	pub extern "C" fn entity_set_position(vm: *mut WrenVM) { unsafe {
		let handle = entity_get(vm, 0);
		let position = vector3_get(vm, 1);
		if let Some(entity) = handle.upgrade() {
			entity.borrow_mut().next_translate = *position;
		}
	} }
	pub extern "C" fn entity_get_scale(vm: *mut WrenVM) { unsafe {
		let handle = entity_get(vm, 0);
		let scale = if let Some(entity) = handle.upgrade() {
			entity.borrow().next_translate
		} else {
			f3(0.0, 0.0, 0.0)
		};
		vector3_new_real(vm, scale);
	} }
	pub extern "C" fn entity_set_scale(vm: *mut WrenVM) { unsafe {
		let handle = entity_get(vm, 0);
		let scale = vector3_get(vm, 1);
		if let Some(entity) = handle.upgrade() {
			entity.borrow_mut().scale = *scale;
		}
	} }
	pub extern "C" fn entity_set_tick(vm: *mut WrenVM) { unsafe {
		let game = game_get(vm);
		let handle = entity_get(vm, 0);
		let r#type = wrenGetSlotType(vm, 1);
		let tick = if r#type == WrenType::Null {
			std::ptr::null_mut()
		} else {
			debug_assert!(r#type == WrenType::Unknown);
			wrenGetSlotHandle(vm, 1)
		};
		game.entities.bind(vm, handle, tick);
	} }

	unsafe fn vector3_new_real(vm: *mut WrenVM, value: F3) {
		let data = wrenGetUserData(vm) as *const Data;
		debug_assert!(!data.is_null());
		let data = &*data;
		wrenEnsureSlots(vm, 1);
		wrenSetSlotHandle(vm, 0, data.class_vector3);
		let allocation = wrenSetSlotNewForeign(vm, 0, 0, std::mem::size_of::<F3>()) as *mut F3;
		debug_assert!(!allocation.is_null());
		std::ptr::write(allocation, value);
	}
	unsafe fn vector3_get(vm: *mut WrenVM, slot: c_int) -> &'static mut F3 {
		let allocation = wrenGetSlotForeign(vm, slot) as *mut F3;
		debug_assert!(!allocation.is_null());
		&mut *allocation
	}
	pub extern "C" fn vector3_allocate(_vm: *mut WrenVM) { panic!("script: use Vector3.new() instead") }
	pub extern "C" fn vector3_new_0(vm: *mut WrenVM) { unsafe {
		debug_assert!(wrenGetSlotCount(vm) == 1);
		vector3_new_real(vm, f3(0.0, 0.0, 0.0));
	} }
	pub extern "C" fn vector3_new_1(vm: *mut WrenVM) { unsafe {
		debug_assert!(wrenGetSlotCount(vm) == 2);
		let value = wrenGetSlotDouble(vm, 1) as f32;
		vector3_new_real(vm, f3(value, value, value));
	} }
	pub extern "C" fn vector3_new_3(vm: *mut WrenVM) { unsafe {
		debug_assert!(wrenGetSlotCount(vm) == 4);
		let x = wrenGetSlotDouble(vm, 1) as f32;
		let y = wrenGetSlotDouble(vm, 2) as f32;
		let z = wrenGetSlotDouble(vm, 3) as f32;
		vector3_new_real(vm, f3(x, y, z));
	} }
	pub extern "C" fn vector3_get_x(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		wrenSetSlotDouble(vm, 0, this.x as f64)
	} }
	pub extern "C" fn vector3_get_y(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		wrenSetSlotDouble(vm, 0, this.y as f64)
	} }
	pub extern "C" fn vector3_get_z(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		wrenSetSlotDouble(vm, 0, this.z as f64)
	} }
	pub extern "C" fn vector3_set_x(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		this.x = wrenGetSlotDouble(vm, 1) as f32;
	} }
	pub extern "C" fn vector3_set_y(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		this.y = wrenGetSlotDouble(vm, 1) as f32;
	} }
	pub extern "C" fn vector3_set_z(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		this.z = wrenGetSlotDouble(vm, 1) as f32;
	} }
	pub extern "C" fn vector3_negate(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		vector3_new_real(vm, -*this);
	} }
	pub extern "C" fn vector3_add(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		let other = vector3_get(vm, 1);
		vector3_new_real(vm, *this+*other);
	} }
	pub extern "C" fn vector3_subtract(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		let other = vector3_get(vm, 1);
		vector3_new_real(vm, *this-*other);
	} }
	pub extern "C" fn vector3_multiply(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		let value = if wrenGetSlotType(vm, 1) == WrenType::Number {
			*this*wrenGetSlotDouble(vm, 1) as f32
		} else {
			*this**vector3_get(vm, 1)
		};
		vector3_new_real(vm, value);
	} }
	pub extern "C" fn vector3_divide(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		let value = if wrenGetSlotType(vm, 1) == WrenType::Number {
			*this/wrenGetSlotDouble(vm, 1) as f32
		} else {
			*this / *vector3_get(vm, 1)
		};
		vector3_new_real(vm, value);
	} }
	pub extern "C" fn vector3_cross(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		let other = vector3_get(vm, 1);
		vector3_new_real(vm, this.cross(*other));
	} }
	pub extern "C" fn vector3_dot(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		let other = vector3_get(vm, 1);
		wrenSetSlotDouble(vm, 1, this.dot(*other) as f64);
	} }
	pub extern "C" fn vector3_length(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		wrenSetSlotDouble(vm, 0, this.length() as f64);
	} }
	pub extern "C" fn vector3_length2(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		wrenSetSlotDouble(vm, 0, this.length2() as f64);
	} }
	pub extern "C" fn vector3_lerp(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		let other = vector3_get(vm, 1);
		let factor = wrenGetSlotDouble(vm, 2) as f32;
		vector3_new_real(vm, this.lerp(*other, factor));
	} }
	pub extern "C" fn vector3_unit_zero(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		vector3_new_real(vm, this.unit_zero());
	} }
	pub extern "C" fn vector3_to_string(vm: *mut WrenVM) { unsafe {
		let this = vector3_get(vm, 0);
		let text = format!("({},{},{})", this.x, this.y, this.z);
		let bytes = text.as_bytes();
		wrenSetSlotBytes(vm, 0, bytes.as_ptr() as *const c_char, bytes.len());
	} }
}
