//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use engine::{
	ActiveShader, DynamicMesh, F3, f3, Index, MeshData,
	Primitive, Rotation, StaticMesh, Window,
};
use engine::ffi_gles::*;
use std::convert::TryFrom;

/// Flat quad useful for drawing over the whole screen.
pub struct Fill {
	mesh: StaticMesh<engine::Vertex>,
}
impl Fill {
	pub fn new(window: &mut Window) -> Self {
		let vertices = [
			engine::Vertex{position: [-1.0, -1.0, 1.0], normal: [0.0; 3], texcoord: [0.0; 2]},
			engine::Vertex{position: [1.0, -1.0, 1.0], normal: [0.0; 3], texcoord: [0.0; 2]},
			engine::Vertex{position: [-1.0, 1.0, 1.0], normal: [0.0; 3], texcoord: [0.0; 2]},
			engine::Vertex{position: [1.0, 1.0, 1.0], normal: [0.0; 3], texcoord: [0.0; 2]},
		];
		let indices = [0, 3, 2, 0, 1, 3];
		Self {
			mesh: StaticMesh::new(window, Primitive::Triangles, &indices, &vertices),
		}
	}
}

/// Shadow for one mesh draw by calculating a shadow volume.
/// A shadow volume is a mesh that contains everywhere that is within the
/// shadow.
/// By rendering the volume into the stencil buffer we can mask out all pixels
/// that aren't in areas where the shadow is cast.
///
/// Typical use:
/// - (startup) cache information about casting meshes `Shadow::new()`
/// - (every frame) recalculate shadow volumes `Shadow::update()`
/// - (every frame) draw shadow volumes into stencil `Shadow::draw_start()`, `Shadow::draw_add()`
/// - (every frame) shade areas within stencil (draw a full screen quad) `Shadow::draw_end()`
pub struct Shadow {
	edges: Box<[Edge]>,
	triangles: Box<[Triangle]>,
	vertices: Box<[Vertex]>,
	volume: DynamicMesh<engine::Vertex>,
}
impl Shadow {
	/// Build casting information.
	/// This is non-trivial, so ideally do it once at startup.
	pub fn new(window: &mut Window, mesh: &MeshData) -> Self {
		debug_assert!(mesh.indices.len()%3 == 0);
		debug_assert!(mesh.indices.len()/3 <= 256, "detailed meshes perform badly with shadow volumes");

		// Deduplicate vertices.
		let mut indices = mesh.indices.clone();
		let mut vertices: Vec<Vertex> = Vec::new();
		for indices in indices.chunks_exact_mut(3) {
			indices.iter_mut().for_each(|index| {
				let position = F3::from_array(mesh.vertices[*index as usize].position);
				let limit = 0.0001;
				for (new_index, vertex) in vertices.iter().enumerate() {
					if (vertex.position-position).length2() < limit*limit {
						*index = Index::try_from(new_index).unwrap();
						return;
					}
				}
				*index = Index::try_from(vertices.len()).unwrap();
				vertices.push(Vertex{position, projection: None});
			});
		}

		// Build edge and face information and caches.
		// Edges need to reference both attached triangles, so are temporarily
		// stored in `incomplete_edges` until we've found both triangles.
		let mut edges = Vec::new();
		let mut incomplete_edges = Vec::new();
		let mut triangles = Vec::with_capacity(indices.len()/3);
		let invalid = Index::MAX; // used to mark uninitialised triangle indices
		for indices in indices.chunks_exact(3) {
			let vertices = [
				vertices[indices[0] as usize].position,
				vertices[indices[1] as usize].position,
				vertices[indices[2] as usize].position,
			];
			let normal = if let Some(normal) = (vertices[2]-vertices[0]).cross(vertices[1]-vertices[0]).unit() {
				normal
			} else {
				eprintln!("WARNING: discarded triangle with zero normal from shadow caster");
				continue;
			};
// HACK
			// let normal = (vertices[2]-vertices[0]).cross(vertices[1]-vertices[0])
				// .unit().expect("zero normal in shadow casting mesh");
			let index = Index::try_from(triangles.len()).unwrap();
			debug_assert!(index != invalid);
			triangles.push(Triangle {
				normal,
				offset: vertices[0].dot(normal),
				shadowed: false,
				vertices: [indices[0], indices[1], indices[2]],
			});
			for vertices in [
				[indices[0], indices[1]],
				[indices[1], indices[2]],
				[indices[2], indices[0]],
			].iter().copied() {
				// This might be technically possible to support, but needs testing.
// HACK
				// debug_assert!(!edges.iter().any(|edge: &Edge| 
				// 	(edge.vertices[0] == vertices[0] && edge.vertices[1] == vertices[1])
				// 		|| (edge.vertices[1] == vertices[0] && edge.vertices[0] == vertices[1])),
				// 	"edge shared by more than two triangles");
				debug_assert!(!incomplete_edges.iter()
					.any(|edge: &Edge| edge.vertices[0] == vertices[0] && edge.vertices[1] == vertices[1]),
					"edge shared by more than two triangles, or with one triangle inside out");
				// Look for an edge shared with another face that we already found.
				// The vertices will be in the opposite direction due to winding order
				// (assuming both triangles point outward or inward).
				if let Some(edge) = incomplete_edges.iter()
					.position(|edge| edge.vertices[1] == vertices[0] && edge.vertices[0] == vertices[1])
				{
					// We already saw the other half of this edge, complete it.
					let mut edge = incomplete_edges.swap_remove(edge);
					debug_assert!(edge.triangles[0] != invalid);
					debug_assert!(edge.triangles[0] < index);
					debug_assert!(edge.triangles[1] == invalid);
					edge.triangles[1] = index;
					edges.push(edge);
				} else {
					incomplete_edges.push(Edge {
						vertices,
						triangles: [index, invalid],
					});
				}
			}
		}
		// This might be technically possible to support, but needs testing.
// HACK
		if incomplete_edges.len() > 0 {
			eprintln!("WARNING: discarded {} edges with only one triangle from shadow caster",
				incomplete_edges.len());
		}
		// debug_assert!(incomplete_edges.len() == 0, "edge used in only one triangle");

		Self {
			edges: edges.into_boxed_slice(),
			triangles: triangles.into_boxed_slice(),
			vertices: vertices.into_boxed_slice(),
			volume: DynamicMesh::new(window, Primitive::Triangles),
		}
	}
	/// Create a new shadow using the cached casting mesh from an existing
	/// shadow.
	pub fn clone(&self, window: &mut Window) -> Self {
		Self {
			edges: self.edges.clone(),
			triangles: self.triangles.clone(),
			vertices: self.vertices.clone(),
			volume: DynamicMesh::new(window, Primitive::Triangles),
		}
	}
	/// Recalculate shadow volume.
	/// The model transform (`scale`, `rotate`, `translate`) is needed because shadows are
	/// calculated in model coordinates (so that vertices can be transformed on
	/// the GPU).
	/// `light` is the light position in world coordinates.
	/// `distance` is how far the shadow reaches from the casting mesh in model coordinates.
	pub fn update(&mut self, scale: F3, rotate: Rotation, translate: F3, light: F3, distance: f32) {
		// Build a shadow volume by extruding faces of the casting mesh that are
		// facing away from the light to an arbitrary distance.
		// The light is modelled as a point source.

		// Convert light position to model coordinates.
		let light = rotate.inverse().apply(light-translate);
		let light = f3(light.x/scale.x, light.y/scale.y, light.z/scale.z);

		// Clear cache.
		self.volume.clear();
		self.vertices.iter_mut().for_each(|vertex| vertex.projection = None);

		// Test which side the light is for every face and cache the result.
		// This lets us find out which triangles are back-facing and therefore
		// casting shadows.
		// For triangles which are in shadow, we add the corresponding top and
		// bottom parts of the shadow volume, and cache the extrusions of the
		// vertcies to use for building the sides of the volume below.
		for triangle in self.triangles.iter_mut() {
			// Check if in shadow.
			triangle.shadowed = triangle.normal.dot(light)-triangle.offset < 0.0;
			if triangle.shadowed {
				// Extrude vertices.
				for vertex in triangle.vertices.iter().copied() {
					let vertex = &mut self.vertices[vertex as usize];
					if vertex.projection.is_none() {
						let projection = vertex.position+(vertex.position-light).unit_zero()*distance;
						vertex.projection = Some(projection);
					}
				}
				// Add triangle to top of shadow volume.
				for vertex in triangle.vertices.iter().copied().rev() {
					self.volume.push(engine::Vertex {
						position: self.vertices[vertex as usize].position.to_array(),
						normal: [0.0; 3],
						texcoord: [0.0; 2],
					});
				}
				// Add triangle to bottom of shadow volume.
				for vertex in triangle.vertices.iter().copied() {
					self.volume.push(engine::Vertex {
						position: self.vertices[vertex as usize].projection.unwrap().to_array(),
						normal: [0.0; 3],
						texcoord: [0.0; 2],
					});
				}
			}
		}

		// Find edges that border shadowed triangles and extrude them to build the
		// rest of the shadow volume.
		for edge in self.edges.iter() {
			let shadowed = [
				self.triangles[edge.triangles[0] as usize].shadowed,
				self.triangles[edge.triangles[1] as usize].shadowed,
			];
			if shadowed[0] != shadowed[1] {
				// Add quad to side of shadow volume.
				// The order of the vertices in the edge need flipping depending on which
				// connected triangle is within the shadow (the edge is stored reversed
				// relative to the second triangle).
				// This ensures that the shadow volume's triangles consistently point
				// outward.
				// We choose to make the edge go left to right, as seen from outside the
				// volume when the casting triangle is above and pointed downward.
				//        * light
				//
				//   ===caster===
				// 0 o>>>edge>>>o 1
				//   |        / |
				//   |  A   /   |
				//   |   /   B  |
				//   | /        |
				//   o----------o
				let vertices = if shadowed[0] {[
					&self.vertices[edge.vertices[1] as usize],
					&self.vertices[edge.vertices[0] as usize],
				]} else {[
					&self.vertices[edge.vertices[0] as usize],
					&self.vertices[edge.vertices[1] as usize],
				]};
				// A.
				let mut vertex = engine::Vertex {
					position: vertices[0].position.to_array(),
					normal: [0.0; 3],
					texcoord: [0.0; 2],
				};
				self.volume.push(vertex.clone());
				vertex.position = vertices[0].projection.unwrap().to_array();
				let bottom_left = self.volume.push(vertex.clone());
				vertex.position = vertices[1].position.to_array();
				let top_right = self.volume.push(vertex.clone());
				// B.
				vertex.position = vertices[1].projection.unwrap().to_array();
				self.volume.push(vertex);
				self.volume.repeat(top_right);
				self.volume.repeat(bottom_left);
			}
		}
	}
	/// Setup the stencil buffer for drawing shadow volumes into.
	/// Do not try to do other rendering until `draw_end()` is called, strange
	/// things will occur!
	pub fn draw_start(window: &mut Window) -> ShadowContext { unsafe {
		let gl = window.gl();
		let mut context = ShadowContext {
			depth_test: GL_FALSE,
			stencil_half: 0,
			valid: false,
		};

		// Check we have a usable stencil buffer.
		let mut bits = 0;
		(gl.GetIntegerv)(GL_STENCIL_BITS, &mut bits);
		debug_assert!(bits >= 8);
		if bits < 8 { return context; }
		context.valid = true;

		// Store current state.
		(gl.GetBooleanv)(GL_DEPTH_TEST, &mut context.depth_test);

		// Don't touch colour or depth.
		(gl.ColorMask)(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
		(gl.DepthMask)(GL_FALSE);

		// Count number of layers of shadow we see fail depth test.
		// This is sometimes called Carmack's Reverse.
		(gl.Enable)(GL_DEPTH_TEST);
		(gl.Disable)(GL_CULL_FACE);
		(gl.Enable)(GL_STENCIL_TEST);
		(gl.StencilFunc)(GL_ALWAYS, 0, 0xff);
		(gl.StencilOpSeparate)(GL_BACK, GL_KEEP, GL_INCR, GL_KEEP);
		(gl.StencilOpSeparate)(GL_FRONT, GL_KEEP, GL_DECR, GL_KEEP);

		// Clear count to half of range to avoid loosing out of order operations.
		context.stencil_half = 1<<(bits-1);
		(gl.ClearStencil)(context.stencil_half);
		// Clearing only the stencil buffer on the Raspberry Pi Zero W (revision 1.1)
		// seems to mess up the depth buffer, so we rely on doing it in
		// `Window::depth()` when the depth buffer is cleared instead.
		// This does mean that the clear value won't be set properly during the first
		// frame.
		// It also means that doing more than one separate shadow pass will break.
		// (gl.Clear)(GL_STENCIL_BUFFER_BIT);

		context
	} }
	/// Draw this shadow volume into the stencil buffer using `shader`.
	/// If using a camera transform it must be passed to the shader via uniforms
	/// before you call this function.
	/// Any colour drawn by the shader will be ignored.
	/// This function can be called for multiple different shadows before calling
	/// `draw_end()`.
	pub fn draw_add(&mut self, window: &mut Window, context: &ShadowContext, shader: &ActiveShader<engine::Vertex>) {
		if !context.valid { return; }
		self.volume.draw_keep(window, shader);
	}
	/// Draw a completed shadow onto the scene by filling the screen with
	/// `shader`.
	/// Each pixel is only drawn once.
	/// Any shader uniforms and blend modes should be set before calling this
	/// function.
	/// No model or view transform should be used.
	pub fn draw_end(window: &mut Window, context: ShadowContext, fill: &Fill, shader: &ActiveShader<engine::Vertex>) {
		if !context.valid { return; }
		unsafe {
			// Fill masked areas.
			let gl = window.gl();
			(gl.ColorMask)(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
			(gl.Disable)(GL_DEPTH_TEST);
			(gl.StencilFunc)(GL_LESS, context.stencil_half, 0xff);
			(gl.StencilOp)(GL_KEEP, GL_KEEP, GL_KEEP);
		}
		fill.mesh.draw(window, shader);
		unsafe {
			// Restore state.
			let gl = window.gl();
			(gl.Enable)(GL_CULL_FACE);
			(gl.Disable)(GL_STENCIL_TEST);
			(gl.DepthMask)(GL_TRUE);
			if context.depth_test == GL_TRUE {
				(gl.Enable)(GL_DEPTH_TEST);
			} else {
				(gl.Disable)(GL_DEPTH_TEST);
			}
		}
	}
}

/// Temporary storage for data while drawing shadows.
pub struct ShadowContext {
	/// State of depth test before drawing shadows.
	depth_test: GLboolean,
	/// Half of available range of stencil buffer.
	stencil_half: GLint,
	/// `true` if we were able to setup shadow drawing correctly.
	valid: bool,
}

#[derive(Clone)]
struct Vertex {
	/// Vertex in model coordinates.
	position: F3,
	/// Cache for the extrusion of the vertex away from the light source.
	projection: Option<F3>,
}

#[derive(Clone)]
struct Edge {
	/// Indices of the two triangles that share this edge.
	/// We disallow meshes where an edge isn't shared by exactly two triangles.
	triangles: [Index; 2],
	/// Indices of the vertices at each end of this edge.
	/// In the order they were found in `triangles[0]`, and the inverse of their
	/// order in `triangles[1]`.
	/// This is important for getting the outside of the shadow volume to point in
	/// a useful direction.
	vertices: [Index; 2],
}

#[derive(Clone)]
struct Triangle {
	/// Direction the triangle faces in model coordinates.
	/// Doesn't need to be a unit vector (but is at the moment).
	normal: F3,
	/// Combined with `normal` defines the plane of the triangle.
	/// This allows us to tell which side of the triangle a light is.
	offset: f32,
	/// Cache for whether or not the triangle is facing away from the light and
	/// casting a shadow as a result.
	shadowed: bool,
	/// Indices of the vertices at the points of this triangle.
	vertices: [Index; 3],
}
