//! This file is part of Flight.
//! 
//! Flight is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
//! 
//! Flight is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//! You should have received a copy of the GNU General Public License along with Flight. If not, see <https://www.gnu.org/licenses/>.

use engine::{ActiveShader, DynamicMesh, F3, Primitive, Vertex, Window};
use engine::ffi_gles::*;

pub struct Trail {
	counter: u32,
	mesh: DynamicMesh<Vertex>,
	points: [F3; Self::LENGTH],
}
impl Trail {
	const LENGTH: usize = 8;
	const INTERVAL: u32 = 4;

	pub fn new(window: &mut Window, source: F3) -> Self {
		debug_assert!(Self::LENGTH > 1);
		Self {
			counter: 0,
			mesh: DynamicMesh::new(window, Primitive::Triangles),
			points: [source; Self::LENGTH],
		}
	}
	pub fn update(&mut self, source: F3) {
		// First point is always `source` so that trail stays attached to things.
		self.points[0] = source;

		// Update mesh before shuffling point list so that no two points have the
		// same position.
		// We linear interpolate between points to make it look like the trail moves
		// every frame, but without having to add a new point every frame.
		// Texture coordinates are from x=0 at the source to x=1 at the end of the
		// trail, and y=0 to y=1 from bottom to top of the trail.
		// When at a frame where `counter` hasn't just reset, we end up building
		// quads that look like this (positions of points marked by numbers):
		//                _________________
		// <<<<<< <<<<<< |      / |      / |
		// source motion 0   1    |   2    |   3
		// <<<<<< <<<<<< |_/______|_/______|
		//
		// The upward direction added to points to find the top and bottom edge of
		// the trail are computed based on where the camera is.
		// The trail behaves a bit like a billboard sprite in that it tries to always
		// face the viewer.
		// In practice, this billboarding effect is applied in the vertex shader
		// because the GPU is likely to be able to do it more efficiently, and to
		// avoid recalculating the mesh for every camera.
		self.mesh.clear();
		let fraction = (self.counter+1) as f32/Self::INTERVAL as f32;
		debug_assert!(fraction > 0.0);
		debug_assert!(fraction <= 1.0);
		let u_step = 1.0/(Self::LENGTH-1) as f32;
		let mut u = 0.0;
		let direction = self.points[1]-self.points[0];
		self.mesh.push(Vertex {
			position: self.points[0].to_array(),
			normal: direction.to_array(),
			texcoord: [u, 1.0],
		});
		let mut last_bottom = self.mesh.push(Vertex {
			position: self.points[0].to_array(),
			normal: direction.to_array(),
			texcoord: [u, 0.0],
		});
		for index in 2..self.points.len() {
			u += u_step;
			let direction = self.points[index]-self.points[index-1];
			let point = (self.points[index-1]-self.points[index])*fraction+self.points[index];
			// Assuming we just pushed a `last_top` and `last_bottom`, in that order.
			// Use those existing vertices to build a quad.
			let top = self.mesh.push(Vertex {
				position: point.to_array(),
				normal: direction.to_array(),
				texcoord: [u, 1.0],
			});
			self.mesh.repeat(last_bottom);
			let bottom = self.mesh.push(Vertex {
				position: point.to_array(),
				normal: direction.to_array(),
				texcoord: [u, 0.0],
			});
			self.mesh.repeat(top);
			// If this isn't the last quad, push vertices to start the next quad.
			if index != self.points.len()-1 {
				self.mesh.repeat(top);
				self.mesh.repeat(bottom);
			}
			last_bottom = bottom;
		}

		// Shuffle points along the array every few frames.
		self.counter += 1;
		if self.counter >= Self::INTERVAL {
			self.counter = 0;
			for index in (1..self.points.len()).rev() {
				self.points[index] = self.points[index-1];
			}
		}
	}
	pub fn draw(&mut self, window: &mut Window, shader: &ActiveShader<Vertex>) {
		unsafe {
			let gl = window.gl();
			(gl.Disable)(GL_CULL_FACE);
		}
		self.mesh.draw_keep(window, shader);
		unsafe {
			let gl = window.gl();
			(gl.Enable)(GL_CULL_FACE);
		}
	}
}
